/*  Copyright © 2017 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 10/03/2017.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_color_evolution_h__
#define __mc_color_evolution_h__ 1


/*   std includes   */


/*   project headers   */
#include <bits/mc-soft-function.h>
#include <bits/mc-engine-opts.h>
#include <bits/mc-sudakov.h>
#include <cxx/random.h>
#include <cxx/stackvector.h>




namespace duct {
  
  namespace __details {

    /*    some simple functions to test color connections    */
    template<ProcessTag _Proc>
    inline int tauL(int i, int l, const color_state<_Proc>& p) {
      return (l == i ? 1 : 0) - (p[l].colorL.prev == i ? 1 : 0);
    }
    
    
    template<ProcessTag _Proc>
    inline int tauR(int i, int l, const color_state<_Proc>& p) {
      return (l == i ? 1 : 0) - (p[l].colorR.prev == i ? 1 : 0);
    }
    
    
    template<ProcessTag _Proc>
    inline int chi_minus(int i, int l, const color_state<_Proc>& p) {
      return (p[l].colorL.prev == i ? 1 : 0) + (p[l].colorR.prev == i ? 1 : 0);
    }
    
    
    template<ProcessTag _Proc>
    inline int chi_plus(int i, int l, const color_state<_Proc>& p) {
      return (p[l].colorL.next == i ? 1 : 0) + (p[l].colorR.next == i ? 1 : 0);
    }
  
    
    template<ProcessTag _Proc>
    inline int chiL_minus(int i, int l, const color_state<_Proc>& p) {
      return (p[l].colorL.prev == i ? 1 : 0);
    }
    
    
    template<ProcessTag _Proc>
    inline int chiL_plus(int i, int l, const color_state<_Proc>& p) {
      return (p[l].colorL.next == i ? 1 : 0);
    }
    
    
    template<ProcessTag _Proc>
    inline int chiR_minus(int i, int l, const color_state<_Proc>& p) {
      return (p[l].colorR.prev == i ? 1 : 0);
    }
    
    
    template<ProcessTag _Proc>
    inline int chiR_plus(int i, int l, const color_state<_Proc>& p) {
      return (p[l].colorR.next == i ? 1 : 0);
    }
    
    
    template<ProcessTag _Proc>
    inline int chi(int k, int l, const color_state<_Proc>& p)
    {
      int ret = 0;
      auto& cL = p[l].colorL;
      auto& cR = p[l].colorR;
      
      if(cL.prev == k) ++ret;
      if(cL.next == k) ++ret;
      
      if(cR.prev == k) ++ret;
      if(cR.next == k) ++ret;
      
      return ret;
    }
    
    /**************************************************************************************************************************/
    
    
    inline auto compare_cvars_real(const splitting_cvars_real& a, const splitting_cvars_real& b) -> bool {
      return std::abs(a.weight) > std::abs(b.weight);
    }
    
    inline auto compare_cvars_virt(const splitting_cvars_virt& a, const splitting_cvars_virt& b) -> bool {
      return std::abs(a.weight) > std::abs(b.weight);
    }

    
    namespace su3 {

      /*    this is the h_{ij}^{kl}({c', c}) color matrix for the Delta H(t) term    */
      template<ProcessTag _Proc>
      inline int h(int i, int j, int l, int k, const color_state<_Proc>& p)
      {
        int h = 0;
        
        
        /*    tldagger|{c}> <{c'}|tk(1-Cdagger)  contribution    */
        int tldagger = (l == i ? 1 : 0) - (p[l].colorL.prev == i ? 1 : 0);
        
        if(tldagger)
        {
          int tk = (p[l].colorR.next == k ? 0 : (p[k].colorR.prev == j ? -1 : 0));
          if(k == j) tk += (p[l].colorR.prev == k ? 0 : 1);
          
          h -= tldagger*tk;
        }
        
        /*    (1-C)tkdagger|{c}> <{c'}|tl  contribution   */
        int tl = (l == j ? 1 : 0) - (p[l].colorR.prev == j ? 1 : 0);
        
        if(tl)
        {
          int tkdagger = (p[l].colorL.next == k ? 0 : (p[k].colorL.prev == i ? -1 : 0));
          if(k == i) tkdagger += (p[l].colorL.prev == k ? 0 : 1);
          
          h -= tl*tkdagger;
        }
        
        
        return h;
      }
      
      template<ProcessTag _Proc>
      inline int lc(int i, int j, int l, int k, const color_state<_Proc>& p)
      {
        return -tauL(i,l, p)*((j == k ? chiR_minus(k,l, p) : 0) - chiR_plus(k,l, p)*chiR_minus(j,k, p))
               -tauR(j,l, p)*((i == k ? chiL_minus(k,l, p) : 0) - chiL_plus(k,l, p)*chiL_minus(i,k, p));
      }
      
      /*    this is the v_{ij}^{kl}({c}) color matrix for the Delta V(t) term    */
      template<ProcessTag _Proc>
      inline int vL(int i, int j, int l, int k, const color_state<_Proc>& p)
      {
        int v = 0;
        
        /*    tk(1-Cdagger)tldagger|{c}> <{c'}|  contribution    */
        int tldagger = (l == i ? 1 : 0) - (p[l].colorL.prev == i ? 1 : 0);
        
        if(tldagger)
        {
          int tk = (p[l].colorL.next == k ? 0 : (p[k].colorL.prev == j ? -1 : 0));
          if(k == j) tk += (p[l].colorL.prev == k ? 0 : 1);
          
          v -= tldagger*tk;
        }
        
        return v;
      }
      
      
      /*    this is the v_{ij}^{kl}({c'}) color matrix for the Delta V(t) term    */
      template<ProcessTag _Proc>
      inline int vR(int i, int j, int l, int k, const color_state<_Proc>& p)
      {
        int v = 0;
        
        /*    |{c}> <{c'}|tl(1-C)tkdagger  contribution    */
        int tl = (l == i ? 1 : 0) - (p[l].colorR.prev == i ? 1 : 0);
        
        if(tl)
        {
          int tkdagger = (p[l].colorR.next == k ? 0 : (p[k].colorR.prev == j ? -1 : 0));
          if(k == j) tkdagger += (p[l].colorR.prev == k ? 0 : 1);
          
          v -= tl*tkdagger;
        }
        
        return v;
      }
    
      
      /*    this is the \tilde{v}_{ij}^{kl}({c}) color matrix for the Delta H(t) term    */
      template<ProcessTag _Proc>
      inline int phL(int i, int j, int l, int k, const color_state<_Proc>& p) {
        return tauL(i,l, p)*tauL(j,k, p);
      }
      
      
      /*    this is the \tilde{v}_{ij}^{kl}({c'}) color matrix for the Delta H(t) term    */
      template<ProcessTag _Proc>
      inline int phR(int i, int j, int l, int k, const color_state<_Proc>& p)
      {
        return tauR(i,l, p)*tauR(j,k, p);
      }
      
   }   //   namespace su3
    
    
    
    namespace u3 {
      
      
      /*    this is the h_{ij}^{kl}({c', c}) color matrix for the Delta H(t) term    */
      template<ProcessTag _Proc>
      inline int h(int i, int j, int l, int k, const color_state<_Proc>& p)
      {
        /*   Glk part is the full color operator    */
        int ret = -tauL(i,l, p)*tauR(j,k, p) - tauL(i,k, p)*tauR(j,l, p);
        
        /*   LC+ part of the full Glk is subtracted    */
        ret -= (l == i && l == j ? chi_plus(k,l, p) : 0);
        ret -= (p[l].colorL.prev == i && p[l].colorR.prev == j ? chi_minus(k,l, p) : 0);
      
        return ret;
      }
    
      
      template<ProcessTag _Proc>
      inline int lc(int i, int j, int l, int k, const color_state<_Proc>& p)
      {
        return (l == i && l == j ? chi_plus(k,l, p) : 0)
        + (p[l].colorL.prev == i && p[l].colorR.prev == j ? chi_minus(k,l, p) : 0);
      }
      

      /*    this is the v_{ij}^{kl}({c}) color matrix for the Delta V(t) term    */
      template<ProcessTag _Proc>
      inline int vL(int i, int j, int l, int k, const color_state<_Proc>& p) {
        return -tauL(i,l, p)*tauL(j,k, p);
      }

      
      /*    this is the v_{ij}^{kl}({c'}) color matrix for the Delta V(t) term    */
      template<ProcessTag _Proc>
      inline int vR(int i, int j, int l, int k, const color_state<_Proc>& p) {
        return -tauR(i,l, p)*tauR(j,k, p);
      }
    
      /*   color matrix in the phase terms    */
      using su3::phL;
      using su3::phR;
      
    }   //  namespace u3
 }  //   namespace __details
  
  
  
  /*    Implementation details of the real shower operator    */
  template<ProcessTag _Proc>
  class real_shower_color_operator
  {
  public:
    /*    default constructor reserves memory   */
    real_shower_color_operator() {
      _TMP_dlk.reserve(50);
      _TMP_leading.reserve(50);
      _TMP_subleading.reserve(50);
    }
    
    
    /*    do the color evolution in the real operator   */
    template<OrderingTag>
    void shower(const shower_engine_opts::color_opts&, random_engine_type&, shower_history_stage<_Proc>&);
    
    
  private:
    /*   engine parameters    */
    random_engine_type *_M_random;
    
    
    /*    all the other weights for emitter l    */
    std::vector<std::pair<int, double>> _TMP_dlk;
    double _TMP_lmd;
    
    /*   temporary color state    */
    color_state<_Proc> _TMP_cstate;
    
    /*   a vector of splitting variables to select
     *   the active operator for the DeltaH(t) term
     */
    std::vector<splitting_cvars_real> _TMP_leading, _TMP_subleading;
    
    
    /*   gluon emission operators   */
    void _M_lcplus_u3(stackvector<splitting_cvars_real, 4u>&, int, const color_state<_Proc>&);
    void _M_lcplus_su3(stackvector<splitting_cvars_real, 4u>&, int, const color_state<_Proc>&);
    void _M_delta_su3(stackvector<splitting_cvars_real, 4u>&, int, const color_state<_Proc>&);
 
    /*   pure collinear operators    */
    void _M_collinear_u3(stackvector<splitting_cvars_real, 4u>&, int, int, int, const color_state<_Proc>&) const;
    void _M_collinear_su3(stackvector<splitting_cvars_real, 4u>&, int, int, int, const color_state<_Proc>&) const;
  };
  
  
  template<ProcessTag _Proc>
  template<OrderingTag _Ordering>
  void real_shower_color_operator<_Proc>::
  shower(const shower_engine_opts::color_opts& opts, random_engine_type& rne, shower_history_stage<_Proc>& next)
  {
    _M_random = std::addressof(rne);


    /*   make the mapping in the momentum and flavor space     */
    auto boost = shower_mapping(next.state, next.vars);
    
    next.current_time = next.next_time;
    next.scale = next.time_to_scale();
    next.qnull = boost(next.qnull);
    
    
    /*   calculate the lambda probabilities ( = lmd_lk'/sum(lmd_lk, k != l) )  */
    const int emitter = next.vars.emitter;

    //------
    {
      const auto& p = next.state;
      const auto& chi = next.averaged_chi;
      
      const int m = p.upper();
      constexpr int lo = 1-static_cast<int>(number_of_incoming_hadrons<_Proc>);
      
      _TMP_dlk.clear();
      _TMP_lmd = 0.0;

      const auto Q = momentum_sum<incomings>(p);
      const auto &pl = p[emitter].momentum, &pm = p[m].momentum;
      double slQ = pl*Q, slm = pl*pm;
      
      for(int k = lo; k < m; k++)
        if(k != emitter)
        {
          const auto &pk = p[k].momentum;
          double skQ = pk*Q, skm = pk*pm;
          double lmdlk = -mag2(pl*skm-pk*slm)/skm*slQ/(slm*skQ + skm*slQ);
          
          _TMP_dlk.emplace_back(k, lmdlk);
          _TMP_lmd += lmdlk*chi(k, emitter);
        }
    }
    
    
    
    /***    Real shower operator    ***/
    shower_color_state<_Proc>& c = next.cstate;
    stackvector<splitting_cvars_real, 4u> cvars;
    
    auto ci = c.begin();
    auto last = --c.end();
    
    
    if(emitter != next.vars.spectator)
    {
      /*    Correcting the coupling in the delta terms.
       *    In the pure soft term we don't need the enhanced coupling.
       */
      const qcd_coupling& as = next.state.get_model().alpha_qcd();
      double tau = as.tauk(next.scale);
      double ask = as.ask(tau);
      double dfac = as.as(tau)/ask;
      
      if constexpr(static_cast<unsigned>(_Ordering::alpha))
        dfac *= 1.0 + static_cast<unsigned>(_Ordering::alpha)*ask*qcd::beta1(as.nf(tau))*std::log(1.0-next.vars.z);
 
      
      /*   making the transformations    */
      do
      {
        bool is_zero = true;
        
        /*   Delta H(t) contributions   */
        if(ci->is_color_u3 == false && ci->size() > 2 && ci->order_of_soft_corrections < opts.soft_corrections)
        {
          _M_delta_su3(cvars, emitter, *ci);
          
          if(cvars.size())
          {
            c.transform_keep(ci, cvars.begin(), cvars.end(),
                             [=](color_state<_Proc>& ci) {
                               ci.order_of_soft_corrections++;
                               ci.order_of_softH_corrections++;
                               ci.weight *= dfac;
                             });
            is_zero = false;
          }
        }
        
        
        /*   LC+ part of the gluon emission    */
        if(ci->is_color_u3) _M_lcplus_u3(cvars, emitter, *ci);
        else _M_lcplus_su3(cvars, emitter, *ci);
        
        
        if(cvars.size())
        {
          c.transform(ci, cvars.begin(), cvars.end());
          is_zero = false;
        }

        /*   this is zero operator   */
        if(is_zero) ci->weight = 0.0;
      } while(ci++ != last);
    }
    
    /*   pure collinear oprators    */
    else
    {
      const int cl = next.vars.cl, cm = next.vars.cm;

      do
      {
        if(ci->is_color_u3) _M_collinear_u3(cvars, emitter, cl, cm, *ci);
        else _M_collinear_su3(cvars, emitter, cl, cm, *ci);
        
        c.transform(ci, cvars.begin(), cvars.end());
      } while(ci++ != last);
    }
    
    /*    remove all the zero color states    */
    c.reduce(1e-15);
  }

  
  template<ProcessTag _Proc>
  void real_shower_color_operator<_Proc>::
  _M_lcplus_su3(stackvector<splitting_cvars_real, 4u>& ret, int emitter, const color_state<_Proc>& c)
  {
    /*   make some space    */
    _TMP_cstate = c;
    _TMP_cstate.resize_by(1);
    ret.clear();
    
    
    /*   try all the possible insertions   */
    const int lo = c.lower(), hi = c.upper();
    const double fac = 0.5*qcd::Cf/_TMP_lmd;
    
    for(int i = lo; i <= hi; ++i)
      for(int j = lo; j <= hi; ++j)
        if(c[i].colorL.is_antiquark() == false && c[j].colorR.is_antiquark() == false)
        {
          /*   weight of the operator   */
          double lc = 0.0;
          
          for(const auto& d : _TMP_dlk)
            lc += __details::su3::lc(i,j, emitter, d.first, c)*d.second;
          
          /*    LC+ operators   */
          if(std::abs(lc) > 1e-15)
          {
            int suppr_idx = ('f' == health_with_gluon_loop(i, j, _TMP_cstate) ? 2 : 0);
            ret.push_back({1,1, i, j, fac*lc, static_cast<unsigned>(suppr_idx), suppr_idx});
          }
        }
  }
  
  
  template<ProcessTag _Proc>
  void real_shower_color_operator<_Proc>::
  _M_lcplus_u3(stackvector<splitting_cvars_real, 4u>& ret, int emitter, const color_state<_Proc>& c)
  {
    /*   make some space    */
    _TMP_cstate = c;
    _TMP_cstate.resize_by(1);
    ret.clear();
    
    
    /*   try all the possible insertions   */
    const int lo = c.lower(), hi = c.upper();
    const double fac = 0.25*qcd::Ca/_TMP_lmd;

    for(int i = lo; i <= hi; ++i)
      for(int j = lo; j <= hi; ++j)
        if(c[i].colorL.is_antiquark() == false && c[j].colorR.is_antiquark() == false)
          if('h' == health_with_gluon_loop(i, j, _TMP_cstate))
          {
            /*   weight of the operator   */
            double lc = 0.0;
            
            for(const auto& d : _TMP_dlk)
              lc += __details::u3::lc(i,j, emitter, d.first, c)*d.second;
            
            
            /*   sorting the current operator   */
            if(std::abs(lc) > 1e-15)
              ret.push_back({1,1, i, j, fac*lc, 0u, 0});
          }
  }
  

  template<ProcessTag _Proc>
  void real_shower_color_operator<_Proc>::
  _M_collinear_su3(stackvector<splitting_cvars_real, 4u>& lcplus, int l, int cl, int cm, const color_state<_Proc>& c) const
  {
    lcplus.clear();
    
    /*  spliting e.g.: g -> g + g    */
    if(cl == 8 && cm == 8)
    {
      int suppr_idx = (c[l].health == 'h' ? 2 : 0);
      
      lcplus.push_back({ 1,  1, l, l,  qcd::Cf/qcd::Ca, 0u, 0});
      lcplus.push_back({-1, -1, l, l,  qcd::Cf/qcd::Ca, 0u, 0});
      lcplus.push_back({ 1, -1, l, l, -qcd::Cf/qcd::Ca, static_cast<unsigned>(suppr_idx), suppr_idx});
      lcplus.push_back({-1,  1, l, l, -qcd::Cf/qcd::Ca, static_cast<unsigned>(suppr_idx), suppr_idx});
    }
    
    /*   splitting e.g.: q -> q + g   */
    else if(cl == 3 && cm == 8) lcplus.push_back({1, 1, l, l, 1.0, 0u, 0});
    
    /*   splitting e.g.: q~ -> q~ + g   */
    else if(cl == -3 && cm == 8) lcplus.push_back({-1,-1, l, l, 1.0, 0u, 0});
    
    /*  spliting e.g.: g -> q + q~   */
    else if(std::abs(cl) == 3 && cm == -cl)
    {
      int sigma = (cl == 3 ? 2 : 3);
      int suppr_idx = (c[l].health == 'f' ? -2 : 0);
      unsigned suppr_uidx = (c[l].health == 'f' ? 0u : 2u);

      bool tadL = (c[l].colorL.next == c[l].colorL.prev ? true : false);
      bool tadR = (c[l].colorR.next == c[l].colorR.prev ? true : false);
      
      lcplus.push_back({ sigma,  sigma, l, l,  qcd::Nc2/qcd::Na, 0u, 0});

      if(!tadL && !tadR) lcplus.push_back({-sigma, -sigma, l, l,  1.0/qcd::Na, suppr_uidx, suppr_idx});
      if(true  && !tadR) lcplus.push_back({ sigma, -sigma, l, l, -qcd::Nc/qcd::Na, suppr_uidx, 1+suppr_idx});
      if(!tadL &&  true) lcplus.push_back({-sigma,  sigma, l, l, -qcd::Nc/qcd::Na, suppr_uidx, 1+suppr_idx});
    }
    
    /*  splitting e.g.: q -> g + q in the initial state   */
    else if(cl == 8 && cm == 3) lcplus.push_back({4, 4, l, l, 1.0, 0u, 0});
    
    /*  splitting e.g.: q~ -> g + q~ in the initial state    */
    else if(cl == 8 && cm == -3) lcplus.push_back({-4, -4, l, l, 1.0, 0u, 0});
    
    
    /*   unkown splitting    */
    else throw "color_evolution<_Proc>::_M_collinear_operators_su3(...) : unable to assign the new color state";
  }
  
  
  template<ProcessTag _Proc>
  void real_shower_color_operator<_Proc>::
  _M_collinear_u3(stackvector<splitting_cvars_real, 4u>& lcplus, int l, int cl, int cm, const color_state<_Proc>&) const
  {
    lcplus.clear();
    
    /*  splitting e.g.: g -> g + g    */
    if(cl == 8 && cm == 8)
    {
      lcplus.push_back({ 1,  1, l, l, 0.5, 0u, 0});
      lcplus.push_back({-1, -1, l, l, 0.5, 0u, 0});
    }
    
    /*   splitting e.g.: q -> q + g   */
    else if(cl == 3 && cm == 8) lcplus.push_back({1, 1, l, l, 1.0, 0u, 0});
    
    /*   splitting e.g.: q~ -> q~ + g   */
    else if(cl == -3 && cm == 8) lcplus.push_back({-1,-1, l, l, 1.0, 0u, 0});
    
    /*  spliting e.g.: g -> q + q~   */
    else if(cl == 3 && cm == -cl) lcplus.push_back({2, 2, l, l, 1.0, 0u, 0});
    else if(cl ==-3 && cm == -cl) lcplus.push_back({3, 3, l, l, 1.0, 0u, 0});
    
    /*  splitting e.g.: q -> g + q in the initial state   */
    else if(cl == 8 && cm == 3) lcplus.push_back({4, 4, l, l, 1.0, 0u, 0});
    
    /*  splitting e.g.: q~ -> g + q~ in the initial state    */
    else if(cl == 8 && cm == -3) lcplus.push_back({-4, -4, l, l, 1.0, 0u, 0});
    
    
    /*   unkown splitting    */
    else throw "color_evolution<_Proc>::_M_lcplus_operators_u3(...) : unable to assign the new color state";
  }
  
  
  
  template<ProcessTag _Proc>
  void real_shower_color_operator<_Proc>::
  _M_delta_su3(stackvector<splitting_cvars_real, 4u>& ret, int emitter, const color_state<_Proc>& c)
  {
    /*   make some space    */
    int lo = c.lower(), hi = c.upper();
    
    _TMP_cstate = c;
    _TMP_cstate.resize_by(1);
    _TMP_leading.clear();
    _TMP_subleading.clear();
    ret.clear();
    
    /*   try all the possible insertions   */
    const double fac = 0.5*qcd::Cf/_TMP_lmd;

    for(int i = lo; i <= hi; ++i)
      for(int j = lo; j <= hi; ++j)
        if(c[i].colorL.is_antiquark() == false && c[j].colorR.is_antiquark() == false)
        {
          /*   weight of the operator   */
          double h = 0.0;
          
          for(const auto& d : _TMP_dlk)
            h += __details::su3::h(i,j, emitter, d.first, c)*d.second;
          
          
          /*    sorting the current operator   */
          if(std::abs(h) > 1e-15)
          {
            if('f' == health_with_gluon_loop(i, j, _TMP_cstate))
              _TMP_subleading.push_back({1,1, i, j, fac*h, 2u, 2u});
            else _TMP_leading.push_back({1,1, i, j, fac*h, 0u, 0});
          }
        }
    
 
    /*   selecting the active leading color vertices   */
    if(!_TMP_leading.empty())
    {
      /*    selecting the active DeltaH(t) operators    */
      std::sort(_TMP_leading.begin(), _TMP_leading.end(), __details::compare_cvars_real);
      
      auto io = select_signed_pair_randomly(*_M_random, _TMP_leading.begin(), _TMP_leading.end(),
                                            [](const splitting_cvars_real& vi) {
                                              return vi.weight;
                                            });
      
      /*   adjusting the weight for positive operator    */
      if(io.selectedP != _TMP_leading.end())
      {
        io.selectedP->weight = io.normP;
        ret.push_back(*io.selectedP);
      }
      
      /*   and for negative operator    */
      if(io.selectedM != _TMP_leading.end())
      {
        io.selectedM->weight = -io.normM;
        ret.push_back(*io.selectedM);
      }
    }
    
    
    /*   selecting the active subleading color vertices   */
    if(!_TMP_subleading.empty())
    {
      /*    selecting the active DeltaH(t) operators    */
      std::sort(_TMP_subleading.begin(), _TMP_subleading.end(), __details::compare_cvars_real);
      
      auto io = select_signed_pair_randomly(*_M_random, _TMP_subleading.begin(), _TMP_subleading.end(),
                                            [](const splitting_cvars_real& vi) {
                                              return vi.weight;
                                            });
      
      /*   adjusting the weight for positive operator    */
      if(io.selectedP != _TMP_subleading.end())
      {
        io.selectedP->weight = io.normP;
        ret.push_back(*io.selectedP);
      }
      
      /*   and for negative operator    */
      if(io.selectedM != _TMP_subleading.end())
      {
        io.selectedM->weight = -io.normM;
        ret.push_back(*io.selectedM);
      }
    }
  }
}   //   namespace duct



namespace duct {
  
  
  /*   This class makes the transformation of the color state psi to xi.    */
  template<ProcessTag _Proc>
  class virtual_shower_color_operator
  {
    using _Diff = std::ptrdiff_t;
    
  public:
    /*    X(t2, t1)*Phi(t2, t1)  operator for shower evolution    */
    template<OrderingTag>
    void shower(const shower_engine_opts::color_opts&, random_engine_type&, shower_history_stage<_Proc>&);
    
    
    /*    X(t2, t1)*Phi(t2, t1)  operator for threshold evolution    */
    template<OrderingTag>
    void threshold(const shower_engine_opts::threshold_opts&, random_engine_type&, shower_history_stage<_Proc>&);
    
    
  private:
    /*    data member to store the exponents    */
    soft_function<_Proc> _M_lmd;
    
    /*    temporary storage for calculating the l and k sums of the soft functions    */
    typename soft_function<_Proc>::result_t _TMP_S, _TMP_iS;
    
    
    /*   temporary color state    */
    color_state<_Proc> _TMP_cstate, _TMP_cstate_exp;
    
    /*    list of possible color operators      */
    std::vector<splitting_cvars_virt> _TMP_leading, _TMP_subleading;
    
    /*   engine parameters    */
    random_engine_type *_M_random;
    
    /*   Sudakov operator   */
    sudakov_operator<_Proc> _M_sudakov;
    
  private:
    /*   setting up the DeltaV operator    */
    void _M_delta_su3(stackvector<splitting_cvars_virt, 4u>&, const color_state<_Proc>&);
  };
  
  
  
  template<ProcessTag _Proc>
  template<OrderingTag _Ordering>
  void virtual_shower_color_operator<_Proc>::
  shower(const shower_engine_opts::color_opts& opts, random_engine_type& rne, shower_history_stage<_Proc>& stage)
  {
    /***   initializing the parameters and the temporaries   ***/
    _M_random = std::addressof(rne);
    
    _M_lmd.template init<typename virtual_color_operator_initializer<_Ordering>::shower>(stage);
    _M_lmd.integral(_TMP_iS);

    
    /***   tranforming the color state by the Phi operator  ***/
    stage.cstate.transform([=, this, ipi = opts.ipi_effects,
                            lpsi = std::exp(-soft_exponent(_TMP_iS, stage.averaged_chi))
                            ](color_state<_Proc>& ci) {
      /*   the inverse of the averaged sudakov factor   */
      ci.weight /= lpsi;
      
      /*   when the color state is already U(3), than we have a simple Phi operator     */
      if(!ipi || ci.is_color_u3) ci.weight *= std::exp(-soft_exponent(this->_TMP_iS, ci));
      
      /*    otherwise we have to exponentiate a non-trivial matrix and trasform the current color state     */
      else this->_M_sudakov(ci, *_M_random, this->_TMP_iS);
    });
    
    
    
    /***   transforming the color state by the X operator    ***/
    if(opts.soft_corrections && _M_lmd.tm_min() < _M_lmd.tm_max())
    {
      stackvector<splitting_cvars_virt, 4u> cvars;
      std::uniform_real_distribution<double> rng(0.0,1.0);
      
      auto& c = stage.cstate;
      auto ic = c.begin(), ilast = c.end();
      double jacT, tmT = _M_lmd.tm_min();
      
      
      for(unsigned k = 1; k <= opts.soft_corrections; k++)
      {
        if(ic == c.end()) break;
        
        jacT = _M_lmd.tm_max() - tmT;
        tmT += jacT*rng(*_M_random);
        ilast = --c.end();
        
        /*    calculate the integral of the soft functions over the evolution range [tmT, tmH]  */
        _M_lmd.function(_TMP_S, tmT);
        _M_lmd.integral(_TMP_iS, tmT);

        
        /*   selecting the active virtual operators    */
        do
        {
          if(ic->is_color_u3 == false && ic->order_of_soft_corrections < opts.soft_corrections)
          {
            _M_delta_su3(cvars, *ic);
            
            if(cvars.size())
              c.transform(ic, cvars.begin(), cvars.end(),
                          [=, this](color_state<_Proc>& ci) {
                            ci.order_of_soft_corrections++;
                            ci.order_of_softV_corrections++;
                            ci.weight *= -jacT;
                          });
          }
        } while(ic++ != ilast);
      }
    }
  }
  
  
  
  template<ProcessTag _Proc>
  template<OrderingTag _Ordering>
  void virtual_shower_color_operator<_Proc>::
  threshold(const shower_engine_opts::threshold_opts& opts, random_engine_type& rne, shower_history_stage<_Proc>& stage)
  {
    /***   initializing the parameters and the temporaries   ***/
    _M_random = std::addressof(rne);
    
    _M_lmd.template init<typename virtual_color_operator_initializer<_Ordering>::threshold>(stage);
    _M_lmd.integral(_TMP_iS);
    
    
    /***   tranforming the color state    ***/
    stage.cstate.transform([=, this, lpsi = soft_exponent(_TMP_iS, stage.averaged_chi)](color_state<_Proc>& ci) {
      ci.weight *= std::exp(lpsi-soft_exponent(this->_TMP_iS, ci));
    });
    
    
    /*    transforming phi to xi    */
    if(opts.order_of_soft_corrections && _M_lmd.tm_min() < _M_lmd.tm_max())
    {
      stackvector<splitting_cvars_virt, 4u> cvars;
      std::uniform_real_distribution<double> rng(0.0,1.0);
      
      auto& c = stage.cstate;
      auto ic = c.begin(), ilast = c.end();
      double jacT, tmT = _M_lmd.tm_min();
      
      
      for(unsigned k = 1; k <= opts.order_of_soft_corrections; k++)
      {
        if(ic == c.end()) break;
        
        jacT = _M_lmd.tm_max() - tmT;
        tmT += jacT*rng(*_M_random);
        ilast = --c.end();
        
        /*    calculate the integral of the soft functions over the evolution range [tmT, tmH]  */
        _M_lmd.function(_TMP_S, tmT);
        _M_lmd.integral(_TMP_iS, tmT);
        
        
        /*   selecting the active virtual operators    */
        do
        {
          /*   delta insertions    */
          if(ic->order_of_softT_corrections < opts.order_of_soft_corrections)
          {
            _M_delta_su3(cvars, *ic);
            
            if(cvars.size())
              c.transform(ic, cvars.begin(), cvars.end(),
                          [=](color_state<_Proc>& ci) {
                            ci.order_of_softT_corrections++;
                            ci.weight *= -jacT;
                          });
          }
        } while(ic++ != ilast);
      }
    }
  }
  
  
  
  template<ProcessTag _Proc>
  void virtual_shower_color_operator<_Proc>::
  _M_delta_su3(stackvector<splitting_cvars_virt, 4u>& ret, const color_state<_Proc>& ci)
  {
    /*    initialize temporaries    */
    int lo = ci.lower(), hi = ci.upper();
    
    _TMP_cstate = ci;
    _TMP_cstate.resize(hi+1);
    _TMP_leading.clear();
    _TMP_subleading.clear();
    ret.clear();
   
    
    /*   a lambda expression to check the vertices    */
    auto check_vertex = [&, this, lcc = soft_exponent(_TMP_iS, ci)](splitting_cvars_virt vi) -> void
    {
      shower_mapping(_TMP_cstate_exp, ci, vi);
      
      if(!check_for_tadpoles(_TMP_cstate_exp))
      {
        vi.weight *= std::exp(lcc-soft_exponent(_TMP_iS, _TMP_cstate_exp));
        
        if(vi.state_idx == 0u) _TMP_leading.push_back(vi);
        else _TMP_subleading.push_back(vi);
      }
    };

    
    /*    we have to check all the possible insertions    */
    double v1 = 0.0;

    for(int i = lo; i <= hi; ++i)
      if(ci.is_antitriplet(i) == false)
      {
        /*   off-diagonal components    */
        for(int j = lo; j <= hi; ++j)
          if(i != j && ci.is_antitriplet(j) == false)
          {
            /*   calculate the v coefficients   */
            double vL = 0.0, vR = 0.0;
            
            for(int l = lo; l <= hi; ++l)
              for(int k = lo; k <= hi; ++k)
                if(k != l)
                {
                  double Slk = 0.25*_TMP_S[l][k];
                  
                  vL += Slk*__details::su3::vL(i,j, l,k, ci);
                  vR += Slk*__details::su3::vR(i,j, l,k, ci);
                }
            
            /*   weight of the unit operator    */
            v1 += vL+vR;
            
            
            /*   nothing to do if the gluon insertion leads to tadpoles   */
            if(ci[i].colorL.next == j || ci[j].colorL.next == i) vL = 0.0;
            if(ci[i].colorR.next == j || ci[j].colorR.next == i) vR = 0.0;
            if(std::abs(vL) < 1e-15 && std::abs(vR) < 1e-15) continue;
            
            
            /*    sorting the vertices   */
            int suppr_idx = ('f' == health_with_gluon_loop(i, j, _TMP_cstate) ? 1 : -1);
            
            if(std::abs(vL) > 1e-15) check_vertex({'G','L', i, j, vL, static_cast<unsigned>(suppr_idx+1), suppr_idx});
            if(std::abs(vR) > 1e-15) check_vertex({'G','R', i, j, vR, static_cast<unsigned>(suppr_idx+1), suppr_idx});
          }
      }
    
    /*    subleading color part      */
    if(std::abs(v1) > 1e-15) _TMP_subleading.push_back({'U','-', -200, -200, -v1/qcd::Nc, 2u, 0u});
 
    
    /*   selecting the active leading color vertices   */
    if(!_TMP_leading.empty())
    {
      /*    selecting the active DeltaH(t) operators    */
      std::sort(_TMP_leading.begin(), _TMP_leading.end(), __details::compare_cvars_virt);
      
      auto io = select_signed_pair_randomly(*_M_random, _TMP_leading.begin(), _TMP_leading.end(),
                                            [](const splitting_cvars_virt& vi) {
                                              return vi.weight;
                                            });
      
      /*   adjusting the weight for positive operator    */
      if(io.selectedP != _TMP_leading.end())
      {
        io.selectedP->weight = io.normP;
        ret.push_back(*io.selectedP);
      }
      
      /*   and for negative operator    */
      if(io.selectedM != _TMP_leading.end())
      {
        io.selectedM->weight = -io.normM;
        ret.push_back(*io.selectedM);
      }
    }
    

    /*   selecting the active subleading color vertices   */
    if(!_TMP_subleading.empty())
    {
      /*    selecting the active DeltaH(t) operators    */
      std::sort(_TMP_subleading.begin(), _TMP_subleading.end(), __details::compare_cvars_virt);
      
      auto io = select_signed_pair_randomly(*_M_random, _TMP_subleading.begin(), _TMP_subleading.end(),
                                            [](const splitting_cvars_virt& vi) {
                                              return vi.weight;
                                            });
      
      /*   adjusting the weight for positive operator    */
      if(io.selectedP != _TMP_subleading.end())
      {
        io.selectedP->weight = io.normP;
        ret.push_back(*io.selectedP);
      }
      
      /*   and for negative operator    */
      if(io.selectedM != _TMP_subleading.end())
      {
        io.selectedM->weight = -io.normM;
        ret.push_back(*io.selectedM);
      }
    }
  }
}   //   namespace duct

#endif
