/*  Copyright © 2015 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 13/10/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_splitfunc_approx_h__
#define __mc_splitfunc_approx_h__ 1



/*   std includes   */

/*   project headers   */
#include <bits/mc-vz-mapping.h>
#include <bits/mc-splitfunc-base.h>



namespace duct {
  
  namespace __details {
    
    /*   This class for representing the approximation of the splitting function
     *   and it can be used for doing the veto algorithm. I don't think I can
     *   provide a generic template definition. The design of the overshooting functions
     *   is a tricky business.
     */
    template<class _Type, class _Ordering>
    class splitting_functions_approx;
    
    
    /*   This is a class to approximate the splittting functions in the initial stae.
     *   Fortunately I can give a general template definition that looks OK for
     *   lambda, kT and angular ordering.
     */
    template<class _Ordering>
    class splitting_functions_approx<ini, _Ordering> : public v_mapping<ini>
    {
      /*   Size of the interpolation spline.
       *   Be careful!!! Large grid can slow down your code...
       */
      static constexpr unsigned _N_grid = 4u;
      
      /*   some type aliases   */
      using _Base = v_mapping<ini>;
      
    public:
      /*   public types    */
      using ordering_type = _Ordering;
      
      /*   default constructor    */
      splitting_functions_approx()
      : _M_soft{_N_grid}, _M_coll{_N_grid}, _M_fccoll{_N_grid} {}
      
      
      /*   binding to the current hard configuration    */
      bool bind(const hard_params<ini> *pars, double tauL, double tauH)
      {
        bool is_frozen = false;
        double frozen = 0.0;
        
        return _Base::bind(ordering_type{}, pars, tauL, tauH,
                           
                           /*   z integral as function of tau    */
                           [=, this, &is_frozen, &frozen](double tau) -> double {
                             if(std::abs(_Base::_M_pars->fl) > 3 && tau > tauT()*0.95)
                             {
                               if(is_frozen) return frozen;
                               else {
                                 tau = tauT()*0.95;
                                 is_frozen = true;
                               }
                             }
                             
                             this->bind(_Base::_M_pars->mdl.alpha_qcd().q2(tau)/_Base::_M_pars->Q2);
                             frozen = this->integral();
                             
                             return frozen;
                           });
      }
      
      /*   binding to the current y variable    */
      void bind(double);
      
      /*    generating the z distribution according to the approximated splitting functions   */
      std::pair<double,double> mapping_chi2z(splitting_type, double) const;
      
      /*   for heavy flavors it returns the threshold tau value   */
      double tauT() const { return _Base::_M_tauT;}
      
      /*    the z integral of the full distribution   */
      double integral() const {
        return _M_soft.integral() + _M_coll.integral() + _M_fccoll.integral();
      }

    private:
      /*   using a constant spline for the various contribution */
      spline<spline_type::constant> _M_soft, _M_coll, _M_fccoll;
      
      /*   maps out the pdfs and the z dependent part of the coupling   */
      z_mapping_ini _M_zmap;
    };
    
    
    template<class _Ordering>
    void splitting_functions_approx<ini, _Ordering>::bind(double v)
    {
      using _Split = __details::splitting_functions_base<ini>;

      /*   binding the z-mapping   */
      _M_zmap.bind<_Ordering>(_Base::_M_pars, v);

      /*    setting up the soft spline    */
      _M_soft.assign(unitweight{}, [=, this](double w) -> double {
        auto z = _M_zmap.w2z(splitting_type::soft, w);
        auto y = ordering_traits<ini,_Ordering>::main::mapping_to_y(_Base::_M_pars, v, z.first);

        return y.second*z.second*_Split::parton_gluon(_Base::_M_pars, y.first, z.first);
      });
      
      
      /*    setting up the collinear spline    */
      _M_coll.assign(unitweight{}, [=, this](double w) -> double {
        auto z = _M_zmap.w2z(splitting_type::collinear, w);
        auto y = ordering_traits<ini,_Ordering>::main::mapping_to_y(_Base::_M_pars, v, z.first);
        
        return y.second*z.second*(_Base::_M_pars->fl
                                  ? _Split::quark_gluon(y.first, z.first)
                                  : _Split::gluon_gluon(y.first, z.first));
      });
      
      
      /*    setting up the flavor changing collinear spline    */
      _M_fccoll.assign(unitweight{}, [=, this](double w) -> double {
        auto z = _M_zmap.w2z(splitting_type::flavor_changing_collinear, w);
        auto y = ordering_traits<ini,_Ordering>::main::mapping_to_y(_Base::_M_pars, v, z.first);
      
        return y.second*z.second*(_Base::_M_pars->fl
                                  ? _Split::quark_antiquark(y.first, z.first)
                                  : _Split::gluon_quark(y.first, z.first));
      });
    }

    
    template<class _Ordering> std::pair<double,double>
    splitting_functions_approx<ini, _Ordering>::mapping_chi2z(splitting_type idx, double chi) const
    {
      /*   chi ==> w using the splines    */
      double w = 0.5, jac = 0.0;
      
      switch(idx) {
        case splitting_type::soft:
          w = _M_soft.integral_inverse(chi*_M_soft.integral());
          jac = _M_soft.integral()/_M_soft(w);
          break;
          
        case splitting_type::collinear:
          w = _M_coll.integral_inverse(chi*_M_coll.integral());
          jac = _M_coll.integral()/_M_coll(w);
         break;
          
        case splitting_type::flavor_changing_collinear:
          w = _M_fccoll.integral_inverse(chi*_M_fccoll.integral());
          jac = _M_fccoll.integral()/_M_fccoll(w);
         break;
      }
      
      /*   from w ==> z  using the pdf integrals    */
      auto z = _M_zmap.w2z(idx, w);
      z.second *= jac;
      
      return z;
    }
 
    
    
    /*    Final state splitting function approximation for any ordering.    */
    template<class _Ordering>
    class splitting_functions_approx<fin, _Ordering> : public v_mapping<fin>
    {
      /*   size of the interpolation grid    */
      static constexpr unsigned int _N_grid = 8u;
      
      /*   some type aliases   */
      using _Base = v_mapping<fin>;
      
    public:
      /*   public types    */
      using ordering_type = _Ordering;
      
      /*   default constructor    */
      splitting_functions_approx() : _M_soft{_N_grid}, _M_coll{_N_grid} {}
      

      /*   binding to the current hard configuration    */
      bool bind(const hard_params<fin> *pars, double tauL, double tauH)
      {
        return _Base::bind(ordering_type{}, pars, tauL, tauH,
                           
                           /*   z integral as function of tau    */
                           [=, this](double tau) -> double {
                             this->bind(_Base::_M_pars->mdl.alpha_qcd().q2(tau)/_Base::_M_pars->pQ);
                             return this->integral();
                           });
      }
      
      /*   binding to the current y variable    */
      void bind(double);
      
      /*    generating the z distribution according to the approximated splitting functions   */
      std::pair<double,double> mapping_chi2z(splitting_type, double) const;
      
      /*    returns the z integral    */
      double integral() const {
        return _M_soft.integral() + _M_coll.integral();
      }
      
    private:
      /*   using a constant spline for the various contribution */
      spline<spline_type::constant> _M_soft, _M_coll;
    
      /*   maps out the z dependent alpha_s and the soft pole     */
      z_mapping_fin<_Ordering> _M_zmap;
    };

    
    template<class _Ordering>
    void splitting_functions_approx<fin, _Ordering>::bind(double v)
    {
      using _Split = __details::splitting_functions_base<fin>;
      
      /*   binding the z-mapping   */
      _M_zmap.bind(_Base::_M_pars, v);
      
 
      /*    setting up the soft spline    */
      _M_soft.assign(unitweight{}, [=, this](double w) -> double {
        auto z = _M_zmap.w2z(splitting_type::soft, w);
        auto y = ordering_traits<fin,_Ordering>::main::mapping_to_y(_Base::_M_pars, v, z.first);
        return y.second*z.second*_Split::parton_gluon(_Base::_M_pars, y.first, z.first);
      });
      
      
      /*    setting up the collinear spline    */
      _M_coll.assign(unitweight{}, [=, this](double w) -> double {
        auto z = _M_zmap.w2z(splitting_type::collinear, w);
        auto y = ordering_traits<fin,_Ordering>::main::mapping_to_y(_Base::_M_pars, v, z.first);

        return y.second*z.second*(_Base::_M_pars->fl
                                  ?  _Split::quark_gluon(_Base::_M_pars, y.first, z.first)
                                  : (_Split::gluon_gluon(_Base::_M_pars, y.first, z.first)
                                     + _Split::quark_antiquark(_Base::_M_pars, y.first, z.first)));
      });
    }
    
    
    template<class _Ordering> std::pair<double,double>
    splitting_functions_approx<fin, _Ordering>::mapping_chi2z(splitting_type idx, double chi) const
    {
      /*   chi ==> w using the splines    */
      double w, jac;
      
      switch(idx) {
        case splitting_type::soft:
          w = _M_soft.integral_inverse(chi*_M_soft.integral());
          jac = _M_soft.integral()/_M_soft(w);
          break;
          
        case splitting_type::collinear:
          w = _M_coll.integral_inverse(chi*_M_coll.integral());
          jac = _M_coll.integral()/_M_coll(w);
          break;
      
        default:
          return {0.5, 0.0};
      }
      
      /*   from w ==> z  using the pdf integrals    */
      auto z = _M_zmap.w2z(idx, w);
      z.second *= jac;
      
      return z;
    }
    
   
  
    /*    Specialization of the splitting function approximation
     *    in the final state for angular ordering.
     */
    template<>
    class splitting_functions_approx<fin, ordering::angular> : public v_mapping<fin>
    {
      /*   size of the interpolation grid    */
      static constexpr unsigned int _N_grid = 4u;
      
      /*   some type aliases   */
      using _Base = v_mapping<fin>;
      
    public:
      /*   public types    */
      using ordering_type = ordering::angular;
      
      /*   default constructor    */
      splitting_functions_approx()
      : _M_softL{_N_grid}, _M_collL{_N_grid}
      , _M_softH{_N_grid}, _M_collH{_N_grid} {}
      
      
      /*   binding to the current hard configuration    */
      bool bind(const hard_params<fin> *pars, double tauL, double tauH)
      {
        return _Base::bind(ordering_type{}, pars, tauL, tauH,
                           
                           /*   z integral as function of tau    */
                           [=, this](double tau) -> double {
                             this->bind(_Base::_M_pars->mdl.alpha_qcd().q2(tau)/_Base::_M_pars->pQ);
                             return this->integral();
                           });
      }
      
      /*   binding to the current y variable    */
      void bind(double);
      
      /*    generating the z distribution according to the approximated splitting functions   */
      std::pair<double,double> mappingL_chi2z(splitting_type, double) const;
      std::pair<double,double> mappingH_chi2z(splitting_type, double) const;
      
      /*    returns the z integral    */
      double integral() const {
        return _M_softL.integral() + _M_collL.integral() + _M_softH.integral() + _M_collH.integral();
      }
      
    private:
      /*   using a constant spline for the various contribution */
      spline<spline_type::constant> _M_softL, _M_collL, _M_softH, _M_collH;
      
      /*   maps out the z dependent alpha_s and the soft pole     */
      z_mapping_fin<ordering_type> _M_zmapL, _M_zmapH;
    };
    
    
  }   //  namespace __details
  
  
}

#endif
