/*  Copyright (C) 2014 Zoltán Nagy
 *  Created by Zoltán Nagy on 2/4/14.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_splitting_kernel_h__
#define __mc_splitting_kernel_h__ 1


/*   std includes   */

/*   project headers   */
#include <bits/mc-event.h>
#include <bits/mc-splitfunc.h>
#include <cxx/random.h>



namespace duct {
  
  
  /*   Empty declaration of the splitting_kernel template class    */
  template<class _Type, OrderingTag _Ordering>
  class splitting_kernel;

  
  
  /*
   *    Splitting kernel for the INITIAL STATE evolution.
   */
  template<OrderingTag _Ordering>
  class splitting_kernel<ini, _Ordering>
  {
    /*    private types   */
    using _Split = typename splitting_functions<_Ordering>::initial_state;
    using _Approx = typename splitting_functions<_Ordering>::approx::initial_state;
    
    /*   probability and varaibles of a possible splitting channel   */
    struct _Channel
    {
      /*   probability of the channel   */
      double prob;
      
      /*   flavors after the splitting   */
      int fl, fm;
      
      /*   index of the spectator parton   */
      int spectator;
      
      /*   hardness and momentum fraction variables   */
      double v, z;
      
      /*   pointer to the soft state to be able to get
       *   the azimuthal angle when it is needed
       */
      const typename _Split::soft_state *soft;
      
      /*   generate the azimuthal angle   */
      double azimuthal_angle(double rho) const {
        if(soft) return soft->azimuthal_angle(rho);
        return 2.0*M_PI*rho;
      }
    };
    
  public:
    /*    public types   */
    using ordering_type = _Ordering;
    
    /*    default constructor   */
    splitting_kernel() { _M_zeta.reserve(100);}
    
    
    /*   calculate the hit-or-miss ration and sets up the list of possible splitting channels   */
    double hit_or_miss_ratio(const hard_params<ini> *, double, double);
   
    /*   this function make a trial emission  */
    bool make_trial(double&, splitting_vars&, random_engine_type&, const hard_params<ini> *, double, double);
    
  private:
    /*   the splitting function and it approximation    */
    _Split _M_sfunc;
    _Approx _M_approx;
    
    /*   list of the possible splitting channels   */
    std::vector<_Channel> _M_zeta;
  };
  
  
  template<OrderingTag _Ordering>
  double splitting_kernel<ini, _Ordering>::
  hit_or_miss_ratio(const hard_params<ini> *pars, double tau, double chi)
  {
    /*   some aliases   */
    const int &fl = pars->fl;
 
    /*   cleaning up    */
    _M_zeta.clear();
    
    
    /*   safety cut to avoid strange numerical errors   */
    if(std::abs(fl) > 3 && tau > _M_approx.tauT()*0.95)
      tau = _M_approx.tauT()*0.95;
    
    /*    hardness variable   */
    double v = pars->mdl.alpha_qcd().q2(tau)/pars->Q2;
   
    /*    check if it is in the physical range    */
    if(v < _M_approx.vmin() || v >= _M_approx.vmax())
      return 0.0;
    
    /*   binding to the current hardness variable   */
    _M_approx.bind(v);

    
    /*   soft contribution    */
    double weight = 0.0;
    std::pair<double, double> z = _M_approx.mapping_chi2z(splitting_type::soft, chi);
    
    if(_Split::veto(pars, v, z.first) && z.second > 0.0)
    {
      weight += z.second*_M_sfunc(splitting_type::soft, pars, v, z.first);
      
      for(const auto& s : _M_sfunc.soft_states())
        if(s.weight() > 0.0) _M_zeta.push_back(_Channel{z.second*s.weight(), fl, 0, s.spectator(), v, z.first, &s});
    }
  
    
    /*   collinear contribution    */
    z = _M_approx.mapping_chi2z(splitting_type::collinear, chi);
    
    if(_Split::veto(pars, v, z.first) && z.second > 0.0)
    {
      double w = z.second*_M_sfunc(splitting_type::collinear, pars, v, z.first);
      
      if(w > 0.0) {
        _M_zeta.push_back(_Channel{w, fl, 0, pars->emitter, v, z.first, nullptr});
        weight += w;
      }
    }
    
  
    /*   flavor changing collinear contribution    */
    z = _M_approx.mapping_chi2z(splitting_type::flavor_changing_collinear, chi);
    
    if(_Split::veto(pars, v, z.first) && z.second > 0.0)
    {
      double w = z.second*_M_sfunc(splitting_type::flavor_changing_collinear, pars, v, z.first);
     
      if(w > 0)
      {
        /*   quark turning to gluon    */
        if(fl) {
          _M_zeta.push_back(_Channel{w, 0, fl, pars->emitter, v, z.first, nullptr});
          weight += w;
        }
        
        /*   gluon turns to quark   */
        else {
          for(int k = -5; k <= 5; ++k)
            if(k && _M_sfunc.pdf_ratio(k) > 0.0) {
              _M_zeta.push_back(_Channel{w*_M_sfunc.pdf_ratio(k), k, -k, pars->emitter, v, z.first, nullptr});
              weight += w*_M_sfunc.pdf_ratio(k);
            }
        }
      }
    }

    
    /*   the trial emission was vetoed    */
    if(0 == _M_zeta.size()) return 0.0;

   
    /*   normalizing the probabilities    */
    for(auto& c : _M_zeta) c.prob /= weight;
  
    
    /*   returning the hit-or-miss ratio   */
    return weight/_M_approx.integral();
  }
  
  
  template<OrderingTag _Ordering>
  bool splitting_kernel<ini, _Ordering>::
  make_trial(double& tmT, splitting_vars& vars, random_engine_type& rne, const hard_params<ini> *pars, double tmU, double tmL)
  {
    /*   aliases   */
    const auto& as = pars->mdl.alpha_qcd();
    
    /*    shower time interval in variable tau   */
    double vL = std::max(pars->mT, pars->vnull*std::exp(-tmU));
    double vH = pars->vnull*std::exp(-tmL);
    
    if(vL >= vH) return false;
    

    /*   binding to the hard state   */
    double tauU = as.tauk(vL*pars->Q2), tauL = as.tauk(vH*pars->Q2);

    if(!_M_approx.bind(pars, tauL, tauU)) return false;
    
    
    /*   generating tau variable according to the Sudakov factor    */
    double tauT = tauL;
    std::uniform_real_distribution<double> rng(0.0,1.0);

    do {
      /*   generate tau  */
      auto __tau = _M_approx.generate_tau(tauT, tauU, -std::log(rng(rne)));
    
      if(false == __tau.first) return false;
      
      tauT = __tau.second;
    } while(rng(rne) > hit_or_miss_ratio(pars, tauT, rng(rne)));
  
    
    /*   the splitting is possible from this emitter => select a channel  */
    auto iz = select_randomly(_M_zeta.cbegin(), _M_zeta.cend(), 0.99999*rng(rne),
                              [](const _Channel& zeta) {
                                return zeta.prob;
                              });
    
    /*   setting up the splitting variables    */
    vars.emitter = pars->emitter;
    vars.spectator = iz->spectator;
    
    vars.fl = iz->fl;
    vars.fm = iz->fm;
    
    vars.cl = pars->mdl.color(iz->fl);
    vars.cm = pars->mdl.color(iz->fm);

    vars.z = iz->z;
    vars.y = ordering_traits<ini,_Ordering>::main::mapping_to_y_nojac(pars, iz->v, iz->z);
    vars.phi = iz->azimuthal_angle(rng(rne));
    
    /*   transforming variable v to shower time   */
    tmT = std::log(pars->vnull/iz->v);
    

    return true;
  }
 
  
  
  /*
   *    Splitting kernel for the FINAL STATE evolution.
   */
  template<OrderingTag _Ordering>
  class splitting_kernel<fin, _Ordering>
  {
    static constexpr bool _is_angular_ordered = std::is_same<_Ordering, ordering::angular>::value;
    
    /*    private types   */
    using _Split = typename splitting_functions<_Ordering>::final_state;
    using _Approx = typename splitting_functions<_Ordering>::approx::final_state;
    
    /*   probability and varaibles of a possible splitting channel   */
    struct _Channel
    {
      /*   probability of the channel   */
      double prob;
      
      /*   flavors after the splitting   */
      int fl, fm;
      
      /*   index of the spectator parton   */
      int spectator;
      
      /*   hardness and momentum fraction variables   */
      double v, z;
      
      /*   pointer to the soft state to be able to get
       *   the azimuthal angle when it is needed
       */
      const typename _Split::soft_state *soft;
      
      /*   generate the azimuthal angle   */
      double azimuthal_angle(double rho) const {
        if(soft) return soft->azimuthal_angle(rho);
        return 2.0*M_PI*rho;
      }
    };
    
  public:
    /*    public types   */
    using ordering_type = _Ordering;
    
    /*    default constructor    */
    splitting_kernel() { _M_zeta.reserve(_is_angular_ordered ? 200 : 100);}
    
    /*   calculate the hit-or-miss ration and sets up the list of possible splitting channels   */
    double hit_or_miss_ratio(const hard_params<fin> *, double, double);
    
    /*   this function make a trial emission  */
    bool make_trial(double&, splitting_vars&, random_engine_type&, const hard_params<fin> *, double, double);
 
  private:
    /*   the splitting function and it approximation    */
    _Split _M_sfunc[_is_angular_ordered ? 2 : 1];
    _Approx _M_approx;
    
    /*   list of the possible splitting channels   */
    std::vector<_Channel> _M_zeta;
  };
  
  
  
  template<OrderingTag _Ordering>
  double splitting_kernel<fin, _Ordering>::
  hit_or_miss_ratio(const hard_params<fin> *pars, double tau, double chi)
  {
    /*    cleaning up    */
    _M_zeta.clear();
    
    
    /*    hardness variable   */
    double weight = 0.0, v = pars->mdl.alpha_qcd().q2(tau)/pars->pQ;
    
    /*    check if it is in the physical range    */
    if(v < _M_approx.vmin() || v >= _M_approx.vmax())
      return 0.0;
    
    /*   binding to the current hardness variable   */
    _M_approx.bind(v);
    
    
    /*   soft contribution    */
    auto z = _M_approx.mapping_chi2z(splitting_type::soft, chi);
    
    if(_Split::veto(pars, v, z.first) && z.second > 0.0) {
      weight += z.second*_M_sfunc[0](splitting_type::soft, pars, v, z.first);
    
      for(const auto& s : _M_sfunc[0].soft_states())
        if(s.weight() > 0.0) _M_zeta.push_back(_Channel{z.second*s.weight(), pars->fl, 0, s.spectator(), v, z.first, &s});
    }
    
    
    /*   collinear contribution    */
    z = _M_approx.mapping_chi2z(splitting_type::collinear, chi);
    
    if(_Split::veto(pars, v, z.first) && z.second > 0.0) {
      weight += z.second*_M_sfunc[0](splitting_type::collinear, pars, v, z.first);
      
      const auto& cw = _M_sfunc[0].collinear_weights();
      
      for(int k = 0; k < (int) cw.size(); ++k)
        if(cw[k] > 0.0) _M_zeta.push_back(_Channel{z.second*cw[k], pars->fl+k, -k, pars->emitter, v, z.first, nullptr});
    }
    
    /*   the trial emission was vetoed    */
    if(0 == _M_zeta.size()) return 0.0;
    
    /*   normalizing the probabilities    */
    for(auto& c : _M_zeta) c.prob /= weight;
    
    
    /*   returning the hit-or-miss ratio   */
    return weight/_M_approx.integral();
  }
  
  
  /*    specialization of the hit-or-miss ratio for angular ordering.    */
  template<> double splitting_kernel<fin, ordering::angular>::
  hit_or_miss_ratio(const hard_params<fin> *, double, double);

  
  
  
  template<OrderingTag _Ordering>
  bool splitting_kernel<fin, _Ordering>::
  make_trial(double& tmT, splitting_vars& vars, random_engine_type& rne, const hard_params<fin> *pars, double tmU, double tmL)
  {
    /*   aliases   */
    const auto& as = pars->mdl.alpha_qcd();

    /*    shower time interval in variable tau   */
    double vL = std::max(pars->mT, pars->vnull*std::exp(-tmU));
    double vH = pars->vnull*std::exp(-tmL);
    
    if(vL >= vH) return false;
    
    
    /*   binding to the hard state   */
    double tauU = as.tauk(vL*pars->pQ), tauL = as.tauk(vH*pars->pQ);
    
    if(!_M_approx.bind(pars, tauL, tauU)) return false;
    
    
    /*   generating tau variable according to the Sudakov factor    */
    double tauT = tauL;
    std::uniform_real_distribution<double> rng(0.0,1.0);
    
    do {
      /*   generate tau  */
      auto __tau = _M_approx.generate_tau(tauT, tauU, -std::log(rng(rne)));
      
      if(false == __tau.first) return false;
      tauT = __tau.second;
    } while(rng(rne) > hit_or_miss_ratio(pars, tauT, rng(rne)));
    
    
    /*   the splitting is possible from this emitter => select a channel  */
    auto iz = select_randomly(_M_zeta.cbegin(), _M_zeta.cend(), 0.99999*rng(rne),
                              [](const _Channel& zeta) {
                                return zeta.prob;
                              });
    
    /*   setting up the splitting variables    */
    vars.emitter = pars->emitter;
    vars.spectator = iz->spectator;
    
    vars.fl = iz->fl;
    vars.fm = iz->fm;
    
    vars.cl = pars->mdl.color(iz->fl);
    vars.cm = pars->mdl.color(iz->fm);
    
    vars.z = iz->z;
    vars.y = ordering_traits<fin,_Ordering>::main::mapping_to_y_nojac(pars, iz->v, iz->z);
    vars.phi = iz->azimuthal_angle(rng(rne));

    /*   transforming variable v to shower time   */
    tmT = std::log(pars->vnull/iz->v);
    

    return true;
  }
  
  
}    /*   namespace duct   */

#endif
