/*  Copyright © 2018 Zoltán Nagy. All rights reserved.
 *  Created by Zoltan Nagy on 21.09.18.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_sudakov_h__
#define __mc_sudakov_h__ 1


/*   std includes   */


/*   project headers   */
#include <bits/mc-color-state.h>
#include <bits/mc-soft-function.h>
#include <cxx/buffer.h>
#include <cxx/math-support.h>
#include <cxx/math/linear-algebra.h>


namespace duct {
  
  template<ProcessTag _Proc>
  class sudakov_operator
  {
    /*   some private types   */
    using complex_t = std::complex<double>;
    using subspace_t = buffer<quantum_color_state<_Proc>, 14u>;
    using lambda_t = typename soft_function<_Proc>::result_t;
    
    
    /*   Anomalous matrix & the routine to exponentiate the matrix     */
    template<std::size_t _Dim>
    struct anomalous_t
    {
      using matrix_t = linalg::matrix<complex_t, _Dim, _Dim>;
      
      /*   constructors   */
      anomalous_t() = default;
      
      anomalous_t(std::initializer_list<complex_t> il)
      : exponent{il} {}
      
      /*   do the exponentiation   */
      void sudakov(subspace_t& subspace, const lambda_t& lmd);
      
      /*   anomalous matrix   */
      matrix_t exponent;
    };
    
    
  public:
    /*   make the "sudakov transformation"   */
    template<class _RandomEngine>
    auto operator()(color_state<_Proc>& c, _RandomEngine& rne, const lambda_t& lmd)
    {
      /***   exponentiating the quantum operators    ***/
      _M_quantum_states_of(c);
      _S_span_the_subspace(_M_left, lmd);
      _S_span_the_subspace(_M_right, lmd);
      
      /***   selecting a color basis state (statistical state) randomly   ***/
      _TMP_w.clear();

      double norm = 0.0, wprob;
      complex_t weight;
      
      for(std::size_t i = 0; i < _M_left.size(); i++)
        for(std::size_t j = 0; j < _M_right.size(); j++)
        {
          weight = _M_left[i].weight*std::conj(_M_right[j].weight);
          wprob = std::abs(weight);
          norm += wprob;
          
          _TMP_w.push_back({i, j, weight, wprob});
        }
      
      auto ic = select_randomly(rne, _TMP_w.begin(),  _TMP_w.end(),
                                [=](const state_ref& r) {
                                  return r.wprob/norm;
                                });
      
      /*   reconstructing the statistical color state    */
      c.weight *= ic->weight;
      c.weight *= norm/ic->wprob;
      
      /*     copying the color vectors   */
      const auto &cL = _M_left[ic->i], &cR = _M_right[ic->j];
      
      for(int l = c.lower(); l <= c.upper(); ++l) {
        c[l].colorL = cL[l];
        c[l].colorR = cR[l];
      }
    }
    
  private:
    /*   buffers of the basis states   */
    subspace_t _M_left, _M_right;

    
    /*    helper functions    */
    void _M_quantum_states_of(const color_state<_Proc>&);
    static void _S_span_the_subspace(subspace_t&, const lambda_t&);

  private:
    /*   temporary members to select states randomly   */
    struct state_ref
    {
      std::size_t i, j;
      complex_t weight;
      double wprob;
    };
    
    std::vector<state_ref> _TMP_w;
  };
  
  
  template<ProcessTag _Proc>
  void sudakov_operator<_Proc>::_M_quantum_states_of(const color_state<_Proc>& ci)
  {
    int lo = ci.lower(), hi = ci.upper();

    /*   making space   */
    _M_left.resize(1);
    _M_left[0].resize(hi);
    
    _M_right.resize(1);
    _M_right[0].resize(hi);
    
    _M_left[0].weight = _M_right[0].weight = 1.0;
    
    
    /*   copying the color vectors   */
    for(int l = lo; l <= hi; ++l)
    {
      _M_left[0][l] = ci[l].colorL;
      _M_right[0][l] = ci[l].colorR;
    }
  }
  

  
  template<ProcessTag _Proc>
  void sudakov_operator<_Proc>::
  _S_span_the_subspace(subspace_t& sub, const lambda_t& lmd)
  {
    /*   trivial case when the phase term is zero    */
    if(std::abs(lmd.omega) < 1e-16) {
      anomalous_t<1u> n;
      n.sudakov(sub, lmd);
      return;
    }
    
    
    /*   the case when we have to work hard     */
    const quantum_color_state<_Proc> &c = sub[0];
    int a = -1, b = 0;
    
    const complex_t om{0.0,lmd.omega}, nc{0.0, qcd::Nc*lmd.omega};
    const complex_t tnc{0.0, 2.0*qcd::Nc*lmd.omega}, onc{0.0, lmd.omega/qcd::Nc};
    
    
    /*   gluon gluon incomings   */
    if(c.is_octet(a) && c.is_octet(b))
    {
      /*  special case: only 'a' and 'b' are on a closed string    */
      if(c[a].prev == b && c[a].next == b)
      {
        anomalous_t<1u> n;
        
        n.exponent[0][0] = tnc;
        n.sudakov(sub, lmd);
        
        return;
      }
      
      /*  special case : 'b' is next to 'a'    */
      if(c[a].next == b)
      {
        sub.push_back(c);
        
        /*   setting up the new state   */
        quantum_color_state<_Proc> &d = sub[1];
        
        d[d[a].prev].next = d[b].next;
        d[d[b].next].prev = d[a].prev;
        d[a].prev = b;
        d[b].next = a;
        
        /*   imaginary part of the anomalous dimension   */
        anomalous_t<2u> n = {
          nc, 0.0,
          om, tnc
        };
        
        /*   do the exponentiation   */
        n.sudakov(sub, lmd);

        return;
      }
      
      /*  special case : 'a' is next to 'b'    */
      if(c[b].next == a)
      {
        sub.push_back(c);
        
        /*   setting up the new state   */
        quantum_color_state<_Proc> &d = sub[1];
        
        d[d[b].prev].next = d[a].next;
        d[d[a].next].prev = d[b].prev;
        d[b].prev = a;
        d[a].next = b;
        
        /*   imaginary part of the anomalous dimension   */
        anomalous_t<2u> n = {
          nc, 0.0,
          om, tnc
        };
        
        /*   do the exponentiation   */
        n.sudakov(sub, lmd);
        
        return;
      }
      
      /*    general case is more complicated we have 14 basis vectors   */
      {
        /*  C1 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[1];
          
          d[d[a].prev].next = b;
          d[d[a].next].prev = b;
          
          d[d[b].prev].next = a;
          d[d[b].next].prev = a;
          
          std::swap(d[a].prev, d[b].prev);
          std::swap(d[a].next, d[b].next);
        }
        
        /*  C2 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[2];
          
          d[d[a].next].prev = b;
          d[d[b].next].prev = a;
          
          std::swap(d[a].next, d[b].next);
        }
        
        /*  C3 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[3];
          
          d[d[a].prev].next = b;
          d[d[b].prev].next = a;
          
          std::swap(d[a].prev, d[b].prev);
        }
        
        /*  C4 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[4];
          
          d[d[a].next].prev = d[b].prev;
          d[d[b].prev].next = d[a].next;
          
          d[a].next = b;
          d[b].prev = a;
        }
        
        /*  C5 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[5];
          
          d[d[b].next].prev = d[a].prev;
          d[d[a].prev].next = d[b].next;
          
          d[b].next = a;
          d[a].prev = b;
        }
        
        /*  C6 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[6];
          
          d[d[a].prev].next = b;
          d[d[a].next].prev = d[b].prev;
          
          d[d[b].prev].next = d[a].next;
          d[d[b].next].prev = a;
          
          d[b].prev = d[a].prev;
          d[a].next = d[b].next;
          
          d[b].next = a;
          d[a].prev = b;
        }
        
        /*  C7 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[7];
          
          d[d[a].prev].next = d[b].next;
          d[d[a].next].prev = b;
          
          d[d[b].prev].next = a;
          d[d[b].next].prev = d[a].prev;
          
          d[a].prev = d[b].prev;
          d[b].next = d[a].next;
          
          d[a].next = b;
          d[b].prev = a;
        }
        
        /*  C8 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[8];
          
          d[d[b].prev].next = d[b].next;
          d[d[b].next].prev = d[b].prev;
          
          d[d[a].next].prev = b;
          d[b].next = d[a].next;
          d[b].prev = a;
          d[a].next = b;
        }
        
        /*  C9 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[9];
          
          d[d[a].prev].next = d[a].next;
          d[d[a].next].prev = d[a].prev;
          
          d[d[b].next].prev = a;
          d[a].next = d[b].next;
          d[b].next = a;
          d[a].prev = b;
          
        }
        
        /*  C10 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[10];
          
          d[d[b].prev].next = d[b].next;
          d[d[b].next].prev = d[b].prev;
          
          d[d[a].prev].next = b;
          d[b].prev = d[a].prev;
          d[a].prev = b;
          d[b].next = a;
        }
        
        /*  C11 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[11];
          
          d[d[a].prev].next = d[a].next;
          d[d[a].next].prev = d[a].prev;
          
          d[d[b].prev].next = a;
          d[a].prev = d[b].prev;
          d[b].prev = a;
          d[a].next = b;
        }
        
        /*  C12 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[12];
          
          d[d[a].prev].next = d[b].next;
          d[d[a].next].prev = d[b].prev;
          
          d[d[b].prev].next = d[a].next;
          d[d[b].next].prev = d[a].prev;
          
          d[a].next = d[a].prev = b;
          d[b].next = d[b].prev = a;
        }
        
        /*  C13 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[13];
          
          d[d[a].prev].next = d[a].next;
          d[d[a].next].prev = d[a].prev;
          
          d[d[b].prev].next = d[b].next;
          d[d[b].next].prev = d[b].prev;
          
          d[a].next = d[a].prev = b;
          d[b].next = d[b].prev = a;
        }
      }
      
      
      /*   imaginary part of the anomalous dimension   */
      anomalous_t<14u> n = {
        0.0, 0.0, -om, -om, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, -om, -om, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        -om, -om, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        -om, -om, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
         om, 0.0, 0.0, 0.0,  nc, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
         om, 0.0, 0.0, 0.0, 0.0,  nc, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0,  om, 0.0, 0.0, 0.0, 0.0,  nc, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0,  om, 0.0, 0.0, 0.0, 0.0, 0.0,  nc, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0, 0.0,  om, 0.0, 0.0, 0.0, 0.0, 0.0,  nc, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0, 0.0,  om, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,  nc, 0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0,  om, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,  nc, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0,  om, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,  nc, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0,  om,  om,  om,  om, 0.0, 0.0, 0.0, 0.0, tnc, 0.0,
        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,  om,  om,  om,  om, 0.0, tnc
      };
      
      
      /*   do the exponentiation   */
      n.sudakov(sub, lmd);

      
      return;
    }  //  end of gluon-gluon case
    
    
    /*   quark-gluon case   */
    if((c.is_triplet(a) && c.is_octet(b)) || (c.is_triplet(b) && c.is_octet(a)))
    {
      if(c.is_triplet(b)) std::swap(a,b);

      /*   special case when 'b' is emitted by 'a'   */
      if(c[a].next == b)
      {
        anomalous_t<1u> n;
        
        n.exponent[0][0] = nc;
        n.sudakov(sub, lmd);
        
        return;
      }
      
      /*    in the general case we have 4 basis vectors    */
      {
        /*  C1 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[1];
          
          d[d[a].next].prev = b;
          d[d[b].next].prev = a;
          
          std::swap(d[b].next, d[a].next);
        }
        
        /*  C2 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[2];
          
          d[d[a].next].prev = b;
          d[d[b].next].prev = d[b].prev;
          d[d[b].prev].next = d[b].next;
          
          d[b].next = d[a].next;
          d[b].prev = a;
          d[a].next = b;
        }
        
        /*  C3 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[3];
          
          d[d[a].next].prev = d[b].prev;
          d[d[b].prev].next = d[a].next;
          
          d[b].prev = a;
          d[a].next = b;
        }
      }
      
      
      /*   imaginary part of the anomalous dimension   */
      anomalous_t<4u> n = {
        0.0, -om, 0.0, 0.0,
        -om, 0.0, 0.0, 0.0,
        0.0,  om,  nc, 0.0,
         om, 0.0, 0.0,  nc
      };
      
      /*   do the exponentiation   */
      n.sudakov(sub, lmd);

      
      return;
    }   //  end of quark-gluon case
    
    
    /*   antiquark-gluon case   */
    if((c.is_antitriplet(a) && c.is_octet(b)) || (c.is_antitriplet(b) && c.is_octet(a)))
    {
      if(c.is_antitriplet(b)) std::swap(a,b);
      
      /*   special case when 'b' is emitted by 'a'   */
      if(c[a].prev == b)
      {
        anomalous_t<1u> n;
        
        n.exponent[0][0] = nc;
        n.sudakov(sub, lmd);

        return;
      }
      
      /*    in the general case we have 4 basis vectors    */
      {
        /*  C1 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[1];
          
          d[d[a].prev].next = b;
          d[d[b].prev].next = a;
          
          std::swap(d[b].prev, d[a].prev);
        }
        
        /*  C2 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[2];
          
          d[d[a].prev].next = b;
          d[d[b].next].prev = d[b].prev;
          d[d[b].prev].next = d[b].next;
          
          d[b].prev = d[a].prev;
          d[b].next = a;
          d[a].prev = b;
        }
        
        /*  C3 :  */
        {
          sub.push_back(c);
          
          /*   setting up the new state   */
          quantum_color_state<_Proc> &d = sub[3];
          
          d[d[a].prev].next = d[b].next;
          d[d[b].next].prev = d[a].prev;
          
          d[b].next = a;
          d[a].prev = b;
        }
      }
      
  
      /*   imaginary part of the anomalous dimension   */
      anomalous_t<4u> n = {
        0.0, -om, 0.0, 0.0,
        -om, 0.0, 0.0, 0.0,
        0.0, om,   nc, 0.0,
         om, 0.0, 0.0,  nc
      };
      
      /*   do the exponentiation   */
      n.sudakov(sub, lmd);
 
      
     return;
    }   //  end of antiquark-gluon case

    
    /*    quark-quark case    */
    if(c.is_triplet(a) && c.is_triplet(b))
    {
      /*  C1 :  */
      {
        sub.push_back(c);
        
        /*   setting up the new state   */
        quantum_color_state<_Proc> &d = sub[1];
        
        d[d[a].next].prev = b;
        d[d[b].next].prev = a;
        
        std::swap(d[a].next, d[b].next);
      }
      
      /*   imaginary part of the anomalous dimension   */
      anomalous_t<2u> n = {
        onc, -om,
        -om, onc
      };
      
      /*   do the exponentiation   */
      n.sudakov(sub, lmd);
      
      return;
    }   //  end of quark-quark case
    
    
    /*    antiquark-antiquark case    */
    if(c.is_antitriplet(a) && c.is_antitriplet(b))
    {
      /*  C1 :  */
      {
        sub.push_back(c);
        
        /*   setting up the new state   */
        quantum_color_state<_Proc> &d = sub[1];
        
        d[d[a].prev].next = b;
        d[d[b].prev].next = a;
        
        std::swap(d[a].prev, d[b].prev);
      }
      
      /*   imaginary part of the anomalous dimension   */
      anomalous_t<2u> n = {
        onc, -om,
        -om, onc
      };
      
      /*   do the exponentiation   */
      n.sudakov(sub, lmd);
      
      return;
    }   //  end of antiquark-antiquark case
    
    
    /*   quark-antiquark case   */
    if((c.is_triplet(a) && c.is_antitriplet(b)) || (c.is_triplet(b) && c.is_antitriplet(a)))
    {
      if(c.is_triplet(b)) std::swap(a,b);
      
      /*   special case when 'b' is next 'a'   */
      if(c[a].next == b)
      {
        anomalous_t<1u> n;
        
        n.exponent[0][0] = nc-onc;
        n.sudakov(sub, lmd);
        
        return;
      }
      
      /*  C1 :  */
      {
        sub.push_back(c);
        
        /*   setting up the new state   */
        quantum_color_state<_Proc> &d = sub[1];
        
        d[d[a].next].prev = d[b].prev;
        d[d[b].prev].next = d[a].next;
        
        d[a].next = b;
        d[b].prev = a;
      }
      
      /*   imaginary part of the anomalous dimension   */
      anomalous_t<2u> n = {
        -onc, 0.0,
        om, nc-onc
      };
      
      /*   do the exponentiation   */
      n.sudakov(sub, lmd);
      
      return;
    }   //  end of quark-antiquark case
  }
  
  
  
  template<ProcessTag _Proc> template<std::size_t _Dim>
  void sudakov_operator<_Proc>::anomalous_t<_Dim>::sudakov(subspace_t& subspace, const lambda_t& lmd)
  {
    /*   real part of the anomalous dimension for the original state    */
    exponent[0][0] += soft_exponent(lmd, subspace[0]);
   
    
    /*   trivial case : 1 dimensional subspace    */
    if constexpr(1u == _Dim)
    {
      subspace[0].weight = std::exp(-exponent[0][0]);
      return;
    }
    
    
    /*   ignore the tadpole graphs      */
    for(std::size_t k = 1; k < subspace.size(); ++k)
      if(check_for_tadpole(subspace[k]))
      {
        for(std::size_t l = 0; l < _Dim; ++l)
          exponent[k][l] = exponent[l][k] = 0.0;
      }
      else {
        exponent[k][k] += soft_exponent(lmd, subspace[k]);
      }
    
    
    /*    diagonalize the anomalous matrix    */
    linalg::matrix<complex_t, _Dim,_Dim> u, d;
    
    if(auto nsweep = linalg::eigensystem(exponent, d, u); nsweep > 0)
    {
      /*    exponential of the eigenvalues   */
      linalg::ketvector<complex_t, _Dim> n;
      
      for(std::size_t i = 0; i < _Dim; ++i)
        n[i] = std::exp(-d[i][i])*u[i][0];
      
      auto nn = *inverse(u)*n;

      for(std::size_t l = 0; l < _Dim; ++l)
        subspace[l].weight = nn[l];
    }

    /*    the matrix is not diagonalizable we try to do something else    */
    else
    {
      /*   ignore the imaginary part    */
      subspace[0].weight = std::exp(-std::real(exponent[0][0]));
      
      for(std::size_t l = 1; l < _Dim; ++l)
        subspace[l].weight = 0.0;
      
      std::cerr<<"Sudakov operator : Cannot diagonalize the anomalous matrix.\n This message shouldn't appear often.\n";
      std::cerr<<exponent<<"\n"<<"========================================================="<<std::endl;
    }
  }
  
}   //   namespace duct

#endif /* __mc_sudakov_h__ */
