/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 3/18/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __nlo_model_h__
#define __nlo_model_h__ 1

/*   std headers   */
#include <map>
#include <random>
#include <string>
#include <climits>
#include <cmath>
#include <cstdlib>
#include <fstream>
#include <functional>
#include <initializer_list>

/*   project headers    */
#include <bits/duct-tags.h>
#include <bits/mc-coupling.h>
#include <bits/mc-shower-aux.h>
#include <cxx/filesystem.h>



namespace duct {
  
  
  /*   Structure to describe a particle */
  struct particle_info
  {
    /*    default constructor   */
    particle_info()
    : key(INT_MAX), name("---"), mass(0.0), mass2(0.0), color(0), spin(0) {}

    /*    constructor    */
    particle_info(int k, const char *n, energy<double> m, int c, int s)
    : key(k), name(n), mass(m), mass2(m*m), color(c), spin(s) {}
    
    //  id of the particle
    int key;
    
    //   name of the particle
    std::string name;
    
    //   mass of the particle
    energy<double> mass;
    energy_sqr<double> mass2;
    
    //   color representation (-3, 0, 3, 8)
    int color;
    
    //   2*spin (0, 1, 2)
    int spin;
  };
  
  
  /*   Structure to describe a hadron. Note that the hadron can play the
   *   role of an incomig hadron, thus we need the pdf information.
   */
  struct hadron_info {
    /*    default constructors    */
    hadron_info() : key{INT_MAX}, name{"---"}{}
    
    /*    constructor    */
    hadron_info(int k, const char *n)
    : key{k}, name{n} {}
    
    /*   id of the hadron   */
    int key;
    
    /*   name of the hadron  */
    std::string name;
    
    /*   pdf and pdf related auxilary data for shower
     *    This is not the standard MSbar pdf!
     */
    std::array<shower_aux, 3> aux;
    
    /*   MSbar PDFs   */
    std::array<qcd::dglap, 2> pdf;
  };
  
 
  /*    Some pre defined QCD settings distributed with the library.    */
  namespace qcd {
    
    struct predefined : input {
    };
    
    
    /*    HERA PDF
     *      It is from Voica. (This is for testing only.)
     */
    extern const predefined herapdflo;
    
    
    /*   This is the PDF function of CT14n at the fitting scale
     *   Q = 1.295GeV. The alpha_s(Mz) = 0.118.
     */
    extern const predefined ct14n;
    
    /*    List predefined QCD inputs    */
    extern const predefined * default_inputs[];
  }
  
  
  
  /*   Base class to define particles and hadron and basic things
   *   those are important to have parton shower implemented.
   *   All the model extension should be inhereted from this.
   */
  class model
  {
  public:
    /*  types  */
    typedef std::map<int, particle_info> ptbl_type;
    typedef std::map<int, hadron_info>   htbl_type;
    
    /*    constructors    */
    model();
    explicit model(const qcd::input&);
    
    template<OrderingTag _Ordering>
    model(_Ordering, const qcd::input&);
    
    
    /*    creating from qcd::input and saving it   */
    model(fs::path, const qcd::input&);
    
    template<OrderingTag _Ordering>
    model(fs::path, _Ordering, const qcd::input&);
    
    /*    creating from predefined qcd::input and saving it   */
    explicit model(const qcd::predefined& __in)
    : model{fs::path{std::getenv("HOME")}/".deductor/models", __in} {}
    
    template<OrderingTag _Ordering>
    model(_Ordering o, const qcd::predefined& __in)
    : model{fs::path{std::getenv("HOME")}/".deductor/models", o, __in} {}
    
    
    
    /*  construct from list of particles and hadrons  */
    model(const particle_info *, const hadron_info *);

    /*   adding more particles from initializer list   */
    model(std::initializer_list<particle_info> pl, std::initializer_list<hadron_info> hl) {
      for(const particle_info& p : pl) particle_table[p.key] = p;
      for(const hadron_info& h : hl) hadron_table[h.key] = h;
    }
    
     
    /*    return the mass and mass2 of a particle   */
    energy<double> mass(int key) const { return particle_table.at(key).mass;}
    energy_sqr<double> mass2(int key) const { return particle_table.at(key).mass2;}
 
    /*    return the color and spin of a particle   */
    int color(int key) const { return particle_table.at(key).color;}
    int spin(int key) const { return particle_table.at(key).spin;}
   
    /*    return the name of a particle or hadron   */
    const std::string& name(int key) const { return particle_table.at(key).name;}
    const std::string& name(const hadron& ih) const { return hadron_table.at(ih.idx).name;}
    
    /*   add new elements to the tables or modify them   */
    particle_info& operator[](int k) { return particle_table[k];}
    hadron_info& operator[](const hadron& ih) { return hadron_table[ih.idx];}
    
    
    /*    returns a pointer to the auxiliary data   */
    template<OrderingTag _Ordering>
    const shower_aux * aux(const hadron& ih) const {
      return std::addressof(hadron_table.at(ih.idx).aux[static_cast<unsigned>(_Ordering::alpha)]);
    }
    
    /*    returns a pointer to the auxiliary data   */
    template<qcd::order _Ord>
    const qcd::dglap * pdf(const hadron& ih) const {
      return std::addressof(hadron_table.at(ih.idx).pdf[static_cast<unsigned>(_Ord)]);
    }
    
    
    /*   This is the particle table. We use a map for this and particles can be added easily.
     */
    std::map<int, particle_info> particle_table;
    
    /*   Similarly we need a hadron table.
     *   At the moment we have to describe only the incomming hadrons.
     */
    std::map<int, hadron_info> hadron_table;
    
    /*   Strong coupling   */
    qcd_coupling alpha_qcd;

    /*   Simple electromagnetic coupling 
     *   It is virtual, thus, it can be overridden. 
     */
    virtual double alpha_qed(energy_sqr<double>) const;

  private:
    /*   we need some private function...   */
    void read_core(std::istream&);
    void write_core(std::ostream&) const;
 
    void read_pdf(int, std::istream&);
    template<OrderingTag> void read_aux(int, std::istream&);

    /*   QCD coupling   */
    void qcd_core(const qcd::input& __in) {
      qcd::coupling as{__in.order, __in.asref/(2.0*M_PI), sqr(__in.qref), sqr(__in.mc), sqr(__in.mb), sqr(__in.mt)};
      alpha_qcd = qcd_coupling{0.25_GeV2, 1e+15_GeV2, as};
    }
    
    /*   MSbar evolution   */
    void qcd_msbar_pdf(const qcd::input&);
    void qcd_msbar_pdf(int, const qcd::input&);

    /*    Shower pdf evolution and integrals    */
    template<OrderingTag> void qcd_shower_aux(const qcd::input&);
    template<OrderingTag> void qcd_shower_aux(int, const qcd::input&);
  };
  
  
  
  template<OrderingTag _Ordering>
  inline model::model(_Ordering, const qcd::input& __in) : model{}
  {
    qcd_core(__in);
    qcd_msbar_pdf(__in);
    qcd_shower_aux<_Ordering>(__in);
  }
  
  
  inline model::model(const qcd::input& __in) : model{}
  {
    qcd_core(__in);
    qcd_msbar_pdf(__in);
    qcd_shower_aux<ordering::kT>(__in);
    qcd_shower_aux<ordering::lambda>(__in);
    qcd_shower_aux<ordering::angular>(__in);
  }
  
  
  template<OrderingTag _Ordering>
  model::model(fs::path dir, _Ordering, const qcd::input& __in) : model{}
  {
    dir /= __in.name;
    
    /*   create the directory if it doesn't exists    */
    if(!fs::is_directory(dir))
      fs::create_directories(dir);
    
    
    /*   CORE part     */
    fs::path fn{dir/"core.mdl"};
    
    /*   if we have core.mdl file then we read it   */
    if(fs::exists(fn)) {
      std::fstream file{fn.c_str(), std::ios::in};
      read_core(file);
    }
    
    /*   otherwise save the default     */
    else {
      std::fstream file{fn.c_str(), std::ios::out};
      qcd_core(__in);
      write_core(file);
    }
    
    
    /*   PDFs
     *   we read or create if there is any in the qcd::input
     */
    if(__in.pdf.empty() == false)
    {
      std::string oname{_Ordering::name}; oname += '-';
      
      for(auto& h : __in.pdf)
      {
        /*   MSbar PDFs     */
        int key = h.first;
        std::string pdg{"msbar-"};
        
        pdg += std::to_string(key);
        pdg += ".pdg";
        
        fn = dir/pdg;
        
        if(fs::exists(fn)) {
          std::fstream file{fn.c_str(), std::ios::in};
          read_pdf(key, file);
        }
        
        /*   otherwise save the default     */
        else {
          std::fstream file{fn.c_str(), std::ios::out};
          qcd_msbar_pdf(key, __in);
          duct::write(file, hadron_table.at(key).pdf);
        }
        
        
        /*   Shower PDFs    */
        pdg = oname;
        pdg += std::to_string(key);
        pdg += ".pdg";
        
        fn = dir/pdg;
        
        
        if(fs::exists(fn)) {
          std::fstream file{fn.c_str(), std::ios::in};
          read_aux<_Ordering>(key, file);
        }
        
        /*   otherwise save the default     */
        else {
          std::fstream file{fn.c_str(), std::ios::out};
          qcd_shower_aux<_Ordering>(key, __in);
          hadron_table.at(key).aux[static_cast<unsigned>(_Ordering::alpha)].write(file);
        }
      }
    }
  }



  template<OrderingTag _Ordering>
  void model::qcd_shower_aux(int key, const qcd::input& qin)
  {
    unsigned k = static_cast<unsigned>(_Ordering::alpha);
    
    if(qin.pdf.empty() == false)
    {
      for(auto& h : qin.pdf)
        if(key == h.first)
        {
          /*   calculate the shower PDFs   */
          dglap __dglap{sqr(qin.mT), sqr(qin.mc), sqr(qin.mb), alpha_qcd};
          
          __dglap.evolve(_Ordering::alpha, h.second);
          
          hadron_table[-key].aux[k] = hadron_table[key].aux[k] = shower_aux{__dglap};
          hadron_table[-key].aux[k].anti();
          
          /*   just for safety if we provide a hadron that is not in the list   */
          hadron_table[key].key = key;
          hadron_table[-key].key = -key;
        }
    }
  }
  
  
  template<OrderingTag _Ordering>
  void model::read_aux(int key, std::istream& is)
  {
    shower_aux aux;
    unsigned k = static_cast<unsigned>(_Ordering::alpha);
    
    aux.read(is, alpha_qcd);
    hadron_table[-key].aux[k] = hadron_table[key].aux[k] = aux;
    
    hadron_table[-key].aux[k].anti();
    
    /*   just for safety if we provide a hadron that is not in the list   */
    hadron_table[key].key = key;
    hadron_table[-key].key = -key;
  }
  
  
  template<OrderingTag _Ordering>
  void model::qcd_shower_aux(const qcd::input& qin)
  {
    unsigned k = static_cast<unsigned>(_Ordering::alpha);
    
    if(qin.pdf.empty() == false)
    {
      for(auto& h : qin.pdf)
      {
        int key = h.first;
        
        /*   calculate the shower PDFs   */
        dglap __dglap{sqr(qin.mT), sqr(qin.mc), sqr(qin.mb), alpha_qcd};
        
        __dglap.evolve(_Ordering::alpha, h.second);
        
        hadron_table[-key].aux[k] = hadron_table[key].aux[k] = shower_aux{__dglap};
        hadron_table[-key].aux[k].anti();
        
        /*   just for safety if we provide a hadron that is not in the list   */
        hadron_table[key].key = key;
        hadron_table[-key].key = -key;
      }
    }
  }
  


  /*    Reference to a model.
   *    We use this class to pass the model to the hard part.
   *    The model is binded to a given ordering scheme.
   */
  class model_ref
  {
  public:
    /*   if it is not specified it will use kT ordering when we ask for pdf   */
    explicit model_ref(const model *mdl = nullptr) : model_ref{ordering::lambda{}, mdl} {}
    explicit model_ref(const model& mdl) : model_ref{ordering::lambda{}, &mdl} {}
    
    /*   tagged the model with ordering   */
    template<class _Ordering>
    model_ref(_Ordering, const model *mdl)
    : _M_mdl{mdl}, _M_alpha{_Ordering::alpha} {}
    
    template<class _Ordering>
    model_ref(_Ordering, const model& mdl)
    : _M_mdl{&mdl}, _M_alpha{_Ordering::alpha} {}
    

    /*    return the mass and mass2 of a particle   */
    energy<double> mass(int key) const { return _M_mdl->mass(key);}
    energy_sqr<double> mass2(int key) const { return _M_mdl->mass2(key);}
    
    
    /*    return the color and spin of a particle   */
    int color(int key) const { return _M_mdl->color(key);}
    int spin(int key) const { return _M_mdl->spin(key);}
    
    
    /*    return the name of a particle or hadron   */
    const std::string& name(int key) const { return _M_mdl->name(key);}
    const std::string& name(const hadron& ih) const { return _M_mdl->name(ih);}
    
    /*    return the strong coupling    */
    const qcd_coupling& alpha_qcd() const { return _M_mdl->alpha_qcd;}
    
    /*    returns a pointer to the pdfs   */
    const qcd::dglap * pdf(qcd::order ord, int key) const {
      return std::addressof(_M_mdl->hadron_table.at(key).pdf[static_cast<unsigned>(ord)]);
    }

    const std::array<qcd::dglap,2>&  pdf(int key) const {
      return _M_mdl->hadron_table.at(key).pdf;
    }
    
   /*    returns a pointer to the pdf helpers  */
    const shower_aux * aux(const hadron& ih) const {
      return std::addressof(_M_mdl->hadron_table.at(ih.idx).aux[static_cast<unsigned>(_M_alpha)]);
    }

    
    
    /*    threshold weights for an incoming parton   */
    double hard_thrs_weight(int key, int f, double eta, energy_sqr<double> q2) const {
      const auto& h = _M_mdl->hadron_table.at(key);
      return h.aux[static_cast<unsigned>(_M_alpha)].parton(f, eta, q2)/h.pdf[0].parton(f, eta, q2);
    }
    
    
    /*   converting to reference to model  */
    explicit operator const model& () const { return *_M_mdl;}
    const model& get() const { return *_M_mdl;}
    
    
    /*    couplings   */
    double alpha_qcd(energy_sqr<double> q2) const { return _M_mdl->alpha_qcd(q2);}
    double alpha_qed(energy_sqr<double> q2) const { return _M_mdl->alpha_qed(q2);}
    
    /*    returns the ordering scheme    */
    ordering::scheme scheme() const { return _M_alpha;}
    
  private:
    const model * _M_mdl;
    ordering::scheme _M_alpha;
  };
  
  
  /*   returns a pointer to the default model    */
  const model * get_default_model();
  
  /*   set the default model, if the argument is nullptr,
   *   it resets the model to the sytem default
   */
  void set_default_model(const model * = nullptr);
  
}

#endif
