/*  Copyright © 2021 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 12.01.21 at 23:39.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_color_basis_h__
#define __mc_color_basis_h__ 1


/*   std includes   */


/*   project headers   */
#include <bits/mc-statistical-basis.h>
#include <bits/qcd-color.h>



/*     Here we define component that are needed for
 *     put together a complete color state for shower evolution.
 */
namespace duct {
  
  /*    color state of a single parton   */
  struct parton_cstate
  {
    /*   color connections   */
    struct colorlink
    {
      /*   make a dead link, it means the particle is not QCD parton   */
      static colorlink dead() {
        return {INT_MAX, INT_MIN};
      }
      
      /*    make a quark    */
      static colorlink quark(int n) {
        return {quarkend, n};
      }
      
      /*   make an antiquark    */
      static colorlink antiquark(int p) {
        return {p, antiquarkend};
      }
      
      //----- test if it is a quark -----
      bool is_quark() const { return prev == quarkend;}
      bool is_triplet() const { return prev == quarkend;}
      
      //----- test if it is a quark -----
      bool is_antiquark() const { return next == antiquarkend;}
      bool is_antitriplet() const { return next == antiquarkend;}
      
      //----- test if it is a quark -----
      bool is_gluon() const { return next > antiquarkend && prev < quarkend;}
      bool is_octet() const { return next > antiquarkend && prev < quarkend;}
      
      //----- test if it is connection ----
      bool is_qcd() const {
        return next > INT_MIN && prev < INT_MAX && !(prev == quarkend && next == antiquarkend);
      }
      
      /*   test if it is connected to parton i  */
      bool is_connected(int i) const {
        return prev == i || next == i;
      }
      
      //---- labeling the quark and antiquark ends -----
      static constexpr int quarkend =  INT_MAX-1;
      static constexpr int antiquarkend = INT_MIN+1;
      
      /*  The label of the previous and next parton on the color string.
       *  this should follow our labeling convetion (tagged integers).
       */
      int prev, next;
    } colorL, colorR;
    
    
    /*   test if it is triplet, antitriplet or octet   */
    bool is_octet() const { return colorL.is_octet();}
    bool is_triplet() const { return colorL.is_triplet();}
    bool is_antitriplet() const { return colorL.is_antitriplet();}

    bool is_gluon() const { return colorL.is_gluon();}
    bool is_quark() const { return colorL.is_quark();}
    bool is_antiquark() const { return colorL.is_antiquark();}

    int color() const
    {
      if(is_octet()) return 8;
      if(is_triplet()) return 3;
      if(is_antitriplet()) return -3;
      
      return 0;
    }

    
    /*  Healthiness of a gluon:
     *    => '-' not applicable
     *    => 'f' frail gluon (or color 8 object)
     *    => 'h' healthy gluon (or color 8 object)
     */
    char health;
    
    /*   helps to calculate the color suppression index   */
    mutable bool active;
  };

  /*    fancy print    */
  inline std::ostream& operator<<(std::ostream& os, const parton_cstate::colorlink& c)
  {
    if(c.is_quark()) os<<"[Q, "<<c.next<<")";
    else if(c.is_antiquark()) os<<"("<<c.prev<<", A]";
    else if(c.is_gluon()) os<<"("<<c.prev<<","<<c.next<<")";
    else os<<"(-,-)";
    
    return os;
  }

  inline std::ostream& operator<<(std::ostream& os, const parton_cstate& c) {
    return os<<"|"<<c.colorL<<"><"<<c.colorR<<"|  "<<c.health;
  }
  
  
   
  /*    check the LC+ the color connections   */
  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate>
  && requires(_Ext e) { {e.is_color_u3} -> std::convertible_to<bool>;}
  double chi_cfac(int k, int l, const statistical_basis<_Proc, _Parton, _Ext>& p)
  {
    double cfac = (!p.is_color_u3 && !p[l].is_octet() ? 0.5*qcd::Cf : 0.25*qcd::Ca);
    
    if(l == k) return (p[l].is_octet() ? 4.0 : 2.0)*cfac;
    
    int ret = 0;
    auto& cL = p[l].colorL;
    auto& cR = p[l].colorR;
    
    if(cL.prev == k) ++ret;
    if(cL.next == k) ++ret;
    
    if(cR.prev == k) ++ret;
    if(cR.next == k) ++ret;
    
    
    return ret*cfac;
  }
  
  
  /*        Elementary color operators       */
  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate>
  void color_a_dagger(int sigma, int emitter, statistical_basis<_Proc, _Parton, _Ext>& p)
  {
    static constexpr int QUARK = parton_cstate::colorlink::quarkend;
    static constexpr int ANTI_QUARK = parton_cstate::colorlink::antiquarkend;
    
    int spectator;
    auto [_,m] = p.bounds();
    
    switch(sigma) {
      case 1:
        spectator = p[emitter].colorL.next;
        p[emitter].colorL.next  = m;
        p[m].colorL.prev = emitter;
        p[m].colorL.next = spectator;
        p[spectator].colorL.prev = m;
        break;
      case -1:
        spectator = p[emitter].colorL.prev;
        p[spectator].colorL.next = m;
        p[m].colorL.prev = spectator;
        p[m].colorL.next = emitter;
        p[emitter].colorL.prev = m;
        break;
      case 2:
        spectator = p[emitter].colorL.prev;
        p[emitter].colorL.prev  = QUARK;
        p[spectator].colorL.next  = m;
        p[m].colorL.prev = spectator;
        p[m].colorL.next = ANTI_QUARK;
        break;
      case -2:
        p[p[emitter].colorL.prev].colorL.next = p[emitter].colorL.next;
        p[p[emitter].colorL.next].colorL.prev = p[emitter].colorL.prev;
        p[emitter].colorL.prev = QUARK;
        p[emitter].colorL.next  = m;
        p[m].colorL.prev = emitter;
        p[m].colorL.next = ANTI_QUARK;
        break;
      case 3:
        spectator = p[emitter].colorL.next;
        p[emitter].colorL.next = ANTI_QUARK;
        p[spectator].colorL.prev = m;
        p[m].colorL.next = spectator;
        p[m].colorL.prev = QUARK;
        break;
      case -3:
        p[p[emitter].colorL.prev].colorL.next = p[emitter].colorL.next;
        p[p[emitter].colorL.next].colorL.prev = p[emitter].colorL.prev;
        p[emitter].colorL.next = ANTI_QUARK;
        p[emitter].colorL.prev  = m;
        p[m].colorL.next = emitter;
        p[m].colorL.prev = QUARK;
        break;
      case 4:
        p[emitter].colorL.prev  = m;
        p[m].colorL.next = emitter;
        p[m].colorL.prev = QUARK;
        break;
      case -4:
        p[emitter].colorL.next  = m;
        p[m].colorL.prev = emitter;
        p[m].colorL.next = ANTI_QUARK;
        break;
      default:
        //   do nothing
        break;
    }
  }
  
  
  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate>
  void color_a(int sigma, int emitter, statistical_basis<_Proc, _Parton, _Ext>& p)
  {
    static constexpr int QUARK = parton_cstate::colorlink::quarkend;
    static constexpr int ANTI_QUARK = parton_cstate::colorlink::antiquarkend;
    
    int spectator;
    auto [_,m] = p.bounds();

    switch(sigma) {
      case 1:
        spectator = p[emitter].colorR.next;
        p[emitter].colorR.next  = m;
        p[m].colorR.prev = emitter;
        p[m].colorR.next = spectator;
        p[spectator].colorR.prev = m;
        break;
      case -1:
        spectator = p[emitter].colorR.prev;
        p[spectator].colorR.next = m;
        p[m].colorR.prev = spectator;
        p[m].colorR.next = emitter;
        p[emitter].colorR.prev = m;
        break;
      case 2:
        spectator = p[emitter].colorR.prev;
        p[emitter].colorR.prev  = QUARK;
        p[spectator].colorR.next  = m;
        p[m].colorR.prev = spectator;
        p[m].colorR.next = ANTI_QUARK;
        break;
      case -2:
        p[p[emitter].colorR.prev].colorR.next = p[emitter].colorR.next;
        p[p[emitter].colorR.next].colorR.prev = p[emitter].colorR.prev;
        p[emitter].colorR.prev = QUARK;
        p[emitter].colorR.next  = m;
        p[m].colorR.prev = emitter;
        p[m].colorR.next = ANTI_QUARK;
        break;
      case 3:
        spectator = p[emitter].colorR.next;
        p[emitter].colorR.next = ANTI_QUARK;
        p[spectator].colorR.prev = m;
        p[m].colorR.next = spectator;
        p[m].colorR.prev = QUARK;
        break;
      case -3:
        p[p[emitter].colorR.prev].colorR.next = p[emitter].colorR.next;
        p[p[emitter].colorR.next].colorR.prev = p[emitter].colorR.prev;
        p[emitter].colorR.next = ANTI_QUARK;
        p[emitter].colorR.prev  = m;
        p[m].colorR.next = emitter;
        p[m].colorR.prev = QUARK;
        break;
      case 4:
        p[emitter].colorR.prev  = m;
        p[m].colorR.next = emitter;
        p[m].colorR.prev = QUARK;
        break;
      case -4:
        p[emitter].colorR.next  = m;
        p[m].colorR.prev = emitter;
        p[m].colorR.next = ANTI_QUARK;
        break;
      default:
        //  do nothing
        break;
    }
  }
  

  /*    Inserting color strings   */
  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate>
  void insert_closed_stringL(const int *begin, const int *end, statistical_basis<_Proc, _Parton, _Ext>& p)
  {
    const int *i;
    
    p[*begin].colorL.prev = *(end-1);
    p[*begin].colorL.next = *(begin+1);
    
    for(i = begin+1; i < end-1; i++) {
      p[*i].colorL.prev = *(i-1);
      p[*i].colorL.next = *(i+1);
    }
    
    p[*i].colorL.prev = *(i-1);
    p[*i].colorL.next = *begin;
  }
  
  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate>
  void insert_closed_stringR(const int *begin, const int *end, statistical_basis<_Proc, _Parton, _Ext>& p)
  {
    const int *i;
    
    p[*begin].colorR.prev = *(end-1);
    p[*begin].colorR.next = *(begin+1);
    
    for(i = begin+1; i < end-1; i++) {
      p[*i].colorR.prev = *(i-1);
      p[*i].colorR.next = *(i+1);
    }
    
    p[*i].colorR.prev = *(i-1);
    p[*i].colorR.next = *begin;
  }
  
  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate>
  void insert_open_stringL(const int *begin, const int *end, statistical_basis<_Proc, _Parton, _Ext>& p)
  {
    const int *i;
    
    p[*begin].colorL.prev = parton_cstate::colorlink::quarkend;
    p[*begin].colorL.next = *(begin+1);
    
    for(i = begin+1; i < end-1; i++) {
      p[*i].colorL.prev = *(i-1);
      p[*i].colorL.next = *(i+1);
    }
    
    p[*i].colorL.prev = *(i-1);
    p[*i].colorL.next = parton_cstate::colorlink::antiquarkend;
  }
  
  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate>
  void insert_open_stringR(const int *begin, const int *end, statistical_basis<_Proc, _Parton, _Ext>& p)
  {
    const int *i;
    
    p[*begin].colorR.prev = parton_cstate::colorlink::quarkend;
    p[*begin].colorR.next = *(begin+1);
    
    for(i = begin+1; i < end-1; i++) {
      p[*i].colorR.prev = *(i-1);
      p[*i].colorR.next = *(i+1);
    }
    
    p[*i].colorR.prev = *(i-1);
    p[*i].colorR.next = parton_cstate::colorlink::antiquarkend;
  }
  
  
  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate>
  inline void insert_open_stringL(std::initializer_list<int> ll, statistical_basis<_Proc, _Parton, _Ext>& p) {
    insert_open_stringL(ll.begin(), ll.end(), p);
    
  }

  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate>
  inline void insert_open_stringR(std::initializer_list<int> ll, statistical_basis<_Proc, _Parton, _Ext>& p) {
    insert_open_stringR(ll.begin(), ll.end(), p);
  }

  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate>
  inline void insert_closed_stringL(std::initializer_list<int> ll, statistical_basis<_Proc, _Parton, _Ext>& p) {
    insert_closed_stringL(ll.begin(), ll.end(), p);
  }

  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate>
  inline void insert_closed_stringR(std::initializer_list<int> ll, statistical_basis<_Proc, _Parton, _Ext>& p) {
    insert_closed_stringR(ll.begin(), ll.end(), p);
  }

  
  /*    making gluon health report   */
  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate>
  inline char health_report(int l, const statistical_basis<_Proc, _Parton, _Ext>& p)
  {
    //---- the health is defined only for color 8 objects -----
    if(!(p[l].is_octet())) return '-';
    
    //---- it is a color 8, find the heath status -----
    int rprev = l, lnext;
    do {
      if((lnext = p[rprev].colorL.next) == l) return 'f';
      if((rprev = p[lnext].colorR.prev) == l) return 'h';
    } while(true);
  }
  
  
  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate>
  inline void make_health_report(statistical_basis<_Proc, _Parton, _Ext>& p)
  {
    auto[lo, hi] = bounds_of<colored>(p);
    
    for(int l = lo; l <= hi; ++l)
    {
      //---- the health is defined only for color 8 objects -----
      if(!(p[l].is_octet())) p[l].health = '-';
      
      //---- it is a color 8, find the heath status -----
      else {
        int rprev = l, lnext;
        do {
          if((lnext = p[rprev].colorL.next) == l) { p[l].health = 'f'; break;}
          if((rprev = p[lnext].colorR.prev) == l) { p[l].health = 'h'; break;}
        } while(true);
      }
    }
  }
  


  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate>
  char health_with_gluon_loop(int i, int j, statistical_basis<_Proc, _Parton, _Ext>& p)
  {
    using link = parton_cstate::colorlink;
    
    /*   insert a gluon after i and j   */
    auto[_,m] = bounds_of<colored>(p);
    
    /*   a^\dagger_+(i) operator   */
    int s = p[i].colorL.next;
    link ci = p[i].colorL, cs = p[s].colorL;
    
    p[i].colorL.next = m;
    p[m].colorL.prev = i;
    p[m].colorL.next = s;
    p[s].colorL.prev = m;
    
    
    /*   a_+(emitter) operator   */
    int r = p[j].colorR.next;
    link cj = p[j].colorR, cr = p[r].colorR;
    
    p[j].colorR.next = m;
    p[m].colorR.prev = j;
    p[m].colorR.next = r;
    p[r].colorR.prev = m;
    
    
    /*   find out the health of the inserted gluon    */
    char ret = '-';
    int rprev = m, lnext;
    
    do {
      if((lnext = p[rprev].colorL.next) == m) { ret = 'f'; break;}
      if((rprev = p[lnext].colorR.prev) == m) { ret = 'h'; break;}
    } while(true);
    
    
    /*   restore the color state    */
    p[i].colorL = ci; p[s].colorL = cs;
    p[j].colorR = cj; p[r].colorR = cr;
    
    
    return ret;
  }
  
  
  /*    calculates the color suppresssion index    */
  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate>
  int color_suppression_index_u3(const statistical_basis<_Proc, _Parton, _Ext>& c)
  {
    auto[lo,hi] = bounds_of<colored>(c);
    
    /*    fresh start   */
    int nloops = 0;
    
    for(int l = lo; l <= hi; ++l)
      if(!c[l].is_antiquark())
      {
        c[l].active = true;
        --nloops;
      } else c[l].active = false;
      
    
    /*    counting loops    */
    for(int l = lo; l <= hi; ++l)
      if(c[l].active)
      {
        int i = l, n;
        
        do
        {
          c[i].active = false;
          n = c[i].colorL.next;
          i = c[n].colorR.prev;
          c[i].active = false;
        } while(i != l);
        
        ++nloops;
      }
    
    return -nloops;
  }
    
  
  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate>
  bool check_for_tadpoles(const statistical_basis<_Proc, _Parton, _Ext>& c)
  {
    auto[lo,hi] = bounds_of<colored>(c);
    
    for(auto k = lo; k <= hi; ++k)
    {
      if(c[k].colorL.next == k && c[k].colorL.prev == k) return true;
      if(c[k].colorR.next == k && c[k].colorR.prev == k) return true;
    }
    
    return false;
  }
  
  
  
  /*   simple extension for statistical_basis to have a full color state    */
  struct statistical_basis_cext
  {
    /*   It is a good idea to store the color weight here.    */
    double color_weight = 1.0;
    
    /*   It is true if the state is consired as U(Nc) state.   */
    bool is_color_u3 = false;

    /*   Color suppression index of the "operator".
     *   This is basically the suppression index that always increses
     *   and helps us to truncate the 1/Nc^2 series.
     */
    unsigned color_suppression_index = 0u;
    
    /*   Color suppression index of the color overlap.
     *   This helps us to optimize the probability of the basis states.
     */
    int overlap_color_suppression_index;
  };
  

  
  /*  Splitting variables
   *  This class is a collection of variables for reconstracting
   *  the event after a splitting in a real splitting.
   */
  struct splitting_cvars
  {
    int sigmaL, sigmaR;
    int emitL, emitR;
    
    double weight;
    unsigned state_idx;
    int overlap_idx;
  };
  
  inline std::ostream& operator<<(std::ostream& os, const splitting_cvars& v)
  {
    os<<"["<<v.sigmaL<<", "<<v.sigmaR<<", "<<v.emitL<<", "<<v.emitR<<", "<<v.weight<<", "<<v.overlap_idx<<", "<<v.state_idx<<"]";
    return os;
  }
  
  
 
  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate> && std::derived_from<_Ext, statistical_basis_cext>
  void shower_mapping(statistical_basis<_Proc, _Parton, _Ext>& p, const splitting_cvars& vars, bool do_resize = true)
  {
    /*   resize and copy   */
    if(do_resize) p.resize_by(1);
    
    /*    assigning the new color state    */
    color_a_dagger(vars.sigmaL, vars.emitL, p);
    color_a(vars.sigmaR, vars.emitR, p);
    
    /*    issuing the health report for the new state    */
    make_health_report(p);
    
    /*   control parameters   */
    p.color_suppression_index += vars.state_idx;
    p.overlap_color_suppression_index += vars.overlap_idx;

    /*   weight parameters   */
    p.color_weight *= vars.weight;
  }
  
  
  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate> && std::derived_from<_Ext, statistical_basis_cext>
  inline void shower_mapping(statistical_basis<_Proc, _Parton, _Ext>& p,
                             const statistical_basis<_Proc, _Parton, _Ext>& q,
                             const splitting_cvars& vars)
  {
    p = q;
    shower_mapping(p, vars);
  }
  
  
 
  /*   a type alias   */
  using real_color_splitting_vars = splitting_cvars;
  
  
  /*  Splitting variables
   *  This class is a collection of variables for reconstracting
   *  the event after a splitting in a virtual splitting.
   */
  struct virtual_color_splitting_vars
  {
    char oper, side;
    int emit, absorb;
    
    double weight;
    unsigned state_idx;
    int overlap_idx;
  };
  

  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate> && std::derived_from<_Ext, statistical_basis_cext>
  void shower_mapping(statistical_basis<_Proc, _Parton, _Ext>& p, const virtual_color_splitting_vars& vars)
  {
    p.color_weight *= vars.weight;
    p.color_suppression_index += vars.state_idx;
    p.overlap_color_suppression_index += vars.overlap_idx;

    
    /*   trivial case when we basically insert a self energy on the fermion line   */
    if(vars.oper == 'G')
    {
      if(vars.emit == vars.absorb)
      {
        p.color_weight *= qcd::Nc;
      }
      
      /*   otherwise we make an extra closed loop   */
      else
      {
        /*    assigning the new color state    */
        const int &i = vars.emit, &j = vars.absorb;
        
        if(vars.side == 'L')
        {
          p[p[i].colorL.next].colorL.prev = j;
          p[p[j].colorL.next].colorL.prev = i;
          
          std::swap(p[i].colorL.next, p[j].colorL.next);
        }
        
        else if(vars.side == 'R')
        {
          p[p[i].colorR.next].colorR.prev = j;
          p[p[j].colorR.next].colorR.prev = i;
          
          std::swap(p[i].colorR.next, p[j].colorR.next);
        }
        
        /*    issuing the health report for the new state    */
        make_health_report(p);
      }
    }
  }
  
  
  template<ProcessTag _Proc, class _Parton, class _Ext>
  requires std::derived_from<_Parton, parton_cstate> && std::derived_from<_Ext, statistical_basis_cext>
  inline void shower_mapping(statistical_basis<_Proc, _Parton, _Ext>& p,
                             const statistical_basis<_Proc, _Parton, _Ext>& q,
                             const virtual_color_splitting_vars& vars)
  {
    p = q;
    shower_mapping(p, vars);
  }
  
  
}   //   namespace duct



/*    Some extra color goodies...   */
namespace duct {
  namespace color {
    
    /*    some simple functions to test color connections    */
    template<ProcessTag _Proc, class _Parton, class _Ext>
    requires std::derived_from<_Parton, parton_cstate>
    inline int tauL(int i, int l, const statistical_basis<_Proc, _Parton, _Ext>& p) {
      return (l == i ? 1 : 0) - (p[l].colorL.prev == i ? 1 : 0);
    }
    
    
    template<ProcessTag _Proc, class _Parton, class _Ext>
    requires std::derived_from<_Parton, parton_cstate>
    inline int tauR(int i, int l, const statistical_basis<_Proc, _Parton, _Ext>& p) {
      return (l == i ? 1 : 0) - (p[l].colorR.prev == i ? 1 : 0);
    }
    
    
    template<ProcessTag _Proc, class _Parton, class _Ext>
    requires std::derived_from<_Parton, parton_cstate>
    inline int chi_minus(int i, int l, const statistical_basis<_Proc, _Parton, _Ext>& p) {
      return (p[l].colorL.prev == i ? 1 : 0) + (p[l].colorR.prev == i ? 1 : 0);
    }
    
    
    template<ProcessTag _Proc, class _Parton, class _Ext>
    requires std::derived_from<_Parton, parton_cstate>
    inline int chi_plus(int i, int l, const statistical_basis<_Proc, _Parton, _Ext>& p) {
      return (p[l].colorL.next == i ? 1 : 0) + (p[l].colorR.next == i ? 1 : 0);
    }
    
    
    template<ProcessTag _Proc, class _Parton, class _Ext>
    requires std::derived_from<_Parton, parton_cstate>
    inline int chiL_minus(int i, int l, const statistical_basis<_Proc, _Parton, _Ext>& p) {
      return (p[l].colorL.prev == i ? 1 : 0);
    }
    
    
    template<ProcessTag _Proc, class _Parton, class _Ext>
    requires std::derived_from<_Parton, parton_cstate>
    inline int chiL_plus(int i, int l, const statistical_basis<_Proc, _Parton, _Ext>& p) {
      return (p[l].colorL.next == i ? 1 : 0);
    }
    
    
    template<ProcessTag _Proc, class _Parton, class _Ext>
    requires std::derived_from<_Parton, parton_cstate>
    inline int chiR_minus(int i, int l, const statistical_basis<_Proc, _Parton, _Ext>& p) {
      return (p[l].colorR.prev == i ? 1 : 0);
    }
    
    
    template<ProcessTag _Proc, class _Parton, class _Ext>
    requires std::derived_from<_Parton, parton_cstate>
    inline int chiR_plus(int i, int l, const statistical_basis<_Proc, _Parton, _Ext>& p) {
      return (p[l].colorR.next == i ? 1 : 0);
    }
    
    
    template<ProcessTag _Proc, class _Parton, class _Ext>
    requires std::derived_from<_Parton, parton_cstate>
    inline int chi(int k, int l, const statistical_basis<_Proc, _Parton, _Ext>& p)
    {
      int ret = 0;
      auto& cL = p[l].colorL;
      auto& cR = p[l].colorR;
      
      if(cL.prev == k) ++ret;
      if(cL.next == k) ++ret;
      
      if(cR.prev == k) ++ret;
      if(cR.next == k) ++ret;
      
      return ret;
    }
    
    
    
    
    namespace su3 {
      
      /*    this is the h_{ij}^{kl}({c', c}) color matrix for the Delta H(t) term    */
      template<ProcessTag _Proc, class _Parton, class _Ext>
      requires std::derived_from<_Parton, parton_cstate>
      inline int h(int i, int j, int l, int k, const statistical_basis<_Proc, _Parton, _Ext>& p)
      {
        int h = 0;
        
        
        /*    tldagger|{c}> <{c'}|tk(1-Cdagger)  contribution    */
        int tldagger = (l == i ? 1 : 0) - (p[l].colorL.prev == i ? 1 : 0);
        
        if(tldagger)
        {
          int tk = (p[l].colorR.next == k ? 0 : (p[k].colorR.prev == j ? -1 : 0));
          if(k == j) tk += (p[l].colorR.prev == k ? 0 : 1);
          
          h -= tldagger*tk;
        }
        
        /*    (1-C)tkdagger|{c}> <{c'}|tl  contribution   */
        int tl = (l == j ? 1 : 0) - (p[l].colorR.prev == j ? 1 : 0);
        
        if(tl)
        {
          int tkdagger = (p[l].colorL.next == k ? 0 : (p[k].colorL.prev == i ? -1 : 0));
          if(k == i) tkdagger += (p[l].colorL.prev == k ? 0 : 1);
          
          h -= tl*tkdagger;
        }
        
        
        return h;
      }
      
      template<ProcessTag _Proc, class _Parton, class _Ext>
      requires std::derived_from<_Parton, parton_cstate>
      inline int lc(int i, int j, int l, int k, const statistical_basis<_Proc, _Parton, _Ext>& p)
      {
        return -tauL(i,l, p)*((j == k ? chiR_minus(k,l, p) : 0) - chiR_plus(k,l, p)*chiR_minus(j,k, p))
        -tauR(j,l, p)*((i == k ? chiL_minus(k,l, p) : 0) - chiL_plus(k,l, p)*chiL_minus(i,k, p));
      }
      
      /*    this is the v_{ij}^{kl}({c}) color matrix for the Delta V(t) term    */
      template<ProcessTag _Proc, class _Parton, class _Ext>
      requires std::derived_from<_Parton, parton_cstate>
      inline int vL(int i, int j, int l, int k, const statistical_basis<_Proc, _Parton, _Ext>& p)
      {
        int v = 0;
        
        /*    tk(1-Cdagger)tldagger|{c}> <{c'}|  contribution    */
        int tldagger = (l == i ? 1 : 0) - (p[l].colorL.prev == i ? 1 : 0);
        
        if(tldagger)
        {
          int tk = (p[l].colorL.next == k ? 0 : (p[k].colorL.prev == j ? -1 : 0));
          if(k == j) tk += (p[l].colorL.prev == k ? 0 : 1);
          
          v -= tldagger*tk;
        }
        
        return v;
      }
      
      
      /*    this is the v_{ij}^{kl}({c'}) color matrix for the Delta V(t) term    */
      template<ProcessTag _Proc, class _Parton, class _Ext>
      requires std::derived_from<_Parton, parton_cstate>
      inline int vR(int i, int j, int l, int k, const statistical_basis<_Proc, _Parton, _Ext>& p)
      {
        int v = 0;
        
        /*    |{c}> <{c'}|tl(1-C)tkdagger  contribution    */
        int tl = (l == i ? 1 : 0) - (p[l].colorR.prev == i ? 1 : 0);
        
        if(tl)
        {
          int tkdagger = (p[l].colorR.next == k ? 0 : (p[k].colorR.prev == j ? -1 : 0));
          if(k == j) tkdagger += (p[l].colorR.prev == k ? 0 : 1);
          
          v -= tl*tkdagger;
        }
        
        return v;
      }
      
      
      /*    this is the \tilde{v}_{ij}^{kl}({c}) color matrix for the Delta H(t) term    */
      template<ProcessTag _Proc, class _Parton, class _Ext>
      requires std::derived_from<_Parton, parton_cstate>
      inline int phL(int i, int j, int l, int k, const statistical_basis<_Proc, _Parton, _Ext>& p) {
        return tauL(i,l, p)*tauL(j,k, p);
      }
      
      
      /*    this is the \tilde{v}_{ij}^{kl}({c'}) color matrix for the Delta H(t) term    */
      template<ProcessTag _Proc, class _Parton, class _Ext>
      requires std::derived_from<_Parton, parton_cstate>
      inline int phR(int i, int j, int l, int k, const statistical_basis<_Proc, _Parton, _Ext>& p)
      {
        return tauR(i,l, p)*tauR(j,k, p);
      }
    }   //   namespace su3
    
    
    
    namespace u3 {
      
      
      /*    this is the h_{ij}^{kl}({c', c}) color matrix for the Delta H(t) term    */
      template<ProcessTag _Proc, class _Parton, class _Ext>
      requires std::derived_from<_Parton, parton_cstate>
      inline int h(int i, int j, int l, int k, const statistical_basis<_Proc, _Parton, _Ext>& p)
      {
        /*   Glk part is the full color operator    */
        int ret = -tauL(i,l, p)*tauR(j,k, p) - tauL(i,k, p)*tauR(j,l, p);
        
        /*   LC+ part of the full Glk is subtracted    */
        ret -= (l == i && l == j ? chi_plus(k,l, p) : 0);
        ret -= (p[l].colorL.prev == i && p[l].colorR.prev == j ? chi_minus(k,l, p) : 0);
        
        return ret;
      }
      
      
      template<ProcessTag _Proc, class _Parton, class _Ext>
      requires std::derived_from<_Parton, parton_cstate>
      inline int lc(int i, int j, int l, int k, const statistical_basis<_Proc, _Parton, _Ext>& p)
      {
        return (l == i && l == j ? chi_plus(k,l, p) : 0)
        + (p[l].colorL.prev == i && p[l].colorR.prev == j ? chi_minus(k,l, p) : 0);
      }
      
      
      /*    this is the v_{ij}^{kl}({c}) color matrix for the Delta V(t) term    */
      template<ProcessTag _Proc, class _Parton, class _Ext>
      requires std::derived_from<_Parton, parton_cstate>
      inline int vL(int i, int j, int l, int k, const statistical_basis<_Proc, _Parton, _Ext>& p) {
        return -tauL(i,l, p)*tauL(j,k, p);
      }
      
      
      /*    this is the v_{ij}^{kl}({c'}) color matrix for the Delta V(t) term    */
      template<ProcessTag _Proc, class _Parton, class _Ext>
      requires std::derived_from<_Parton, parton_cstate>
      inline int vR(int i, int j, int l, int k, const statistical_basis<_Proc, _Parton, _Ext>& p) {
        return -tauR(i,l, p)*tauR(j,k, p);
      }
      
      /*   color matrix in the phase terms    */
      using su3::phL;
      using su3::phR;
    }   //  namespace u3
  }  //   namespace color
}  //  namespace duct


#endif
