/*  Copyright © 2017 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 10/03/2017.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_shower_operator_h__
#define __mc_shower_operator_h__ 1


/*   std includes   */


/*   project headers   */
#include <cxx/random.h>
#include <cxx/stackvector.h>
#include <bits/mc-shower-stage.h>



/*   Function and classes for doing the color mappings      */
namespace duct {
   
  
  /*    Implementation details of the real shower operator    */
  template<ProcessTag _Proc>
  class select_lcplus_operator
  {
  public:
    /*    do the color evolution in the real operator   */
    void operator()(random_engine_type&, shower_history_stage<_Proc>&);
    
    
  private:
    /*    sets of possible color operators    */
    stackvector<splitting_cvars, 4u> _M_cvars;
    
    
    /*   gluon emission operators   */
    void _M_lcplus_u3(const shower_history_stage<_Proc>&);
    void _M_lcplus_su3(const shower_history_stage<_Proc>&);
 
    /*   pure collinear operators    */
    void _M_collinear_u3(const shower_history_stage<_Proc>&);
    void _M_collinear_su3(const shower_history_stage<_Proc>&);
  };
  
  
  template<ProcessTag _Proc>
  void select_lcplus_operator<_Proc>::operator()(random_engine_type& rne, shower_history_stage<_Proc>& next)
  {
    /*   findig all the posssible color operators    */
    if(next.vars.emitter != next.vars.spectator)
    {
      if(next.state.is_color_u3) _M_lcplus_u3(next);
      else _M_lcplus_su3(next);
    }
    
    /*   pure collinear oprators    */
    else
    {
      if(next.state.is_color_u3) _M_collinear_u3(next);
      else _M_collinear_su3(next);
    }

    /*   there is only one possible operator   */
    if(1u == _M_cvars.size()) {
      next.cvars = _M_cvars.front();
      return;
    }
    
    
    /*    otherwise choose an operator randomly  */
    /*   we have more than one basis state   */
    double norm = 0.0;
    
    for(const auto& v : _M_cvars)
      norm += std::abs(v.weight)/std::pow(qcd::Nc, v.overlap_idx);
    
    auto vi = select_randomly(rne, _M_cvars.begin(), _M_cvars.end(),
                              [=](const splitting_cvars& ci) -> double {
                                return std::abs(ci.weight)/norm/std::pow(qcd::Nc, ci.overlap_idx);
                              });
 
    vi->weight = std::copysign(norm, vi->weight);
    vi->weight *= std::pow(qcd::Nc, vi->overlap_idx);
    
    /*   do the transformation   */
    next.cvars = *vi;
  }

  
  template<ProcessTag _Proc>
  void select_lcplus_operator<_Proc>::_M_lcplus_su3(const shower_history_stage<_Proc>& next)
  {
    int l = next.vars.emitter, k = next.vars.spectator;
    auto& c = next.state;
    
    _M_cvars.clear();
    
    /*  spliting e.g.: g -> g + g    */
    if(c[l].is_octet())
    {
      const int pL = color::chiL_plus(k,l, c), pR = color::chiR_plus(k,l, c);
      const int mL = color::chiL_minus(k,l, c), mR = color::chiR_minus(k,l, c);
      const double cfac = 2.0*qcd::Cf/qcd::Ca/(pL+pR + mL+mR);
      
      int suppr_idx = (c[l].health == 'h' ? 2 : 0);
      
      _M_cvars.push_back({ 1,  1, l, l,  cfac*(pL+pR), 0u, 0});
      _M_cvars.push_back({-1, -1, l, l,  cfac*(mL+mR), 0u, 0});
      _M_cvars.push_back({ 1, -1, l, l, -cfac*(pL+mR), static_cast<unsigned>(suppr_idx), suppr_idx});
      _M_cvars.push_back({-1,  1, l, l, -cfac*(mL+pR), static_cast<unsigned>(suppr_idx), suppr_idx});
    }
    
    /*   splitting e.g.: q -> q + g   */
    else if(c[l].is_triplet()) _M_cvars.push_back({1, 1, l, l, 1.0, 0u, 0});
    
    /*   splitting e.g.: q~ -> q~ + g   */
    else if(c[l].is_antitriplet()) _M_cvars.push_back({-1,-1, l, l, 1.0, 0u, 0});

    /*   unkown splitting    */
    else throw "color_evolution<_Proc>::_M_lcplus_su3(...) : colorless parton cannot split";
  }
  
  
  template<ProcessTag _Proc>
  void select_lcplus_operator<_Proc>::_M_lcplus_u3(const shower_history_stage<_Proc>& next)
  {
    int l = next.vars.emitter, k = next.vars.spectator;
    auto& c = next.state;
    
    _M_cvars.clear();
 
    /*  spliting e.g.: g -> g + g    */
    if(c[l].is_octet())
    {
      const int plus = color::chi_plus(k,l, c), minus = color::chi_minus(k,l, c);
      const double cfac = 1.0/(plus + minus);
      
      _M_cvars.push_back({ 1,  1, l, l,  cfac*plus, 0u, 0});
      _M_cvars.push_back({-1, -1, l, l,  cfac*minus, 0u, 0});
    }
    
    /*   splitting e.g.: q -> q + g   */
    else if(c[l].is_triplet()) _M_cvars.push_back({1, 1, l, l, 1.0, 0u, 0});
    
    /*   splitting e.g.: q~ -> q~ + g   */
    else if(c[l].is_antitriplet()) _M_cvars.push_back({-1,-1, l, l, 1.0, 0u, 0});

    /*   unkown splitting    */
    else throw "color_evolution<_Proc>::_M_lcplus_u3(...) : colorless parton cannot split";
  }
  

  template<ProcessTag _Proc>
  void select_lcplus_operator<_Proc>::_M_collinear_su3(const shower_history_stage<_Proc>& next)
  {
    int l = next.vars.emitter, cl = next.vars.cl, cm = next.vars.cm;
    auto& c = next.state;
    
    _M_cvars.clear();
    
    /*  spliting e.g.: g -> g + g    */
    if(cl == 8 && cm == 8)
    {
      int suppr_idx = (c[l].health == 'h' ? 2 : 0);
      
      _M_cvars.push_back({ 1,  1, l, l,  qcd::Cf/qcd::Ca, 0u, 0});
      _M_cvars.push_back({-1, -1, l, l,  qcd::Cf/qcd::Ca, 0u, 0});
      _M_cvars.push_back({ 1, -1, l, l, -qcd::Cf/qcd::Ca, static_cast<unsigned>(suppr_idx), suppr_idx});
      _M_cvars.push_back({-1,  1, l, l, -qcd::Cf/qcd::Ca, static_cast<unsigned>(suppr_idx), suppr_idx});
    }
    
    /*   splitting e.g.: q -> q + g   */
    else if(cl == 3 && cm == 8) _M_cvars.push_back({1, 1, l, l, 1.0, 0u, 0});
    
    /*   splitting e.g.: q~ -> q~ + g   */
    else if(cl == -3 && cm == 8) _M_cvars.push_back({-1,-1, l, l, 1.0, 0u, 0});
    
    /*  spliting e.g.: g -> q + q~   */
    else if(std::abs(cl) == 3 && cm == -cl)
    {
      int sigma = (cl == 3 ? 2 : 3);
      int suppr_idx = (c[l].health == 'f' ? -2 : 0);
      unsigned suppr_uidx = (c[l].health == 'f' ? 0u : 2u);

      bool tadL = (c[l].colorL.next == c[l].colorL.prev ? true : false);
      bool tadR = (c[l].colorR.next == c[l].colorR.prev ? true : false);
      
      _M_cvars.push_back({ sigma,  sigma, l, l,  qcd::Nc2/qcd::Na, 0u, 0});

      if(!tadL && !tadR) _M_cvars.push_back({-sigma, -sigma, l, l,  1.0/qcd::Na, suppr_uidx, suppr_idx});
      if(true  && !tadR) _M_cvars.push_back({ sigma, -sigma, l, l, -qcd::Nc/qcd::Na, suppr_uidx, 1+suppr_idx});
      if(!tadL &&  true) _M_cvars.push_back({-sigma,  sigma, l, l, -qcd::Nc/qcd::Na, suppr_uidx, 1+suppr_idx});
    }
    
    /*  splitting e.g.: q -> g + q in the initial state   */
    else if(cl == 8 && cm == 3) _M_cvars.push_back({4, 4, l, l, 1.0, 0u, 0});
    
    /*  splitting e.g.: q~ -> g + q~ in the initial state    */
    else if(cl == 8 && cm == -3) _M_cvars.push_back({-4, -4, l, l, 1.0, 0u, 0});
    
    
    /*   unkown splitting    */
    else throw "color_evolution<_Proc>::_M_collinear_operators_su3(...) : unable to assign the new color state";
  }
  
  
  template<ProcessTag _Proc>
  void select_lcplus_operator<_Proc>::_M_collinear_u3(const shower_history_stage<_Proc>& next)
  {
    int l = next.vars.emitter, cl = next.vars.cl, cm = next.vars.cm;
    
    _M_cvars.clear();
    
    /*  splitting e.g.: g -> g + g    */
    if(cl == 8 && cm == 8)
    {
      _M_cvars.push_back({ 1,  1, l, l, 0.5, 0u, 0});
      _M_cvars.push_back({-1, -1, l, l, 0.5, 0u, 0});
    }
    
    /*   splitting e.g.: q -> q + g   */
    else if(cl == 3 && cm == 8) _M_cvars.push_back({1, 1, l, l, 1.0, 0u, 0});
    
    /*   splitting e.g.: q~ -> q~ + g   */
    else if(cl == -3 && cm == 8) _M_cvars.push_back({-1,-1, l, l, 1.0, 0u, 0});
    
    /*  spliting e.g.: g -> q + q~   */
    else if(cl == 3 && cm == -cl) _M_cvars.push_back({2, 2, l, l, 1.0, 0u, 0});
    else if(cl ==-3 && cm == -cl) _M_cvars.push_back({3, 3, l, l, 1.0, 0u, 0});
    
    /*  splitting e.g.: q -> g + q in the initial state   */
    else if(cl == 8 && cm == 3) _M_cvars.push_back({4, 4, l, l, 1.0, 0u, 0});
    
    /*  splitting e.g.: q~ -> g + q~ in the initial state    */
    else if(cl == 8 && cm == -3) _M_cvars.push_back({-4, -4, l, l, 1.0, 0u, 0});
    
    
    /*   unkown splitting    */
    else throw "color_evolution<_Proc>::_M_lcplus_operators_u3(...) : unable to assign the new color state";
  }
  
  
}   //   namespace duct


#endif
