/*  Copyright © 2016 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 15/01/16.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __core_mc_dglap_h__
#define __core_mc_dglap_h__ 1


/*   std includes   */

/*   project headers   */
#include <bits/duct-tags.h>
#include <bits/mc-coupling.h>
#include <bits/qcd-dglap.h>



namespace duct {
 
 
  
  /*    This class can be used to solve the DGLAP evolution equation.
   */
  class dglap
  {
  public:
    /*    constructors   */
    dglap() : _M_m2T{1.0_GeV2}, _M_m2c{1.69_GeV2}, _M_m2b{20.25_GeV2} {}
    
    dglap(energy_sqr<double> m2T, energy_sqr<double> m2c, energy_sqr<double> m2b, qcd_coupling&& as)
    : _M_as{std::move(as)}, _M_m2T{m2T}, _M_m2c{m2c}, _M_m2b{m2b} {}
    
    dglap(energy_sqr<double> m2T, energy_sqr<double> m2c, energy_sqr<double> m2b, const qcd_coupling& as)
    : _M_as{as}, _M_m2T{m2T}, _M_m2c{m2c}, _M_m2b{m2b} {}
    
    
    /*   do the evolution  */
    void evolve(ordering::scheme, std::function<void(double, double *)>);
    
    
    /*   calculate the pdf for a given flavor
     *
     *      81, -81 => u+ubar + d+dbar + s+sbar + c+cbar + b+bbar
     *      82, -82 => c+cbar
     *      83, -83 => b+bbar
     *
     *   Important: the scale parameter is tauk, thus we have 
     *   (f, x, tauk) arguments.
     */
    double parton(int, double, double) const;
    double parton_for_mc(int, double, double) const;
  
    /*   It is the approximation of the heavy quarks near to the threshold scale.
     *   This function is rather slow since it is based on two nested gauss integral.
     */
    double heavy_quark_approx(int, double, double) const;
    
    
    /*   usual way to call the pdf function (f, x, q2)   */
    double operator()(int f, double x, energy_sqr<double> q2) const {
      return parton(f, x, _M_as.tauk(q2));
    }
    
    
    /*   returns the factorization scheme    */
    ordering::scheme scheme() const { return _M_scheme;}
    
    /*   return the absolute heavy flavor thresholds    */
    energy_sqr<double> thrs(int fl) const
    {
      switch(std::abs(fl)) {
        case 4: case 12: case 82: return _M_m2c; break;
        case 5: case 13: case 83: return _M_m2b; break;
        case 6: return 1.0e+100_GeV2;
      }
      
      return 0.0_GeV2;
    }
    
    energy_sqr<double> thrsT(int fl) const
    {
      switch(std::abs(fl)) {
        case 4: case 12: case 82: return std::max(_M_m2T, _M_m2c); break;
        case 5: case 13: case 83: return std::max(_M_m2T, _M_m2b); break;
        case 6: return 1.0e+100_GeV2;
      }
      
      return _M_m2T;
    }
    
    /*   return the heavy flavor thresholds    */
    energy_sqr<double> thrs(int f, double eta) const
    {
      switch(_M_scheme) {
        case ordering::scheme::kT:      return thrs(f); break;
        case ordering::scheme::lambda:  return thrs(f)/(1.0-eta); break;
        case ordering::scheme::angular: return thrs(f)/sqr(1.0-eta); break;
      }
      
      return 0.0_GeV2;
    }
    
    energy_sqr<double> thrsT(int f, double eta) const
    {
      switch(_M_scheme) {
        case ordering::scheme::kT:      return thrsT(f); break;
        case ordering::scheme::lambda:  return thrsT(f)/(1.0-eta); break;
        case ordering::scheme::angular: return thrsT(f)/sqr(1.0-eta); break;
      }
      
      return _M_m2T;
    }
    

    /*   returns the lower x bound    */
    double xmin() const { return _M_x[0];}
    
    /*    returns the fitting scale   */
    energy_sqr<double> q2min() const { return _M_m2T;}
    
    /*   returns the coupling object    */
    const qcd_coupling& coupling() const { return _M_as;}
    
    
    /*   I/O  operations   */
    void read(std::istream& is) {
      unsigned sch;
      duct::read(is, _M_as, sch, _M_m2T, _M_m2c, _M_m2b, _M_x, _M_t, _M_h);
      _M_scheme = static_cast<ordering::scheme>(sch);
    }
    
    void write(std::ostream& os) const {
      duct::write(os, _M_as, static_cast<unsigned>(_M_scheme), _M_m2T, _M_m2c, _M_m2b, _M_x, _M_t, _M_h);
    }
    
  private:
    /*   strong coupling   */
    qcd_coupling _M_as;
    
    /*    factorization scheme   */
    ordering::scheme _M_scheme;
    
    /*   transverse momentum cutoff (starting scale)    */
    energy_sqr<double> _M_m2T, _M_m2c, _M_m2b;
    
    /*   t grid     */
    std::vector<double> _M_x, _M_t;
    
    /*   the interpolation tables  */
    __details::pdf_matrix _M_h;
    
    /*   private functions to set up the grid    */
    void _M_xt_grid();
    
    /*   function to calculate the weights in the grid points    */
    __details::pdf_kernel _M_weights(ordering::scheme, unsigned, unsigned, unsigned) const;

    /*    accessing to the grid directly   */
    double _M_parton(int, unsigned, unsigned) const;
  };

  
}   //   namespace duct

#endif
