/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 4/24/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_engine_h__
#define __mc_engine_h__ 1



/*   std includes   */


/*   project headers   */
#include <bits/mc-engine-opts.h>
#include <bits/mc-splitting-prob.h>
#include <bits/mc-shower-history.h>
#include <bits/qcd-color-overlap.h>
#include <cxx/cloned-ptr.h>


namespace duct {
  
  
  /*  This is the core of the shower evolution code. It uses a given shower probability class to make
   *  trial emission. An instance of the shower probability class is stored in this object.
   */
  template<ProcessTag _Proc, OrderingTag _Ordering>
  class shower_engine
  {
  public:
    /*    constructors  */
    shower_engine() = default;
    
    explicit shower_engine(model_ref mdl, const shower_engine_opts& opts, std::size_t s)
    : _M_random{s}, _M_opts{opts}, _M_prob{mdl, sqr(opts.kT_cutoff)} {}
    
    
    /*   Evolution operator of the threshold part.    */
    void threshold_stage(shower_history_stage<_Proc>&) {}

    /*    This is the soft stage of the shower evolution. It needs to be implemented!!!    */
    constexpr bool soft_stage(energy_sqr<double>, shower_history_stage<_Proc>&) {
      return false;
    }
    
    /*   make a trial emission and transforms the color state    */
    bool lcplus_stage(energy_sqr<double> v2F, shower_history_stage<_Proc>& current)
    {
      /*   no emission is possible    */
      if(v2F >= current.next_scale || std::abs(current.state.color_weight) < 1e-25)
        return false;
      
      
      /*   make a trial emission   */
      bool is_possible = _M_prob.make_trial(_M_random, current, v2F, current.next_scale);
      
      
      /*   do the real emission   */
      if(is_possible)
      {
        shower_mapping(current.state, current.vars, current.cvars);
        current.scale = current.current_scale = current.next_scale;
      }
      
      /*   transform to U(3) if it is needed    */
      _M_overlap.convert_to_u3(_M_opts.max_color_suppression, current.state);
      
      
      return is_possible;
    }
    
    
    /*   Post-shower stage. Here we consider further MC reduction to speed up the user routine.    */
    constexpr void postshower_stage(shower_history_stage<_Proc>&)
    {}
    
  private:
    /*   random engine   */
    random_engine_type _M_random;
    
     
    /*   engine parameters    */
    shower_engine_opts _M_opts;
 
    
    /*   the underlying splitting probability   */
    splitting_prob<_Proc, _Ordering> _M_prob;
    
    
    /*   calculates color overlaps in the evolution    */
    color_overlap _M_overlap;
  };

  
  
  
  /*  This is the shower evolution code. The type of the ordering class is erased.    */
  template<ProcessTag _Proc>
  class shower_evolution
  {
  public:
    /*    default constructor    */
    shower_evolution() = default;
    
    /*    create the shower engine from a given shower probability   */
    template<OrderingTag _Ordering>
    explicit shower_evolution(_Ordering, model_ref mdl, const shower_engine_opts& opts, std::size_t s)
    : _M_trial{new obj<_Ordering>{mdl, opts, s}} {}
    
    
    /*   shower evolution    */
    void operator()(shower_history<_Proc>& hist) {
      _M_trial->evolve(hist);
    }
    
    
  private:
 
    /*   abstract class for type erasing   */
    struct base {
      virtual void evolve(shower_history<_Proc>&) = 0;
    };
    
    /*   this stores an instanse of the actual splitting probability   */
    template<OrderingTag _Ordering>
    struct obj : base
    {
      explicit obj(model_ref mdl, const shower_engine_opts& opts, std::size_t s) : engine{mdl, opts, s} {}
      
      void evolve(shower_history<_Proc>& hist)
      {
        using _Engine = shower_engine<_Proc, _Ordering>;
        
        /*   if it is the first step we have to apply the thrshold operator   */
        hist.threshold_stage(&_Engine::threshold_stage, engine);
        
        
        /*   soft shower evolution    */
        while(hist.shower_stage(0.0_GeV2, &_Engine::soft_stage, engine));

        /*   LC+ shower evolution    */
        while(hist.shower_stage(0.0_GeV2, &_Engine::lcplus_stage, engine));
        
        
        /*   post shower stage   */
        hist.postshower_stage(&_Engine::postshower_stage, engine);
      }
      
      
      /*   the underlying splitting probability   */
      shower_engine<_Proc, _Ordering> engine;
    };
    
    
    /*   we put an instance of obj into a cloned_ptr   */
    cloned_ptr<base> _M_trial;
  };
  
}   /*   namespace duct   */

#endif
