/*  Copyright © 2017 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 19/02/2017.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_event_h__
#define __mc_event_h__ 1


/*   std includes   */


/*   project headers   */
#include <bits/qcd-color.h>
#include <bits/duct-model.h>
#include <bits/mc-statistical-basis.h>
#include <bits/mc-color-basis.h>
#include <cxx/bounded-array.h>




/*    Here we define the main event class.
 *    It also consists of a single color state.
 */
namespace duct {
    
  /*  This class is for storing one particle informations.
   *    => flavor/key    : flavor of the particle
   *    => momentum      : four mementum of the particle
   */
  struct hadron_state
  {
    /*   flavor information   */
    int key = 0;

    /*    momentum of the particle    */
    lorentzvector<energy<double>> momentum;
  };
 
  
  /*  This class is for storing one particle informations.
   *    => flavor/key    : flavor of the particle
   *    => momentum      : four mementum of the particle
   *    => color         : from base class
   */
  struct parton_state : public parton_cstate
  {
    /*   flavor information   */
    int flavor = 0;

    /*    momentum of the particle    */
    lorentzvector<energy<double>> momentum;
  };
  
 
  /*    printing a parton state   */
  inline std::ostream& operator<<(std::ostream& os, const parton_state& p) {
    return os<<"["<<p.flavor<<"] "<<p.momentum<<"  "<<mag2(p.momentum)
    <<"  "<<static_cast<const parton_cstate&>(p);
  }
  
  
  /*    extension for event classes
   *      => standard color state extension
   *      => it takes care about the incomming hadrons
   */
  template<ProcessTag _Proc>
  struct statistical_basis_event_ext : public statistical_basis_cext {};
  
  template<>
  struct statistical_basis_event_ext<dis> : public statistical_basis_cext
  {
    bounded_array<hadron_state, 0, 0> hadron;
  };
  
  template<>
  struct statistical_basis_event_ext<hhc> : public statistical_basis_cext
  {
    bounded_array<hadron_state, -1, 0> hadron;
  };
  

  
  /*     definition of a partonic event    */
  template<ProcessTag _Proc>
  using event = statistical_basis<_Proc, parton_state, statistical_basis_event_ext<_Proc>>;

  template<ProcessTag _Proc>
  inline double eta(int i, const event<_Proc>& p)
  {
    /*   DIS   */
    if constexpr(std::same_as<_Proc, dis>)
      return (0 == i ? (p[0].momentum*p[-1].momentum)/(p.hadron[0].momentum*p[-1].momentum) : 1.0);
  
    /*   hadron-hadron collision   */
    else if constexpr(std::same_as<_Proc, hhc>)
    {
      auto s = p.hadron[0].momentum*p.hadron[-1].momentum;
      
      if(0 == i) return (p[0].momentum*p.hadron[-1].momentum)/s;
      else if(-1 == i) return (p[-1].momentum*p.hadron[0].momentum)/s;
    }
    
    return 1.0;
  }

  
  /*   momentum sum over a range   */
  template<ColoredTag _Range, IniFinTag _Type = inifin, ProcessTag _Proc>
  inline auto momentum_sum(const event<_Proc>& p)
  {
    lorentzvector<energy<double>> retval;
    
    for(const auto& ri : p.template crange<_Range, _Type>())
      retval += ri.momentum;
    
    return retval;
  }
  
  template<IniFinTag _Type, ProcessTag _Proc>
  inline auto momentum_sum(const event<_Proc>& p)
  {
    lorentzvector<energy<double>> retval;
    
    for(const auto& ri : p.template crange<uncolored, _Type>())
      retval += ri.momentum;

    for(const auto& ri : p.template crange<colored, _Type>())
      retval += ri.momentum;

    return retval;
  }
  

  /*    print the statistical_basis   */
  template<ProcessTag _Proc, class _Ext>
  std::ostream& operator<<(std::ostream& os, const statistical_basis<_Proc, parton_state, _Ext>& a)
  {
    /*   code for printing the partons    */
    model_ref mdl{get_default_model()};
    
    /*     print the momenta   */
    auto[idx,_] = a.bounds();
    
    os<<"+----\n"
      <<"| Is it a U(3) state?             : "<<a.is_color_u3<<"\n"
      <<"| Color supression index          : "<<a.color_suppression_index<<"\n"
      <<"| Overlap color suppression index : "<<a.overlap_color_suppression_index
      <<", "<<color_suppression_index_u3(a)<<"\n";
    
    os<<"+---- uncolored partons  -----\n";
      for(auto& pi : range<uncolored>(a))
        os<<"|  "<<mdl.name(pi.flavor)<<"("<<idx++<<") : "<<pi<<'\n';
    
    os<<"+---- colored partons -----\n";
    for(auto& pi : range<colored>(a))
      os<<"|  "<<mdl.name(pi.flavor)<<"("<<idx++<<") : "<<pi<<'\n';
    
    os<<"+----\n\n\n";
    
    auto itot = momentum_sum<ini>(a);
    auto otot = momentum_sum<fin>(a);
    
    os<<"Total incoming momentum     : "<<itot<<'\n';
    os<<"Total outgoing momentum     : "<<otot<<'\n';
    os<<"Momentum conservation       : "<<itot-otot<<"\n\n";
    
    return os;
  }
  
}    //   namespace duct



/*    Functions to do the shower mapping for momenta and flavours.
 *    We don't need a separate header file for this.
 */
namespace duct {
  
  /*  Splitting variables
   *  This class is a collection of variables for reconstracting
   *  the event after a splitting in a real splitting.
   */
  struct splitting_vars
  {
    /*  label of the emitter and the possible spectator   */
    int emitter, spectator;
    
    /*  flavor splitting variable   */
    int fl, fm;
    
    /*  color splitting variable   */
    int cl, cm;
    
    /*  momentum splitting variables   */
    double y, z, phi;
  };
  
  
  
  /*    Some helper fuctions to compute the momentum mappings    */
  namespace __details {
    
    /*  Momentum mapping in the final state shower.
     */
    template<ProcessTag _Proc>
    static void momentum_mapping_fin(event<_Proc>& p, const splitting_vars& vars)
    {
       /*   iterators   */
      auto qcd = range<colored,fin>(p);
      auto p1 = qcd.begin(), pn = qcd.end()-1;
      auto pl = p1+vars.emitter-1;
      
      /*    total incoming momentum &  parameters   */
      const auto Q = momentum_sum<ini>(p);
      double al = 0.5*mag2(Q)/(pl->momentum*Q);
      
      
      /*    y dependent variables    */
      const double &y = vars.y, &z = vars.z;
      double lmd = std::sqrt(std::abs(sqr(1.0+y) - 4.0*al*y));
      double x0 = 4.0*al*y/sqr(1.0+y+lmd), Fg = 0.5*(1.0+y + lmd);

      
      /*    momenta kl, nl and Pl    */
      const auto &kl = pl->momentum;
      const auto nl = Q/al - kl;
      const auto Pl = Fg*(kl + x0*nl);
      
      /*    Lorentz boost   */
      double op = 2.0*y/(1.0-y+lmd), om = -y/Fg;
      energy_sqr<double> nk = kl*nl;
      
      auto boost = [=](const lorentzvector<energy<double>>& qnull) -> lorentzvector<energy<double>> {
        return qnull + kl*(op*(nl*qnull)/nk) + nl*(om*(kl*qnull)/nk);
      };
      
      /*    boosting the non-qcd momenta    */
      for(auto& ki : range<uncolored,fin>(p))
        ki.momentum = boost(ki.momentum);
  
      /*    boosting the qcd momenta    */
      for(auto pi = p1; pi != pn; ++pi)
        if(pi != pl) pi->momentum = boost(pi->momentum);
      
      
      /*    the transverse momentum in the c.m. of Pl   */
      const energy<double> mT = sqrt(z*(1.0-z)*x0*2.0*nk);
      lorentzvector<energy<double>> kT{0.0_GeV, mT*std::cos(vars.phi), mT*std::sin(vars.phi), 0.0_GeV};
      
      /*    boosting the transverse momentum back to the original frame    */
      lorentzvector<energy<double>> ql = kl, qk = p[vars.spectator].momentum;
      const threevector<double> bvec = Pl.boostVector();
      
      ql.boost(-bvec);
      double rPhi = ql.phi(), rTheta = ql.theta();
      
      if(vars.spectator != vars.emitter) {
        qk.boost(-bvec);
        qk.rotate(3,-rPhi);
        qk.rotate(2, -rTheta);
        kT.rotate(3, qk.phi());
      }
      
      kT.rotate(2, rTheta);
      kT.rotate(3, rPhi);
      kT.boost(bvec);
      
      /*    the daughter momenta & flavors    */
      pl->flavor = vars.fl;
      pl->momentum = Fg*(z*kl + (1.0-z)*x0*nl - kT);
      
      pn->flavor = vars.fm;
      pn->momentum = Pl-pl->momentum;
    }
    
    
    
    /*  Momentum mapping in the initial state shower.
     */
    template<HadronCollision _Proc>
    static void momentum_mapping_ini(event<_Proc>& p, const splitting_vars& vars)
    {
      /*    iterators    */
      auto ini = range<colored,ini>(p), qcd = range<colored,fin>(p);
      auto pa = ini.begin(), pn = qcd.end()-1;
      auto pb = pa+1, p1 = qcd.begin();
      
  
      /* incoming momentum   */
      if(vars.emitter == 0) std::swap(pa,pb);
      lorentzvector<energy<double>> na = pa->momentum, nb = pb->momentum;
      
      /*    parameters    */
      const energy_sqr<double> Q2 = 2.0*(na*nb);
      const double &z = vars.z, &y = vars.y;
      const double xa = 1.0/z-1.0-y, xb = z*y;
      const double mT = std::sqrt(xa*xb);
      
      /*   Now we have every parameters we can define the momenta   */
      lorentzvector<double> kT{0.0, -mT*std::cos(vars.phi), -mT*std::sin(vars.phi), 0.0};
      
      /*   azimuthal angle is measured from the color connected parton   */
      if(vars.spectator != vars.emitter) kT.rotate(3, p[vars.spectator].momentum.phi());
      
      /*    the daughter momenta    */
      pa->flavor = vars.fl;
      pa->momentum = na/z;
      
      pn->flavor = vars.fm;
      pn->momentum = xa*na + xb*nb + sqrt(Q2)*kT;
      
      
      /*    Lorentz boost    */
      double omega = 1.0/z-xa, nn = na*nb, kT2 = kT.mag2();
      energy<double> rs = std::sqrt(2.0*nn);

      auto boost = [=](const lorentzvector<energy<double>>& qnull) -> lorentzvector<energy<double>>
      {
        double alpha = qnull*nb/nn, beta = qnull*na/nn;
        lorentzvector<energy<double>> pT = qnull - alpha*na - beta*nb;
        
        beta -= kT2*alpha - 2.0*(kT*pT)/rs;
        pT -= alpha*rs*kT;
        
        return (alpha*omega)*na + (beta/omega)*nb + pT;
      };
      
      
      /*      boosting the non-qcd momenta     */
      for(auto& ki : range<uncolored,fin>(p))
        ki.momentum = boost(ki.momentum);
      
      /*    boosting the qcd momenta    */
      for(auto& pi : std::ranges::subrange(p1, pn))
        pi.momentum = boost(pi.momentum);
    }
    
  }   //   namespace __details
  
  
  /*  This functions do the mapping in the color, flavor and momentum space.   */
  template<ProcessTag _Proc>
  inline void shower_mapping(event<_Proc>& p, const splitting_vars& vars, bool do_resize = true)
  {
    /*    making space for the extra parton   */
    if(do_resize) p.resize_by(1);
    
    if(vars.emitter < 1 - static_cast<int>(_Proc::number_of_qcd_in))
      throw "shower_mapping(event<_Proc>&,....) : Unable to do the mapping, emitter is not a QCD parton.";
    
    /*   initial state mapping   */
    if constexpr(is_hadron_collision<_Proc>)
    {
      if(vars.emitter < 1) {
        __details::momentum_mapping_ini(p, vars);
        return;
      }
    }
    
    /*   final state mapping   */
    __details::momentum_mapping_fin(p, vars);
  }
  
  
  template<ProcessTag _Proc>
  inline void shower_mapping(event<_Proc>& p, const splitting_vars& vars, const splitting_cvars& cvars)
  {
    shower_mapping(p, vars);
    shower_mapping(p, cvars, false);
  }
  
  
  /*  This functions do the mapping in the color, flavor and momentum space.   */
  template<ProcessTag _Proc>
  inline void shower_mapping(event<_Proc>& p, const event<_Proc>& q, const splitting_vars& vars)
  {
    p = q;
    shower_mapping(p, vars);
  }
  
  template<ProcessTag _Proc>
  inline void shower_mapping(event<_Proc>& p, const event<_Proc>& q,
                             const splitting_vars& vars, const splitting_cvars& cvars)
  {
    p = q;
    shower_mapping(p, vars, cvars);
  }

}   //   namespace duct


#endif
