/*  Copyright © 2016 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 23/02/16.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_ordering_angle_h__
#define __mc_ordering_angle_h__ 1



/*   std headers   */
#include <cmath>
#include <utility>

/*   project headers   */
#include <bits/mc-hard-params.h>
#include <cxx/math-support.h>



namespace duct {
  
  template<>
  struct ordering_traits<fin, ordering::angular>
  {
    /*   ordering specific properties in the main shower    */
    struct main
    {
      /*  calculates the limits on the variable v    */
      static auto v2limits(const hard_params<fin>& pars) {
        return std::pair{16.0*pars.mT*pars.al*pars.Q2/sqr(1.0+4.0*pars.mT), pars.Q2};
      }
   
      /*    calculates the z limit   */
      static auto zlimits(const hard_params<fin>& pars, energy_sqr<double> v2)
      {
        const double v = v2/pars.Q2;
        const double &al = pars.al, &mT = pars.mT;
        const double r = 0.25*(1.0/v-1.0);
        const double Zmin = mT*sqr(std::sqrt(al) + std::sqrt(al-1.0 + r/mT)) - r;
        const double zp = (Zmin >= 0.25 ? 0.4999999 : 0.5+std::sqrt(0.25-Zmin));
      
        return std::pair{1.0-zp, zp};
      }
      
      /*    mapping the independent variables to y   */
      static auto mapping_to_y(const hard_params<fin>& pars, energy_sqr<double> v2, double z)
      {
        const double v = v2/pars.Q2;
        const double r1 = 4.0*v*z*(1.0-z);
        const double r2 = pars.al*(1.0-v + r1);
        const double s1 = std::sqrt(r2), s2 = std::sqrt(r2-r1);
        
        return std::pair{r1/sqr(s1+s2), pars.al/(s1*s2)};
      }
    };
    
    /*    we might need other components    */
  };
  
  
  
  
  template<>
  struct ordering_traits<ini, ordering::angular>
  {
    /*   ordering specific properties in the main shower    */
    struct main
    {
      /*  calculates the limits on the variable v    */
      static auto v2limits(const hard_params<ini>& pars)
      {
        const double &eta = pars.eta;
        return std::pair{pars.mT*pars.Q2*eta/sqr(1.0-eta), pars.Q2};
      }
      
      /*    calculates the z limit   */
      static auto zlimits(const hard_params<ini>& pars, energy_sqr<double> v2)
      {
        const double x = pars.mT*pars.Q2/v2;
        return std::pair{pars.eta, 4.0/sqr(std::sqrt(x)+std::sqrt(x+4.0))};
      }
      
      /*    mapping the independent variables to y   */
      static auto mapping_to_y(const hard_params<ini>& pars, energy_sqr<double> v2, double z) {
        return std::pair{v2/pars.Q2*(1.0/z-1.0), 1.0};
      }
    };
    
    /*    we might need other components    */
  };


  
}    //   namespace duct


#endif
