/*  Copyright © 2021 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 07.01.21 at 10:01.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_splitting_prob_fin_h__
#define __mc_splitting_prob_fin_h__ 1

/*   std includes   */


/*   project headers   */
#include <bits/mc-ordering.h>
#include <bits/mc-splitting-func.h>
#include <bits/mc-shower-stage.h>
#include <cxx/math-support.h>
#include <cxx/random.h>



namespace duct {

  
  template<class _Ordering>
  class splitting_prob<fin, _Ordering>
  {
    /*   size of the interpolation grid    */
    static constexpr unsigned int _N_zgrid = 8u;
    static constexpr unsigned int _N_vgrid = 20u;

    /*   possible splitting state  */
    struct state
    {
      /*    check if the splitting possible with a given v   */
      bool is_possible(energy_sqr<double> v2) const {
        return (v2min < v2 && v2 < v2max);
      }
      
      /*    hard parameters   */
      hard_params<fin> pars;
      
      /*   available range in variable v   */
      energy_sqr<double> v2min, v2max;
      
      /*   splitting function object  */
      stackvector<splitting_functions<fin>::result, 6u> weights;
      
      /*   z mapping   */
      __details::z_mapping<fin, _Ordering> zmap;
    
      /*   hardness and momentum fraction variables   */
      double y, z;
    };
      
    
    /*   storing the details of the last call   */
    struct lastcall
    {
      /*   make a fresh start   */
      void clear()
      {
        states.clear();
        weight_tot = as = 0.0;
        v2 = -1.0_GeV2;
        u = -1.0;
      }
      
      /*    list of all the possible states    */
      std::vector<state> states;
      
      /*    total weight of the states   */
      double weight_tot = 0.0;
      
      /*   the v and u variables   */
      energy_sqr<double> v2 = -1.0_GeV2;
      double u = -1.0;
      
      /*   strong coupling at the shower scale   */
      double as = 0.0;
    };
    
    
  public:
    /*    public types    */
    using ordering_type = _Ordering;
    
    
    /*   default constructor   */
    splitting_prob()
    : _M_mdl{get_default_model()}, _M_m2T{1.0_GeV2}
    , _M_v2min{1.0e+100_GeV2}, _M_v2max{0.0_GeV2}
    , _M_tauL{1.0e+120}, _M_tauH{0.0}
    , _M_z_spline{_N_zgrid}
    , _M_v_spline{_N_vgrid}
    {}

    /*   with custom model resource    */
    splitting_prob(model_ref mdl, energy_sqr<double> m2T)
    : _M_mdl{mdl}, _M_m2T{m2T}
    , _M_v2min{1.0e+100_GeV2}, _M_v2max{0.0_GeV2}
    , _M_tauL{1.0e+120}, _M_tauH{0.0}
    , _M_z_spline{_N_zgrid}
    , _M_v_spline{_N_vgrid}
    {}
    
    
    /*   make trial emission, azimuthal angle isn't selected   */
    template<ProcessTag _Proc>
    auto make_trial(random_engine_type&, shower_history_stage<_Proc>&, energy_sqr<double>, energy_sqr<double>);

    
  private:
    /*   binding the functions to the hard state   */
    template<ProcessTag _Proc>
    bool bind(const shower_history_stage<_Proc>&, energy_sqr<double>, energy_sqr<double>);
    
    /*    calculates the full splitting function    */
    double splitfunc(energy_sqr<double>, double);
    double splitfunc_sum(energy_sqr<double>, double);

    /*     generating the thau variable   */
    auto generate_tau(double tauL, double lrho) const
    {
      /*  trivial check   */
      if(tauL >= _M_tauH) return std::pair{false, 1.0e+120};
      
      /*   the splitting might lead to tau > tauH, thus it is not possible    */
      lrho += _M_v_spline.integral(tauL);
      
      if(_M_v_spline.integral() < lrho)
        return std::pair{false, 1.0e+120};
      
      return std::pair{true, _M_v_spline.integral_inverse(lrho)};
    }
    

    /*   hit or miss ratio   */
    double hit_or_miss_ratio(energy_sqr<double> v2, double rz)
    {
      /*    initializing the z-spline    */
      double zint = _M_z_spline.assign(unitweight{}, [=, this](double w) -> double {
        return this -> splitfunc_sum(v2, w);
      });
      
      /*    generating the variable for the z integral from a random number    */
      double u = _M_z_spline.integral_inverse(rz*zint);
     
      /*   returning the ration of the exact and the appoximated functions    */
      return splitfunc(v2, u)/_M_z_spline(u);
    }
    
    
  private:
    /*   model   */
    model_ref _M_mdl;
    
    /*   transverse momentum cutoff   */
    energy_sqr<double> _M_m2T;
    
    /*   v limits   */
    energy_sqr<double> _M_v2min, _M_v2max;
 
    /*   v limits   */
    double _M_tauL, _M_tauH;
 
    /*   details of the last call   */
    lastcall _M_last_call;
 
    
  private:
    /*   spline of the approximation     */
    spline<spline_type::constant> _M_z_spline;
 
    /*   spline for the tau grid   */
    spline<spline_type::linear> _M_v_spline;
  };
    
  
  template<OrderingTag _Ordering> template<ProcessTag _Proc>
  bool splitting_prob<fin, _Ordering>::
  bind(const shower_history_stage<_Proc>& hstage, energy_sqr<double> v2L, energy_sqr<double> v2H)
  {
    /*   nice aliases    */
    const auto& as = _M_mdl.alpha_qcd();
    const auto& p = hstage.state;
    auto[lo,hi] = bounds_of<colored>(p);


    /*   make a fresh start   */
    _M_last_call.clear();
    _M_v2min = 1.0e+120_GeV2;
    _M_v2max = 0.0_GeV2;

    
    /*   scale of the previous stage  */
    auto Q = momentum_sum<ini>(p);
    energy_sqr<double> Q2 = mag2(Q);
    energy_sqr<double> k2min = 1.0e+120_GeV2, k2max = 0.0_GeV2;
    
    /*    calculating the hard parameters for each legs    */
    for(int emitter = 1; emitter <= hi; ++emitter)
    {
      state& current = _M_last_call.states.emplace_back();
      hard_params<fin>& pars = current.pars;
      const auto& pl = p[emitter].momentum;
            
      /*   basic parameters   */
      pars.emitter = emitter;
      pars.fl = p[emitter].flavor;
      
      pars.Q2 = Q2;
      pars.pQ = 2.0*(pl*Q);
      
      pars.mT = _M_m2T/pars.pQ;
      pars.al = Q2/pars.pQ;
      pars.ymax = 1.0/sqr(std::sqrt(pars.al)+std::sqrt(std::abs(pars.al-1.0)));
      
      /*   veto parameters   */
      pars.mTmax = hstage.vetos.max_k2T/pars.pQ;
      pars.mYmax = hstage.vetos.max_q2/pars.pQ;

      
      /*   the allowed region in the v variable   */
      auto[v2min, v2max] = ordering_traits<fin, _Ordering>::main::v2limits(pars);
      v2min *=1.005; v2max *=0.995;    // <<== half a percent adjusment to avoid some numerical error
      
      if(k2min > v2min) k2min = v2min;
      if(k2max < v2max) k2max = v2max;
     
      if(v2max > v2H) v2max = v2H;
      if(v2min < v2L) v2min = v2L;

      if(v2min < v2max)
      {
        /*    adjusting the possible limits in v   */
        if(_M_v2min > v2min) _M_v2min = v2min;
        if(_M_v2max < v2max) _M_v2max = v2max;
        
        /*    possible spectators   */
        pars.spectators.clear();
        
        for(int spectator = lo; spectator <= hi; ++spectator)
          if(double cfac = duct::chi_cfac(spectator, emitter, p); cfac > 0.1)
            pars.spectators.push_back({spectator,
              (spectator != emitter ? pars.al*(p[spectator].momentum*pl)/(p[spectator].momentum*Q) : 0.0),
              cfac});
        
        /*   caching the vmin and vmax variables    */
        current.v2min = v2min;
        current.v2max = v2max;
      }
      else _M_last_call.states.pop_back();
    }
 
    /*   the splitting is not possible    */
    if(_M_last_call.states.empty()) return false;
    
    
    /*     setting up the v-spline     */
    double kinL = as.tauk(k2max), kinH = as.tauk(k2min);
    
    _M_tauL = as.tauk(_M_v2max);
    _M_tauH = as.tauk(_M_v2min);
    
    /*    adjustiong the size of the grid and making the grid     */
    _M_v_spline.makegrid(std::lround(_N_vgrid*(_M_tauH-_M_tauL)/(kinH-kinL)), _M_tauL, _M_tauH);
    
    /*    initializing the spline     */
    double vint = _M_v_spline.assign(unitweight{}, [&as, this](double tau) -> double {
      return _M_z_spline.assign(unitweight{}, [v = as.q2(tau), this](double w) -> double {
        return this -> splitfunc_sum(v, w);
      });
    });
    
    return (vint > 1.0e-8 ? true : false);
  }


  template<OrderingTag _Ordering>
  double splitting_prob<fin, _Ordering>::splitfunc(energy_sqr<double> v2, double u)
  {
    /*   trival case  */
    if(v2 < _M_v2min || _M_v2max < v2) return 0.0;
    
    if(_M_last_call.v2 == v2 && _M_last_call.u == u)
      return _M_last_call.weight_tot;
    
    
    /*   binding z mappings to v   */
    const auto& as = _M_mdl.alpha_qcd();

    if(_M_last_call.v2 != v2)
    {
      _M_last_call.v2 = v2;
      _M_last_call.as = as.ask(as.tauk(v2));
      
      for(auto& sf : _M_last_call.states)
        if(sf.is_possible(v2))
          sf.zmap.bind(sf.pars, v2, _M_last_call.as);
    }
    
    
    /*    calculating the function   */
    _M_last_call.u = u;
    _M_last_call.weight_tot = 0.0;
    
    for(auto& sf : _M_last_call.states)
    {
      /*   clear the weights   */
      sf.weights.clear();
      
      if(sf.is_possible(v2) && sf.zmap.is_possible())
      {
        auto[z, zjac] = sf.zmap.w2z(u);
        auto[y, yjac] = ordering_traits<fin, _Ordering>::main::mapping_to_y(sf.pars, v2, z);
        
        sf.z = z;
        sf.y = y;
  
        /*    we have to apply the vetos   */
        if(y < sf.pars.mYmax && z*(1.0-z)*y < sf.pars.mTmax)
        {
          sf.weights = splitting_functions<fin>::splitfunc(sf.pars, y, z);
          
          /*   alpha_s ratio   */
          double as_ratio = as.ask(as.tauk((1.0-z)*y*sf.pars.pQ))/_M_last_call.as;
          
          /*    adjusting the weights   */
          for(auto& w : sf.weights)
            _M_last_call.weight_tot += (w.weight *= zjac*yjac*as_ratio);
        }
      }
    }
    
    return _M_last_call.weight_tot;
  }
    
    
  template<OrderingTag _Ordering>
  double splitting_prob<fin, _Ordering>::splitfunc_sum(energy_sqr<double> v2, double u)
  {
    /*   trival case  */
    if(v2 < _M_v2min || _M_v2max < v2) return 0.0;
    
    if(_M_last_call.v2 == v2 && _M_last_call.u == u)
      return _M_last_call.weight_tot;
    
    
    /*   binding z mappings to v   */
    const auto& as = _M_mdl.alpha_qcd();

    if(_M_last_call.v2 != v2)
    {
      _M_last_call.v2 = v2;
      _M_last_call.as = as.ask(as.tauk(v2));
      
      for(auto& sf : _M_last_call.states)
        if(sf.is_possible(v2))
          sf.zmap.bind(sf.pars, v2, _M_last_call.as);
    }
    
    
    /*    calculating the function   */
    _M_last_call.u = u;
    _M_last_call.weight_tot = 0.0;
    
    for(auto& sf : _M_last_call.states)
      if(sf.is_possible(v2) && sf.zmap.is_possible())
      {
        auto[z, zjac] = sf.zmap.w2z(u);
        auto[y, yjac] = ordering_traits<fin, _Ordering>::main::mapping_to_y(sf.pars, v2, z);
        
        double as_ratio = as.ask(as.tauk((1.0-z)*y*sf.pars.pQ))/_M_last_call.as;
        double weight = splitting_functions<fin>::splitfunc_sum(sf.pars, y, z);
        
        _M_last_call.weight_tot += zjac*yjac*as_ratio*weight;
      }
    
    
    return _M_last_call.weight_tot;
  }
 
  
  template<OrderingTag _Ordering> template<ProcessTag _Proc>
  auto splitting_prob<fin, _Ordering>::
  make_trial(random_engine_type& rne, shower_history_stage<_Proc>& stage, energy_sqr<double> v2L, energy_sqr<double> v2H)
  {
    /*   aliases   */
    const auto& as = _M_mdl.alpha_qcd();

    
    /*   binding to the hard state   */
    if(!bind(stage, v2L, v2H)) return std::pair{false, 0.0_GeV2};
      
    
    /*   generating tau variable according to the Sudakov factor    */
    double tauT = _M_tauL;
    energy_sqr<double> v2 = 0.0_GeV2;
    std::uniform_real_distribution<double> rng(0.0,1.0);
    
    do {
      /*   generate tau  */
      auto[is_ok, tau] = generate_tau(tauT, -std::log(rng(rne)));
      
      if(false == is_ok) return std::pair{false, 0.0_GeV2};
      tauT = tau;
    } while(rng(rne) > hit_or_miss_ratio(v2 = as.q2(tauT), rng(rne)));
    
   
    /*   the splitting is possible, select a channel (no azimuthal angle)  */
    double oprob = 0.0, rho = rng(rne)*_M_last_call.weight_tot;
    
    for(auto& sf : _M_last_call.states)
      if(false == sf.weights.empty())
        for(auto& w : sf.weights)
          if((oprob += w.weight) > rho)
          {
            stage.vars.emitter = w.emitter;
            stage.vars.spectator = w.spectator;
            stage.vars.y = sf.y;
            stage.vars.z = sf.z;
            
            /*    when the gluon splits to q+q~, we have to choose a flavour  */
            if(81 == w.fl)
            {
              std::uniform_int_distribution<> fdist(1, 5);
            
              stage.vars.fl = fdist(rne);
              stage.vars.fm = -stage.vars.fl;
            }
            else
            {
              stage.vars.fl = w.fl;
              stage.vars.fm = w.fm;
            }
            
            /*   color representation of the dougters   */
            stage.vars.cl = _M_mdl.color(stage.vars.fl);
            stage.vars.cm = _M_mdl.color(stage.vars.fm);

            
            return std::pair{true, v2};
          }
    
    return std::pair{false, 0.0_GeV2};
  }
  

}  //  namespace duct

#endif
