/*  Copyright © 2021 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 07.01.21 at 10:02.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_splitting_prob_ini_h__
#define __mc_splitting_prob_ini_h__ 1


/*   std includes   */


/*   project headers   */
#include <bits/mc-ordering.h>
#include <bits/mc-splitting-func.h>
#include <bits/mc-shower-stage.h>
#include <cxx/math-support.h>



namespace duct {
  
  
  template<class _Ordering>
  class splitting_prob<ini, _Ordering>
  {
    /*   size of the interpolation grid    */
    static constexpr unsigned int _N_zgrid = 8u;
    static constexpr unsigned int _N_vgrid = 20u;

    /*   possible splitting state  */
    struct splitting_state
    {
      /*    check if the splitting possible with a given v   */
      bool is_possible() const {
        return (v2min < v2max);
      }
      
      bool is_possible(energy_sqr<double> v2) const {
        return (v2min < v2 && v2 < v2max);
      }
      
      /*    hard parameters   */
      hard_params<ini> pars;
      
      /*   available range in variable v   */
      energy_sqr<double> v2min, v2max, v2T;
      
      /*   splitting function object  */
      stackvector<splitting_functions<ini>::result, 6u> weights;
      
      /*   z mapping   */
      __details::z_mapping<ini, _Ordering> zmap;
    
      /*   hardness and momentum fraction variables   */
      double y, z;
    };
      
    
    /*   storing the details of the last call   */
    struct lastcall
    {
      /*   make a fresh start   */
      void clear()
      {
        state.weights.clear();
        weight_tot = as = 0.0;
        v2 = -1.0_GeV2;
        u = -1.0;
      }
      
      /*    list of all the possible states    */
      splitting_state state;
      
      /*    total weight of the states   */
      double weight_tot = 0.0;
      
      /*   the v and u variables   */
      energy_sqr<double> v2 = -1.0_GeV2;
      double u = -1.0;
      
      /*   strong coupling at the shower scale   */
      double as = 0.0;
    };
    
    
  public:
    /*    public types    */
    using ordering_type = _Ordering;
    
    
    /*   default constructor   */
    splitting_prob()
    : _M_mdl{get_default_model()}, _M_m2T{1.0_GeV2}
    , _M_tauL{1.0e+120}, _M_tauH{0.0}
    , _M_z_spline{_N_zgrid}
    , _M_v_spline{_N_vgrid}
    {}

    /*   with custom model resource    */
    splitting_prob(model_ref mdl, energy_sqr<double> m2T)
    : _M_mdl{mdl}, _M_m2T{m2T}
    , _M_tauL{1.0e+120}, _M_tauH{0.0}
    , _M_z_spline{_N_zgrid}
    , _M_v_spline{_N_vgrid}
    {}
    
    
    /*   make trial emission, azimuthal angle isn't selected   */
    template<ProcessTag _Proc>
    auto make_trial(random_engine_type&, int, shower_history_stage<_Proc>&, energy_sqr<double>, energy_sqr<double>);

    
  private:
    /*   binding the functions to the hard state   */
    template<ProcessTag _Proc>
    bool bind(int, const shower_history_stage<_Proc>&, energy_sqr<double>, energy_sqr<double>);
    
    /*    calculates the full splitting function    */
    double splitfunc(energy_sqr<double>, double);
    double splitfunc_sum(energy_sqr<double>, double);

    /*     generating the thau variable   */
    auto generate_tau(double tauL, double lrho) const
    {
      /*  trivial check   */
      if(tauL >= _M_tauH) return std::pair{false, 1.0e+120};
      
      /*   the v spline is weighted   */
      logweight weight{_M_tauT};
      
      /*   the splitting might lead to tau > tauH, thus it is not possible    */
      lrho += _M_v_spline.integral(weight, tauL);
      
      if(_M_v_spline.integral(weight, _M_tauH) < lrho)
        return std::pair{false, 1.0e+120};
      
      return std::pair{true, _M_v_spline.integral_inverse(weight, lrho)};
    }
    

    /*   hit or miss ratio   */
    double hit_or_miss_ratio(energy_sqr<double> v2, double rz)
    {
      /*    initializing the z-spline    */
      double zint = _M_z_spline.assign(unitweight{}, [=, this](double w) -> double {
        return this -> splitfunc_sum(v2, w);
      });
      
      /*    generating the variable for the z integral from a random number    */
      double u = _M_z_spline.integral_inverse(rz*zint);
     
      /*   returning the ration of the exact and the appoximated functions    */
      return splitfunc(v2, u)/_M_z_spline(u);
    }
    
    
  private:
    /*   model   */
    model_ref _M_mdl;
    const shower_aux *_M_aux;
    
    /*   transverse momentum cutoff   */
    energy_sqr<double> _M_m2T;
    
    /*   v limits   */
    energy_sqr<double> _M_v2min, _M_v2max;
 
    /*   v limits   */
    double _M_tauL, _M_tauH, _M_tauT;
 
    /*   details of the last call   */
    lastcall _M_last_call;
 
    
  private:
    /*   spline of the approximation     */
    spline<spline_type::constant> _M_z_spline;
 
    /*   spline for the tau grid   */
    spline<spline_type::linear> _M_v_spline;
  };
  
  
  template<OrderingTag _Ordering> template<ProcessTag _Proc>
  bool splitting_prob<ini, _Ordering>::
  bind(int emitter, const shower_history_stage<_Proc>& hard, energy_sqr<double> v2L, energy_sqr<double> v2H)
  {
    /*   nice aliases    */
    const auto& as = _M_mdl.alpha_qcd();
    const auto& p = hard.state;
    auto[lo,hi] = bounds_of<colored>(p);
    
    /*     access to the pdf tables   */
    _M_aux = _M_mdl.aux(hadron{p.hadron[emitter].key});

    
    /*   make a fresh start   */
    _M_last_call.clear();
    
    
    /*   scale of the previous stage  */
    auto Q = momentum_sum<ini>(p);
    energy_sqr<double> Q2 = mag2(Q);
    
    /*    calculating the hard parameters for each legs    */
    const auto& pa = p[emitter].momentum;
    auto& current = _M_last_call.state;
    auto& pars = current.pars;
    
    /*   basic parameters   */
    pars.emitter = emitter;
    pars.fl = p[emitter].flavor;
    
    pars.Q2 = Q2;
    pars.eta = eta(emitter, p);
    pars.mT = std::max(_M_aux->thrsT(pars.fl), _M_m2T)/Q2;
    
    /*   veto parameters   */
    pars.mTmax = hard.vetos.max_k2T/Q2;
    pars.mYmax = hard.vetos.max_q2/Q2;

      
    /*   the allowed region in the v variable   */
    auto[k2min, k2max] = ordering_traits<ini, _Ordering>::main::v2limits(pars);

    current.v2T = _M_aux->thrsT(pars.fl, pars.eta);
    current.v2min = std::max({k2min, 1.005*v2L, current.v2T});
    current.v2max = std::min(k2max, 0.995*v2H);


    if(current.v2min < current.v2max)
    {
      /*    possible spectators   */
      pars.spectators.clear();
      
      for(int spectator = lo; spectator <= hi; ++spectator)
        if(double cfac = duct::chi_cfac(spectator, emitter, p); cfac > 0.1)
          pars.spectators.push_back({spectator,
            (spectator != emitter ? (p[spectator].momentum*pa)/(p[spectator].momentum*Q) : 0.0),
            cfac});
    }
    
    /*   the splitting is not possible    */
    else return false;
        
    
    /*     setting up the v-spline     */
    double kinL = as.tauk(k2max), kinH = as.tauk(k2min);
    
    _M_tauL = as.tauk(current.v2max);
    _M_tauH = as.tauk(current.v2min);
    _M_tauT = as.tauk(current.v2T);

    
    /*    adjustiong the size of the grid and making the grid     */
    _M_v_spline.makegrid(std::lround(_N_vgrid*(_M_tauH-_M_tauL)/(kinH-kinL)), _M_tauL, _M_tauH);
    
    /*    initializing the spline     */
    double vint = _M_v_spline.assign(logweight{_M_tauT}, [&as, this](double tau) -> double {
      return _M_z_spline.assign(unitweight{}, [v2 = as.q2(tau), this](double w) -> double {
        return this -> splitfunc_sum(v2, w);
      });
    });
    
    return (vint > 1.0e-8 ? true : false);
  }


  
  template<OrderingTag _Ordering>
  double splitting_prob<ini, _Ordering>::splitfunc(energy_sqr<double> v2, double u)
  {
    /*   trival case  */
    if(!_M_last_call.state.is_possible(v2)) return 0.0;
    
    if(_M_last_call.v2 == v2 && _M_last_call.u == u)
      return _M_last_call.weight_tot;
    
    
    /*   binding z mappings to v   */
    const auto& as = _M_mdl.alpha_qcd();
    auto& sf = _M_last_call.state;
 
    
    if(_M_last_call.v2 != v2)
    {
      _M_last_call.v2 = v2;
      _M_last_call.as = as.ask(as.tauk(v2));
      sf.zmap.bind(sf.pars, v2, _M_last_call.as);
    }
    
    
    /*    calculating the function   */
    _M_last_call.u = u;
    _M_last_call.weight_tot = 0.0;
    
    /*   clear the weights   */
    sf.weights.clear();
      
    if(sf.is_possible(v2) && sf.zmap.is_possible())
    {
      auto[z, zjac] = sf.zmap.w2z(u);
      auto[y, yjac] = ordering_traits<ini, _Ordering>::main::mapping_to_y(sf.pars, v2, z);
      
      sf.z = z;
      sf.y = y;
      
      /*    we have to apply the vetos   */
      if(y < sf.pars.mYmax && (1.0-z)*y < sf.pars.mTmax)
      {
        sf.weights = splitting_functions<ini>::splitfunc(sf.pars, y, z);
        
        /*   alpha_s ratio   */
        double as_ratio = as.ask(as.tauk((1.0-z)*y*sf.pars.Q2))/_M_last_call.as;
       
        /*   PDF ratios    */
        double fden = _M_aux->parton(-sf.pars.fl, sf.pars.eta, v2);
        double pdf_ratio1 = _M_aux->parton(-sf.pars.fl, sf.pars.eta/z, v2)/fden;
        double pdf_ratio2 = _M_aux->parton((0 == sf.pars.fl ? 85 : 0), sf.pars.eta/z, v2)/fden;

        /*    adjusting the weights   */
        for(auto& w : sf.weights)
        {
          /*    as, pdf ratios and jacobians   */
          w.weight *= zjac*yjac*as_ratio;
          w.weight *= (sf.pars.fl == w.fl ? pdf_ratio1 : pdf_ratio2);
          
          /*    summing up the weights     */
          _M_last_call.weight_tot += w.weight;
        }
      }
    }
  
    
    return _M_last_call.weight_tot;
  }
    
 
  template<OrderingTag _Ordering>
  double splitting_prob<ini, _Ordering>::splitfunc_sum(energy_sqr<double> v2, double u)
  {
    /*   trival case  */
    if(!_M_last_call.state.is_possible(v2)) return 0.0;
    
    if(_M_last_call.v2 == v2 && _M_last_call.u == u)
      return _M_last_call.weight_tot;
    
    
    /*   binding z mappings to v   */
    const auto& as = _M_mdl.alpha_qcd();
    auto& sf = _M_last_call.state;
 
    
    if(_M_last_call.v2 != v2)
    {
      _M_last_call.v2 = v2;
      _M_last_call.as = as.ask(as.tauk(v2));
      sf.zmap.bind(sf.pars, v2, _M_last_call.as);
    }
    
    
    /*    calculating the function   */
    _M_last_call.u = u;
    _M_last_call.weight_tot = 0.0;
    
    if(sf.is_possible(v2) && sf.zmap.is_possible())
    {
      auto[z, zjac] = sf.zmap.w2z(u);
      auto[y, yjac] = ordering_traits<ini, _Ordering>::main::mapping_to_y(sf.pars, v2, z);
      auto[w1, w2] = splitting_functions<ini>::splitfunc_sum(sf.pars, y, z);
        
      /*   alpha_s ratio   */
      double as_ratio = as.ask(as.tauk((1.0-z)*y*sf.pars.Q2))/_M_last_call.as;
      
      /*   PDF ratios    */
      double fden = _M_aux->parton(-sf.pars.fl, sf.pars.eta, v2);
      double pdf_ratio1 = _M_aux->parton(-sf.pars.fl, sf.pars.eta/z, v2)/fden;
      double pdf_ratio2 = _M_aux->parton(0 == sf.pars.fl ? 85 : 0, sf.pars.eta/z, v2)/fden;
      
      _M_last_call.weight_tot = zjac*yjac*as_ratio*(pdf_ratio1*w1 + pdf_ratio2*w2);
    }
  
    return _M_last_call.weight_tot;
  }
    

  template<OrderingTag _Ordering>
  template<ProcessTag _Proc>
  auto splitting_prob<ini, _Ordering>::
  make_trial(random_engine_type& rne, int emitter, shower_history_stage<_Proc>& stage,
             energy_sqr<double> v2L, energy_sqr<double> v2H)
  {
    /*   aliases   */
    const auto& as = _M_mdl.alpha_qcd();
    const auto& sf = _M_last_call.state;
    
    
    /*   binding to the hard state   */
    if(!bind(emitter, stage, v2L, v2H)) return std::pair{false, 0.0_GeV2};
      
    
    /*   generating tau variable according to the Sudakov factor    */
    double tauT = _M_tauL;
    energy_sqr<double> v2 = 0.0_GeV2;
    std::uniform_real_distribution<double> rng(0.0,1.0);
    
    do {
      /*   generate tau  */
      auto[is_ok, tau] = generate_tau(tauT, -std::log(rng(rne)));
      
      if(false == is_ok) return std::pair{false, 0.0_GeV2};
      tauT = tau;
    } while(rng(rne) > hit_or_miss_ratio(v2 = as.q2(tauT), rng(rne)));
    
   
    /*   the splitting is possible, select a channel (no azimuthal angle)  */
    double oprob = 0.0, rho = rng(rne)*_M_last_call.weight_tot;
    
    for(const auto& w : sf.weights)
      if((oprob += w.weight) > rho)
      {
        stage.vars.emitter = w.emitter;
        stage.vars.spectator = w.spectator;
        stage.vars.fl = w.fl;
        stage.vars.fm = w.fm;
        stage.vars.y = sf.y;
        stage.vars.z = sf.z;
        
        /*    when the gluon goes to quark we have to choose a quark flavor  */
        if(85 == w.fl)
        {
          double fq[11];
          
          for(int i = -5; i <= 5; ++i)
          {
            if(0 != i)
              fq[5+i] = _M_aux->parton(-i, sf.pars.eta/sf.z, v2);
            else fq[5+i] = 0.0;
          }
          
          std::discrete_distribution<int> fdist{fq, fq+11};
          
          stage.vars.fl = fdist(rne);
          stage.vars.fm = -stage.vars.fl;
        }
      
        /*   color representation of the dougters   */
        stage.vars.cl = _M_mdl.color(stage.vars.fl);
        stage.vars.cm = _M_mdl.color(stage.vars.fm);

        
        return std::pair{true, v2};
      }
    
    return std::pair{false, 0.0_GeV2};
  }
  
  
}  //  namespace duct

#endif
