/*  Copyright (C) 2014 Zoltán Nagy
 *  Created by Zoltán Nagy on 2/5/14.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_splitting_prob_h__
#define __mc_splitting_prob_h__ 1


/*   std includes   */


/*   project headers   */
#include <bits/qcd-color.h>
#include <bits/mc-shower-stage.h>
#include <bits/mc-color-evolution.h>



namespace duct {
  namespace __details {
    
    /*    simple class to handle z mappings    */
    template<IniFinTag _Type, OrderingTag _Ordering>
    struct z_mapping
    {
      /*    binding to the hard parameters and to the v variable    */
      void bind(const hard_params<fin>& pars, energy_sqr<double> v2, double a)
      {
        /*    set the z boundaries   */
        auto[zmin, zmax] = ordering_traits<_Type, _Ordering>::main::zlimits(pars, v2);
        double Lp;
        
        /*   for transverse momentum ordering as = 0  =>  simpler mapping  */
        if constexpr(_Ordering::alpha == ordering::scheme::kT)
        {
          as = 0.0;
          Lm = 1.0-zmin;
          Lp = 1.0-zmax;
          zjac = (zmin < zmax ? std::log(Lm/Lp) : 0.0);
        }
        
        /*   more complicated mapping for lambda and angular ordering  */
        else
        {
          as = a*qcd::beta1(5u)*static_cast<unsigned>(_Ordering::alpha);
          Lm = 1.0 + as*std::log(1.0-zmin);
          Lp = 1.0 + as*std::log(1.0-zmax);
          
          zjac = (zmin < zmax ? std::log(Lm/Lp) : 0.0);
        }
      }
      
      /*    do the mapping    */
      std::pair<double, double> w2z(double w) const
      {
        double L = Lm*std::exp(-w*zjac);
        
        /*   for kT ordering L = 1-z   */
        if constexpr(_Ordering::alpha == ordering::scheme::kT)
          return {1.0-L, L*zjac};
        
        /*   for lambda and angular ordering  L = 1 + as*log(1-z)   */
        else {
          double z = 1.0-std::exp((L-1.0)/as);
          return {z, (1.0-z)*L*zjac/as};
        }
      }
      
      /*   is possible   */
      bool is_possible() const { return (zjac > 1.0e-8);}
      
      double as, zjac, Lm;
    };
    
  }   //   namespace __details
  
  
  /*    empty declaration of the splitting_prob class   */
  template<class, OrderingTag>
  class splitting_prob;
  
} //  namespace duct


/*   partial specializations   */
#include <bits/mc-splitting-prob-fin.h>
#include <bits/mc-splitting-prob-ini.h>



namespace duct {
  namespace __details {
    
    /*    gets the azimuthal angle in a final state splitting     */
    template<ProcessTag _Proc>
    double azimuthal_fin(random_engine_type& rne, const shower_history_stage<_Proc>& stage)
    {
      /*   collinear splitting   */
      std::uniform_real_distribution<double> rng(0.0,1.0);
      
      if(stage.vars.spectator == stage.vars.emitter)
        return 2.0*std::numbers::pi*rng(rne);
      
      
      /*    soft splitting    */
      auto Q = momentum_sum<ini>(stage.state);
      const auto& pl = stage.state[stage.vars.emitter].momentum;
      double al = 0.5*mag2(Q)/(pl*Q);
      
      /*   soft : simple case   */
      if(std::abs(1.0-al) < 1.0e-6 && stage.vars.spectator > 0)
        return 2.0*std::numbers::pi*rng(rne);
      
      
      /*   soft : general case    */
      const auto& pk = stage.state[stage.vars.spectator].momentum;

      return splitting_functions<fin>::azimuthal_angle(stage.vars.spectator, al, al*(pk*pl)/(pk*Q),
                                                       stage.vars.y, stage.vars.z, rng(rne));
    }
    
    
    /*    gets the azimuthal angle in a final state splitting     */
    template<ProcessTag _Proc>
    double azimuthal_ini(random_engine_type& rne, const shower_history_stage<_Proc>& stage)
    {
      /*   collinear splitting   */
      std::uniform_real_distribution<double> rng(0.0,1.0);
      
      if(stage.vars.spectator == stage.vars.emitter)
        return 2.0*std::numbers::pi*rng(rne);
      
      
      /*    soft splitting    */
      auto Q = momentum_sum<ini>(stage.state);
      const auto& pa = stage.state[stage.vars.emitter].momentum;
      const auto& pk = stage.state[stage.vars.spectator].momentum;
      double xi = (pk*pa)/(pk*Q);
      
  
      /*   soft : simple case   */
      if(std::abs(1.0-xi) < 1.0e-6)
        return 2.0*std::numbers::pi*rng(rne);
      
      
      /*   soft : general case    */
      return splitting_functions<ini>::azimuthal_angle(xi, stage.vars.y, stage.vars.z, rng(rne));
    }
    
  }   //   namespace __details
  
  
  
  
  /*   This class implements the splitting probabilities for various processes.
   */
  template<ProcessTag _Proc, OrderingTag _Ordering>
  class splitting_prob<_Proc, _Ordering>
  {
  public:
    /*   public types  */
    using process_type = _Proc;
    using ordering_type = _Ordering;
    
    /*   default constructor   */
    splitting_prob() = default;
    
    /*   with custom model resource    */
    splitting_prob(model_ref mdl, energy_sqr<double> m2T)
    : _M_fin{mdl, m2T} {}
    
    
    /*    make the trial emission    */
    bool make_trial(random_engine_type&, shower_history_stage<_Proc>&, energy_sqr<double>, energy_sqr<double>);
    
  private:
    /*   splitting kernels   */
    splitting_prob<fin, _Ordering> _M_fin;
    select_lcplus_operator<_Proc> _M_lcplus;
  };
  
  
  template<ProcessTag _Proc, OrderingTag _Ordering>
  inline bool splitting_prob<_Proc, _Ordering>::
  make_trial(random_engine_type& rne, shower_history_stage<_Proc>& stage, energy_sqr<double> v2L, energy_sqr<double> v2H)
  {
    auto[is_possible, v2T] = _M_fin.make_trial(rne, stage, v2L, v2H);
    
    /*   not possible   */
    if(false == is_possible) {
      stage.next_scale = v2L;
      return false;
    }
    
    /*   we have splitting and we have to fix the azimuthal angle   */
    stage.next_scale = v2T;
    stage.vars.phi = __details::azimuthal_fin(rne, stage);
    
    /*    select the LC+ color operator   */
    _M_lcplus(rne, stage);
    
    
    return true;
  }
  
 
  
  
  /*
   *     Spliting probability for hadron collisions
   */
  template<ProcessTag _Proc, OrderingTag _Ordering>
  requires HadronCollision<_Proc>
  class splitting_prob<_Proc, _Ordering>
  {
  public:
    /*   public types  */
    using process_type = _Proc;
    using ordering_type = _Ordering;
    
    /*   default constructor   */
    splitting_prob() = default;
    
    /*   with custom model resource    */
    splitting_prob(model_ref mdl, energy_sqr<double> m2T)
    : _M_fin{mdl, m2T}, _M_ini{mdl, m2T} {}
    
    
    /*    make the trial emission    */
    bool make_trial(random_engine_type&, shower_history_stage<_Proc>&, energy_sqr<double>, energy_sqr<double>);
    
  private:
    /*   splitting kernels   */
    splitting_prob<fin, _Ordering> _M_fin;
    splitting_prob<ini, _Ordering> _M_ini;
    select_lcplus_operator<_Proc> _M_lcplus;
  };
  
  
  template<ProcessTag _Proc, OrderingTag _Ordering>
  requires HadronCollision<_Proc>
  inline bool splitting_prob<_Proc, _Ordering>::
  make_trial(random_engine_type& rne, shower_history_stage<_Proc>& stage, energy_sqr<double> v2L, energy_sqr<double> v2H)
  {
    bool is_possible = false;
    
    /*   try to split the incoming parton labeled by 0    */
    if(auto[is_ok, v2T] = _M_ini.make_trial(0, rne, stage, v2L, v2H); is_ok)
    {
      v2L = v2T;
      is_possible = true;
    }
    
    /*   try to split the incoming parton labeled by -1 when it is hhc process   */
    if constexpr(std::same_as<_Proc, hhc>)
      if(auto[is_ok, v2T] = _M_ini.make_trial(-1, rne, stage, v2L, v2H); is_ok)
      {
        v2L = v2T;
        is_possible = true;
      }
    
    /*   trying the final state splittings     */
    if(auto[is_ok, v2T] = _M_fin.make_trial(rne, stage, v2L, v2H); is_ok)
    {
      v2L = v2T;
      is_possible = true;
    }
    
    
    /*   not possible   */
    if(false == is_possible) {
      stage.next_scale = v2L;
      return false;
    }
    
    
    /*   we have splitting and we have to fix the azimuthal angle   */
    stage.next_scale = v2L;
    stage.vars.phi = (stage.vars.emitter > 0
                      ? __details::azimuthal_fin(rne, stage)
                      : __details::azimuthal_ini(rne, stage));
    
    
    /*    select the LC+ color operator   */
    _M_lcplus(rne, stage);
    
    
    return true;
  }
  

}    //    namespace duct

#endif
