/*  Copyright © 2021 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 02.01.21 at 11:04.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_statistical_basis_h__
#define __mc_statistical_basis_h__ 1


/*   std includes   */
#include <algorithm>
#include <iostream>
#include <ranges>
#include <concepts>

/*   project headers   */
#include <bits/duct-tags.h>
#include <bits/duct-lorentzvector.h>
#include <cxx/memory-resource.h>
#include <cxx/binary-io.h>



namespace duct {
  
  /*    empty extention   */
  struct no_statistical_basis_extention{};
  
  
  /*    base class for the statistical_basis classes, we used to have statistical_basis    */
  template<ProcessTag _Proc, class _Parton, class _Ext = no_statistical_basis_extention>
  class statistical_basis : public _Ext
  {
    /*   some private types   */
    using _Diff = std::ptrdiff_t;
    using _Size = std::size_t;
    
    /*   number of incoming partons   */
    static constexpr _Size _n_in = number_of_incomings<_Proc>;
    static constexpr _Size _n_qcd_in = number_of_qcd_in<_Proc>;
    static constexpr _Size _n_reserve = 50u;
    
  public:
    /*   public types   */
    using parton = _Parton;
    using reference = parton&;
    using pointer = parton *;
    using iterator = pointer;
    
    using const_reference = const parton&;
    using const_pointer = const parton *;
    using const_iterator = const_pointer;
    
    using size_type = _Size;
    using difference_type = _Diff;
    using allocator_type = pmr::polymorphic_allocator<parton>;
    
    
    /*    default constructors reserves space for 50 partons   */
    statistical_basis() : statistical_basis{allocator_type{}} {}
    
    explicit statistical_basis(allocator_type a)
    : _M_memory{a}
    {
      _M_create_storage_for(_n_reserve);
    }
    
    /*   destructor   */
    ~statistical_basis() { _M_deallocate();}
    
    
    /*   non-trivial constructor   */
    explicit statistical_basis(_Size _n_qcd, allocator_type a = {})
    : _M_memory{a}
    {
      _M_create_storage_for(_n_in+_n_qcd);
      _M_lay_down_for(0, _n_qcd);
    }
    
    statistical_basis(_Size _n_other, _Size _n_qcd, allocator_type a = {})
    : _M_memory{a}
    {
      _M_create_storage_for(_n_other+_n_in+_n_qcd);
      _M_lay_down_for(_n_other, _n_qcd);
    }
    
    
    /*   copy    */
    statistical_basis& operator=(const statistical_basis&);
    
    statistical_basis(const statistical_basis& x, allocator_type a = {})
    : _Ext{x}, _M_memory{a}
    {
      if(x.begin()) {
        _M_create_storage_for(x.size());
        _M_lay_down_for(x.number_of<uncolored,fin>(), x.number_of<colored,fin>());
        std::copy(x.begin(), x.end(), _M_begin);
      }
      else _M_begin = _M_in = _M_qin = _M_out = _M_end = _M_end_of_storage = nullptr;
    }
    
    
    /*   move    */
    statistical_basis& operator=(statistical_basis&&);
    statistical_basis(statistical_basis&&, allocator_type);
    
    statistical_basis(statistical_basis&& x)
    : _Ext{std::move(x)}
    , _M_memory{x._M_memory}
    , _M_begin{x._M_begin}
    , _M_in{x._M_in}
    , _M_qin{x._M_qin}
    , _M_out{x._M_out}
    , _M_end{x._M_end}
    , _M_end_of_storage{x._M_end_of_storage}
    {
      x._M_begin = x._M_in = x._M_qin = x._M_out = x._M_end = x._M_end_of_storage = nullptr;
    }
    
    
    /*   element access  */
    reference operator[](_Diff k) {
      return _M_out[k-1];
    }
    
    const_reference operator[](_Diff k) const {
      return _M_out[k-1];
    }
 
 
    /*   iterators   */
    iterator begin() { return _M_begin;}
    const_iterator begin() const { return _M_begin;}
    const_iterator cbegin() const { return _M_begin;}
    
    iterator end() { return _M_end;}
    const_iterator end() const { return _M_end;}
    const_iterator cend() const { return _M_end;}
    
    
    /*   direct access to the data   */
    pointer data() { return _M_begin;}
    const_pointer data() const { return _M_begin;}
    const_pointer cdata() const { return _M_begin;}
    
    
    /*   ranges   */
    template<ColoredTag _Range, IniFinTag _Type = inifin>
    auto range()
    {
      if constexpr(std::is_same_v<_Type, ini>)
      {
        if constexpr(std::is_same_v<_Range, colored>) return std::ranges::subrange<iterator>{_M_qin, _M_out};
        if constexpr(std::is_same_v<_Range, uncolored>) return std::ranges::subrange<iterator>{_M_in, _M_qin};
      }
      else if constexpr(std::is_same_v<_Type, fin>)
      {
        if constexpr(std::is_same_v<_Range, colored>) return std::ranges::subrange<iterator>{_M_out, _M_end};
        if constexpr(std::is_same_v<_Range, uncolored>) return std::ranges::subrange<iterator>{_M_begin, _M_in};
      }
      else
      {
        if constexpr(std::is_same_v<_Range, colored>) return std::ranges::subrange<iterator>{_M_qin, _M_end};
        if constexpr(std::is_same_v<_Range, uncolored>) return std::ranges::subrange<iterator>{_M_begin, _M_qin};
      }
    }

    template<ColoredTag _Range, IniFinTag _Type = inifin>
    auto crange() const
    {
      if constexpr(std::is_same_v<_Type, ini>)
      {
        if constexpr(std::is_same_v<_Range, colored>) return std::ranges::subrange<const_iterator>{_M_qin, _M_out};
        if constexpr(std::is_same_v<_Range, uncolored>) return std::ranges::subrange<const_iterator>{_M_in, _M_qin};
      }
      else if constexpr(std::is_same_v<_Type, fin>)
      {
        if constexpr(std::is_same_v<_Range, colored>) return std::ranges::subrange<const_iterator>{_M_out, _M_end};
        if constexpr(std::is_same_v<_Range, uncolored>) return std::ranges::subrange<const_iterator>{_M_begin, _M_in};
      }
      else
      {
        if constexpr(std::is_same_v<_Range, colored>) return std::ranges::subrange<const_iterator>{_M_qin, _M_end};
        if constexpr(std::is_same_v<_Range, uncolored>) return std::ranges::subrange<const_iterator>{_M_begin, _M_qin};
      }
    }

    template<ColoredTag _Range, IniFinTag _Type = inifin>
    auto range() const {
      return crange<_Range, _Type>();
    }
    
    
    /*    observers   */
    bool is_empty() const { return (_M_begin == _M_end);}
    
    _Size size() const { return static_cast<_Size>(_M_end-_M_begin);}
    _Size capacity() const { return static_cast<_Size>(_M_end_of_storage-_M_begin);}
    _Size capacity_colored() const { return static_cast<_Size>(_M_end_of_storage-_M_out);}
    
    
    /*   number of different types    */
    template<ColoredTag _Range, IniFinTag _Type = inifin>
    _Size number_of() const
    {
      if constexpr(std::is_same_v<_Type, ini>)
      {
        if constexpr(std::is_same_v<_Range, colored>) return static_cast<_Size>(_M_out-_M_qin);
        if constexpr(std::is_same_v<_Range, uncolored>) return static_cast<_Size>(_M_qin-_M_in);
      }
      else if constexpr(std::is_same_v<_Type, fin>)
      {
        if constexpr(std::is_same_v<_Range, colored>) return static_cast<_Size>(_M_end-_M_out);
        if constexpr(std::is_same_v<_Range, uncolored>) return static_cast<_Size>(_M_in-_M_begin);
      }
      else
      {
        if constexpr(std::is_same_v<_Range, colored>) return static_cast<_Size>(_M_end-_M_qin);
        if constexpr(std::is_same_v<_Range, uncolored>) return static_cast<_Size>(_M_qin-_M_begin);
      }
    }
    
    template<IniFinTag _Type>
    _Size number_of() const {
      return number_of<uncolored, _Type>() + number_of<colored, _Type>();
    }
    
    
    /*    indices of the ranges   */
    std::pair<_Diff,_Diff> bounds() const {
      return {_M_begin-_M_out+1, _M_end-_M_out};
    }
    
    template<ColoredTag _Range, IniFinTag _Type = inifin>
    std::pair<_Diff,_Diff> bounds_of() const
    {
      auto offset = _M_out-1;
      
      if constexpr(std::is_same_v<_Type, ini>)
      {
        if constexpr(std::is_same_v<_Range, colored>) return {_M_qin-offset, _M_out-_M_out};
        if constexpr(std::is_same_v<_Range, uncolored>) return {_M_in-offset, _M_qin-_M_out};
      }
      else if constexpr(std::is_same_v<_Type, fin>)
      {
        if constexpr(std::is_same_v<_Range, colored>) return {_M_out-offset, _M_end-_M_out};
        if constexpr(std::is_same_v<_Range, uncolored>) return {_M_begin-offset, _M_in-_M_out};
      }
      else
      {
        if constexpr(std::is_same_v<_Range, colored>) return {_M_qin-offset, _M_end-_M_out};
        if constexpr(std::is_same_v<_Range, uncolored>) return {_M_begin-offset, _M_qin-_M_out};
      }
    }

    
    /*   resreving storage    */
    void reserve(_Size _n_qcd)
    {
      if(!_M_begin) _M_create_storage_for(_n_in + _n_qcd);
      else if(capacity_colored() < _n_qcd) _M_rebase_reserve(_n_qcd);
    }
    
    
    /*    resizing the array    */
    void resize(_Size _n_qcd)
    {
      /*   invalid statistical_basiss we just recreate the array   */
      if(!_M_begin) {
        _M_create_storage_for(_n_in + _n_qcd);
        _M_lay_down_for(0, _n_qcd);
        return;
      }
      
      /*   otherwise   */
      if(capacity_colored() < _n_qcd) {
        _M_rebase_resize(number_of<uncolored,fin>(), _n_qcd);
        return;
      }
      
      _M_end = _M_out + _n_qcd;
    }
    
    void resize(_Size _n_other, _Size _n_qcd)
    {
      /*   invalid statistical_basiss we just recreate the array   */
      if(!_M_begin) {
        _M_create_storage_for(_n_other + _n_in + _n_qcd);
        _M_lay_down_for(_n_other, _n_qcd);
        return;
      }
      
      /*   otherwise   */
      if(capacity() < _n_other + _n_in + _n_qcd) {
        _M_rebase_resize(_n_other, _n_qcd);
        return;
      }
      
      _M_lay_down_for(_n_other, _n_qcd);
    }
    
    void resize_by(_Diff k) {
      resize(static_cast<_Size>(std::max<_Diff>(0, (_M_end-_M_out)+k)));
    }
    
    /*   shrink the array   */
    void shrink_to_fit() {
      if(_M_begin && _M_end != _M_end_of_storage)
        _M_rebase_reserve(number_of<colored,fin>());
    }
    
    /*   remove the last parton   */
    void remove_last() {
      if(_M_end) --_M_end;
    }
    
  private:
    /*   memory resource of this object    */
    allocator_type _M_memory;
    
    /*   pointers to mark the statistical_basis layout in the memory   */
    pointer _M_begin, _M_in, _M_qin, _M_out, _M_end, _M_end_of_storage;
    
    
  private:
    /*   create storage    */
    void _M_create_storage_for(_Size _count)
    {
      auto count = std::max(_n_reserve, _count);
      _M_in = _M_qin = _M_out = _M_end = _M_begin = _M_memory.allocate(count);
      _M_end_of_storage = _M_begin + count;
    }
    
    /*   allocate memory    */
    void _M_lay_down_for(_Size _n_other, _Size _n_qcd)
    {
      _M_in = _M_begin + _n_other;
      _M_out = _M_in + _n_in;
      _M_qin = _M_out - _n_qcd_in;
      _M_end = _M_out + _n_qcd;
    }
    
    /*   deallocate memory    */
    void _M_deallocate()
    {
      if(_M_begin)
        _M_memory.deallocate(_M_begin, capacity());
    }
    
    /*    rebase    */
    void _M_rebase_resize(_Size _n_other, _Size _n_qcd)
    {
      auto count = std::max(2*size(), _n_other + _n_in + _n_qcd);
      auto ptr = _M_memory.allocate(count);
      
      std::copy(_M_begin, _M_end, ptr);
      _M_deallocate();
      
      _M_begin = ptr;
      _M_end_of_storage = ptr + count;
      _M_lay_down_for(_n_other, _n_qcd);
    }
    
    void _M_rebase_reserve(_Size nqcd)
    {
      auto _n_qcd = number_of<colored,fin>();
      auto _n_other = number_of<uncolored,fin>();
      auto count = _n_other + _n_in + nqcd;
      auto ptr = _M_memory.allocate(count);
      
      std::copy(_M_begin, _M_end, ptr);
      _M_deallocate();
      
      _M_begin = ptr;
      _M_end_of_storage = ptr + count;
      _M_lay_down_for(_n_other, _n_qcd);
    }
  };
  
  
  /*     copy assignment   */
  template<ProcessTag _Proc, class _Parton, class _Ext>
  statistical_basis<_Proc, _Parton, _Ext>&
  statistical_basis<_Proc, _Parton, _Ext>::operator=(const statistical_basis<_Proc, _Parton, _Ext>& x)
  {
    if(&x != this && x.begin())
    {
      if(auto count = x.size(); capacity() < count) {
        _M_deallocate();
        _M_create_storage_for(count);
      }
      
      _M_lay_down_for(x.number_of<uncolored,fin>(), x.number_of<colored,fin>());
      std::copy(x.begin(), x.end(), _M_begin);
      _Ext::operator=(x);
    }
    
    return *this;
  }
  
  
  /*    move constructor    */
  template<ProcessTag _Proc, class _Parton, class _Ext>
  statistical_basis<_Proc, _Parton, _Ext>::
  statistical_basis(statistical_basis<_Proc, _Parton, _Ext>&& x, allocator_type a)
  : _Ext{std::move(x)}, _M_memory{a}
  {
    _M_begin = _M_in = _M_qin = _M_out = _M_end = _M_end_of_storage = nullptr;
    
    if(x.begin())
    {
      if(_M_memory == x._M_memory)
      {
        std::swap(_M_begin ,x._M_begin);
        std::swap(_M_in, x._M_in);
        std::swap(_M_qin, x._M_qin);
        std::swap(_M_out, x._M_out);
        std::swap(_M_end, x._M_end);
        std::swap(_M_end_of_storage, x._M_end_of_storage);
      }
      else {
        _M_create_storage_for(x.size());
        _M_lay_down_for(x.number_of<uncolored,fin>(), x.number_of<colored,fin>());
        std::copy(x.begin(), x.end(), _M_begin);
      }
    }
  }
  
    
  /*     move assignment   */
  template<ProcessTag _Proc, class _Parton, class _Ext>
  statistical_basis<_Proc, _Parton, _Ext>&
  statistical_basis<_Proc, _Parton, _Ext>::operator=(statistical_basis<_Proc, _Parton, _Ext>&& x)
  {
    if(&x != this && x.begin())
    {
      if(_M_memory == x._M_memory)
      {
        std::swap(_M_begin ,x._M_begin);
        std::swap(_M_in, x._M_in);
        std::swap(_M_qin, x._M_qin);
        std::swap(_M_out, x._M_out);
        std::swap(_M_end, x._M_end);
        std::swap(_M_end_of_storage, x._M_end_of_storage);
      }
      else {
        if(auto count = x.size(); capacity() < count) {
          _M_deallocate();
          _M_create_storage_for(count);
        }
        
        _M_lay_down_for(x.number_of<uncolored,fin>(), x.number_of<colored,fin>());
        std::copy(x.begin(), x.end(), _M_begin);
      }
      
      _Ext::operator=(std::move(x));
    }
    
    return *this;
  }
  
  
  
  /*       ranges   */
  template<ColoredTag _Range, IniFinTag _Type = inifin, ProcessTag _Proc, class _Parton, class _Ext>
  inline auto range(statistical_basis<_Proc, _Parton, _Ext>& p) {
    return p.template range<_Range, _Type>();
  }
  
  
  template<ColoredTag _Range, IniFinTag _Type = inifin, ProcessTag _Proc, class _Parton, class _Ext>
  inline auto crange(const statistical_basis<_Proc, _Parton, _Ext>& p) {
    return p.template crange<_Range, _Type>();
  }
  
  template<ColoredTag _Range, IniFinTag _Type = inifin, ProcessTag _Proc, class _Parton, class _Ext>
  inline auto range(const statistical_basis<_Proc, _Parton, _Ext>& p) {
    return p.template crange<_Range, _Type>();
  }
  
  
  /*   number of different kind of partons    */
  template<ColoredTag _Range, IniFinTag _Type = inifin, ProcessTag _Proc, class _Parton, class _Ext>
  inline auto number_of(const statistical_basis<_Proc, _Parton, _Ext>& p) {
    return p.template number_of<_Range, _Type>();
  }
  
  template<IniFinTag _Type, ProcessTag _Proc, class _Parton, class _Ext>
  inline auto number_of(const statistical_basis<_Proc, _Parton, _Ext>& p) {
    return p.template number_of<uncolored, _Type>() + p.template number_of<colored, _Type>();
  }
  

  /*   indices of the ranges [lower,upper]   */
  template<ColoredTag _Range, IniFinTag _Type = inifin, ProcessTag _Proc, class _Parton, class _Ext>
  inline auto bounds_of(const statistical_basis<_Proc, _Parton, _Ext>& p) {
    return p.template bounds_of<_Range, _Type>();
  }
  


  /*   binary I/O     */
  template<ProcessTag _Proc, class _Parton, class _Ext>
  inline std::ostream& write(std::ostream& os, const statistical_basis<_Proc, _Parton, _Ext>& a)
  {
    duct::write(os, number_of<uncolored,fin>(a), number_of<colored,fin>(a));
    for(auto& pi : a) duct::write(os, pi);
    return os;
  }
  
  template<ProcessTag _Proc, class _Parton, class _Ext>
  inline std::istream& write(std::istream& is, statistical_basis<_Proc, _Parton, _Ext>& a)
  {
    std::size_t nother, nqcd;
    
    duct::read(is, nother, nqcd);
    a.resize(nother, nqcd);
    
    for(auto ip = a.begin(); ip < a.end(); ++ip)
      duct::read(is, *ip);
    
    return is;
  }
  
  
  /*    print the statistical_basis   */
  template<ProcessTag _Proc, class _Parton, class _Ext>
  std::ostream& operator<<(std::ostream& os, const statistical_basis<_Proc, _Parton, _Ext>& a)
  {
    auto[idx,_] = a.bounds();
    
    os<<"+---- colored partons  -----\n";
    for(auto& pi : range<uncolored>(a))
      os<<"|  "<<idx++<<" : "<<pi<<'\n';
    
    os<<"+---- uncolored partons -----\n";
    for(auto& pi : range<colored>(a))
      os<<"|  "<<idx++<<" : "<<pi<<'\n';
    
    os<<"+----\n";
    
    return os;
  }

}  //  namespace duct

#endif
