/*  Copyright (C) 2012-2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 22/2/12.
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */
#ifndef __qcd_color_overlap_h__
#define __qcd_color_overlap_h__ 1


/*   std includes   */
#include <list>
#include <cmath>
#include <vector>


/*   project headers   */
#include <bits/qcd-color.h>
#include <bits/mc-event.h>



namespace duct {
  
  /*   forward declaration of color overlap classes   */
  class color_overlap_su3;

  
  namespace __details {
    
    class color_graph_su3
    {
    public:
      /*    we use the __partile_state color link type   */
      using colorlink = parton_cstate::colorlink;
      
      /*    class for representing a color state      */
      struct parton
      {
        /*  Multiplicity of the parton insertion:
         *    => multi == 0 inactive removed from the list
         *    => multi > 0 the multiplicity of the parton
         *       (we can have many parallel gluon)
         */
        unsigned int multi;
        
        /* These members for gluons. We use multi-gluons and these
         * are the coefficients of the corresponding Fierz identity.
         */
        double A, B;
        
        /*    color connections    */
        colorlink color, colorP;
      };
      
      /*    default constructor    */
      color_graph_su3() : _M_number_of_in{0u}, _M_active{0u}, _M_cf_pow{0}, _M_nc_pow{0}, _M_wgt{1.0} {}
      
      /*   non standard constructor, creating from event<...>    */
      template<ProcessTag _Proc, class _Parton, class _Ext>
      requires std::derived_from<_Parton, parton_cstate>
      explicit color_graph_su3(const statistical_basis<_Proc, _Parton, _Ext>& p) {
        assign(p);
      }
      
      
      /*   assignement operator from event   */
      template<ProcessTag _Proc, class _Parton, class _Ext>
      requires std::derived_from<_Parton, parton_cstate>
      color_graph_su3& operator=(const statistical_basis<_Proc, _Parton, _Ext>& p)
      {
        assign(p);
        return *this;
      }

      
      /*   returns true if it is done and the overlap is calculated   */
      bool is_finished() const { return !_M_active;}
      
      /*   this applys all the simple rules   */
      void reduce();
      
    private:
      /*   assigment functions   */
      template<ProcessTag _Proc, class _Parton, class _Ext>
      requires std::derived_from<_Parton, parton_cstate>
      void assign(const statistical_basis<_Proc, _Parton, _Ext>&);

      
      /*   rules to simplify the graps   */
      void ruleG();
      void ruleQ();
      void ruleA();
      void ruleT();
      void ruleS();
      void ruleF(color_graph_su3&);
      
      /*   some helper   */
      parton * zero() { return _M_graph.data() + _M_number_of_in - 1;}
      parton * start() { return _M_graph.data();}
      parton * end() { return _M_graph.data() + _M_graph.size();}
      
    private:
      /*    storage for the partons   */
      std::vector<parton> _M_graph;
      
      /*   number of incoming colored objects and active partons   */
      std::size_t _M_number_of_in, _M_active;
      
      /*   powers of the Cf and Nc factors   */
      int _M_cf_pow, _M_nc_pow;
      
      /*   weight of the graph   */
      double _M_wgt;
      
      /*   friend declaration   */
      friend class duct::color_overlap_su3;
    };
    
    
    template<ProcessTag _Proc, class _Parton, class _Ext>
    requires std::derived_from<_Parton, parton_cstate>
    void color_graph_su3::assign(const statistical_basis<_Proc, _Parton, _Ext>& p)
    {
      _M_active = _M_cf_pow = _M_nc_pow = 0;
      _M_wgt = 1.0;
   
      /*    allcate memory    */
      _M_number_of_in = number_of<colored,ini>(p);
      _M_graph.resize(number_of<colored>(p));
      
      parton *q = zero();
      auto[lo,hi] = bounds_of<colored>(p);
      
      /*    copy the color connections    */
      _M_wgt = 1.0;
      _M_active = number_of<colored>(p);
      _M_cf_pow = _M_nc_pow = 0;
      
      for(int i = lo; i <= hi; ++i)
      {
        q[i].color = p[i].colorL;
        q[i].colorP = p[i].colorR;
        q[i].multi = 1U;
        
        if(q[i].color.is_gluon())
        {
          --_M_cf_pow;
          q[i].A = qcd::Nc/qcd::Na;
          q[i].B = -1.0/qcd::Na;
        }
        else if(q[i].color.is_quark())
        {
          --_M_nc_pow;
          q[i].A = q[i].B = 0.0;
        }
        else q[i].A = q[i].B = 0.0;
      }
    }

    
  }   //   namespace __details
  
  
  
  /*    callable object class for calculating SU(3) color overlap   */
  class color_overlap_su3
  {
  public:
    /*    calculate the weight of a color configuration    */
    template<ProcessTag _Proc, class _Parton, class _Ext>
    requires std::derived_from<_Parton, parton_cstate>
    double operator()(const statistical_basis<_Proc, _Parton, _Ext>& p)
    {
      /*    if the list is empty then we start a new one    */
      if(_M_graphs.empty())
      {
        _M_graphs.emplace_back(p);
        _M_begin = _M_graphs.begin();
        _M_end = _M_graphs.end();
      }
      /*    use the previous list    */
      else
      {
        _M_end = _M_begin = _M_graphs.begin();
        *(_M_end++) = p;
      }

      return _M_calculate();
    }


  private:
    /*   calculate the overlap of a given graph   */
    double _M_calculate();

    /*     we use list to store the graphs    */
    std::list<__details::color_graph_su3>::iterator _M_begin, _M_end;
    std::list<__details::color_graph_su3> _M_graphs;
  };

  
  
  /*    callable object class for calculatig the color overlap     */
  class color_overlap
  {
  public:
    /*   calculate the U(3) overlap  */
    template<ProcessTag _Proc, class _Parton, class _Ext>
    requires std::derived_from<_Parton, parton_cstate>
    double operator()(const statistical_basis<_Proc, _Parton, _Ext>& c)
    {
      if(c.is_color_u3)
        return std::pow(qcd::Nc, -color_suppression_index_u3(c));
      
      return _M_overlap_su3(c);
    }
    
    /*   converting the color state to U(3)   */
    template<ProcessTag _Proc, class _Parton, class _Ext>
    requires std::derived_from<_Parton, parton_cstate>
    void convert_to_u3(statistical_basis<_Proc, _Parton, _Ext>& c)
    {
      if(c.is_color_u3 == false)
      {
        c.color_weight *= _M_overlap_su3(c);
        c.color_weight *= std::pow(qcd::Nc, color_suppression_index_u3(c));
        c.is_color_u3 = true;
      }
    }
    
    template<ProcessTag _Proc, class _Parton, class _Ext>
    requires std::derived_from<_Parton, parton_cstate>
    void convert_to_u3(unsigned int max_color, statistical_basis<_Proc, _Parton, _Ext>& c)
    {
      if(c.is_color_u3 == false && c.color_suppression_index >= max_color)
      {
        c.color_weight *= _M_overlap_su3(c);
        c.color_weight *= std::pow(qcd::Nc, color_suppression_index_u3(c));
        c.is_color_u3 = true;
      }
    }
    
   
  private:
    /*   underlying engines to calculate the overlaps    */
    color_overlap_su3 _M_overlap_su3;
  };
  
}  //   namespace duct

#endif
