/*  Copyright © 2016 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 14/01/16.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes   */

/*   project headers   */
#include "bits/mc-coupling.h"


namespace duct {
  
  namespace __details {
    
 
    std::pair<double, unsigned> as_msbar(energy_sqr<double> q2)
    {
      static constexpr energy_sqr<double> thr[6] = {0.0_GeV2, 0.0_GeV2, 0.0_GeV2, 0.0_GeV2, 1.69_GeV2, 20.25_GeV2};
      static constexpr energy_sqr<double> lmd[6] = {0.150787_GeV2, 0.150787_GeV2, 0.150787_GeV2, 0.150787_GeV2, 0.111104_GeV2, 0.056656_GeV2};

      unsigned nf = 5;
      while(q2 < thr[nf] && nf > 3) --nf;
      
      /*    computing the NLO alpha_s   */
      double b1 = qcd::beta1(nf);
      double t = b1*log(q2/lmd[nf]);
      double ox, br = qcd::beta2(nf)/b1;
      double x = t+br, lbr = br*std::log(br)-br;
      
      do {
        ox = x;
        x -= x*(x-br*std::log(x)+lbr-t)/(x-br);
      } while(std::abs(ox-x)/ox > 1e-6);
      
      return {1.0/(x-br), nf};
    }
    
    
    double coupling_grid::_M_tauk_inv(double tt) const
    {
      /*   if it is out of range return something reasonable    */
      if(tt <= 0.0) return 0.0;
      
      /*   findig the gridpoint  */
      int lo = 0, hi = _N_grid, mi;
      
      if(tt >= _M_grid[_N_grid].tauk)
        lo = (int) _N_grid-1;
      else {
        while(lo != (mi = (lo+hi)/2)) {
          if(_M_grid[mi].tauk <= tt) lo = mi;
          else hi = mi;
        }
      }
      
      /*   interpolation the inverse function   */
      const auto& g1 = _M_grid[lo];
      double d = (_M_grid[lo+1].ask-g1.ask)/_M_dt;
      
      return lo*_M_dt + (std::sqrt(g1.ask*g1.ask+2.0*d*(tt-g1.tauk))-g1.ask)/d;
    }

    
    unsigned coupling_grid::nf(double t) const
    {
      double kd = t/_M_dtau;
      int k = static_cast<int>(kd);
      
      /*  extrapolating by freezing the basepoints  */
      if(k < 0) k = 0;
      if(k >= (int) _N_grid) k = (int) _N_grid-1;
      
      if(_M_grid[k].nf == _M_grid[k+1].nf) return _M_grid[k].nf;
      
      /*   interpolating the function  */
      double kk = kd-k;
      
      return std::lround((1.0-kk)*_M_grid[k].nf + kk*_M_grid[k+1].nf);
    }
    
    double coupling_grid::as(double t) const
    {
      double kd = t/_M_dtau;
      int k = static_cast<int>(kd);
      
      /*  extrapolating by freezing the basepoints  */
      if(k < 0) k = 0;
      if(k >= (int) _N_grid) k = (int) _N_grid-1;
      
      /*   interpolating the function  */
      double kk = kd-k;
      
      return (1.0-kk)*_M_grid[k].as + kk*_M_grid[k+1].as;
    }
    
    double coupling_grid::ask(double t) const
    {
      double kd = t/_M_dtau;
      int k = static_cast<int>(kd);
      
      /*  extrapolating by freezing the basepoints  */
      if(k < 0) k = 0;
      if(k >= (int) _N_grid) k = (int) _N_grid-1;
      
      /*   interpolating the function  */
      double kk = kd-k;
      
      return (1.0-kk)*_M_grid[k].ask + kk*_M_grid[k+1].ask;
    }
    
    energy_sqr<double> coupling_grid::q2(double t) const
    {
      double kd = t/_M_dtau;
      int k = static_cast<int>(kd);
      
      /*  extrapolating by freezing the basepoints  */
      if(k < 0) k = 0;
      if(k >= (int) _N_grid) k = (int) _N_grid-1;
      
      /*   interpolating the function  */
      double kk = kd-k;
      
      return _M_q2max/std::exp((1.0-kk)*_M_grid[k].t + kk*_M_grid[k+1].t);
    }
    

    double coupling_grid::tauk(energy_sqr<double> q2) const
    {
      double kd = std::log(_M_q2max/q2)/_M_dt;
      int k = static_cast<int>(kd);
      
      /*  extrapolating by freezing the basepoints  */
      if(k < 0) k = 0;
      if(k >= (int) _N_grid) k = (int) _N_grid-1;
      
      /*   interpolating the function  */
      double kk = kd-k;
      
      return (1.0-kk)*_M_grid[k].tauk + kk*_M_grid[k+1].tauk;
    }
    
    
    double coupling_grid::tau(energy_sqr<double> q2) const
    {
      double kd = std::log(_M_q2max/q2)/_M_dt;
      int k = static_cast<int>(kd);
      
      /*  extrapolating by freezing the basepoints  */
      if(k < 0) k = 0;
      if(k >= (int) _N_grid) k = (int) _N_grid-1;
      
      /*   interpolating the function  */
      double kk = kd-k;
      
      return (1.0-kk)*_M_grid[k].tau + kk*_M_grid[k+1].tau;
    }
    
  }   //   namespace __details
  
  
   
  /*   I/O  operations   */
  void qcd_coupling::read(std::istream& is)
  {
    char c;
    
    duct::read(is, c);
    
    /*   if it is there we read the table    */
    if('A' == c) {
      _M_as = std::make_shared<_Base>();
      _M_as -> read(is);
    }
    
    /*   otherwise reset it   */
    else _M_as.reset();
  }
  
  
  void qcd_coupling::write(std::ostream& os) const
  {
    if(_M_as) {
      duct::write(os, 'A');
      _M_as->write(os);
    } else duct::write(os, 'D');
  }
  
 
  
  
}   //  namespace duct
