/*  Copyright © 2016 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 03/03/16.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes   */
#include <set>


/*   project headers   */
#include "bits/qcd-dglap.h"


namespace duct {
  
  namespace __details {
    
    pdf_array pdf_kernel::operator*(const pdf_array& b) const
    {
      pdf_array res;
      
      /*    singlet part  */
      for(unsigned int i = 0U; i < 5U; i++)
        for(unsigned int j = 0U; j < 5U; j++)
          res[i] += _M_elem[5U*i+j]*b[j];
      
      /*    non-singlet part (valence quark) */
      res[5] = _M_elem[25]*b[5];
      res[6] = _M_elem[25]*b[6];
      res[7] = _M_elem[25]*b[7];
      
      /*    nonsinglet (+ components)   */
      res[8] = _M_elem[26]*b[8];
      
      return res;
    }
    
    pdf_kernel pdf_kernel::operator*(const pdf_kernel& b) const
    {
      pdf_kernel res{0.0};
      
      /*    singlet part  */
      for(unsigned int i = 0U; i < 5U; i++)
        for(unsigned int j = 0U; j < 5U; j++)
          for(unsigned int k = 0U; k < 5U; k++)
            res._M_elem[5U*i+j] += _M_elem[5U*i+k]*b._M_elem[5U*k+j];
      
      /*    non-singlet part  */
      res._M_elem[25] = _M_elem[25]*b._M_elem[25];
      res._M_elem[26] = _M_elem[26]*b._M_elem[26];
      
      return res;
    }
    
    void pdf_kernel::inverse()
    {
      pdf_kernel& a = *this;
      
      /*   nonsinglet part   */
      _M_elem[25] = 1.0/_M_elem[25];
      _M_elem[26] = 1.0/_M_elem[26];
      
      /*    This is the Gauss-Jordan Elemination from Numerical Recepes for the 5x5 part.   */
      int indxc[5], indxr[5], ipiv[5] = {-1, -1, -1, -1, -1};
      int icol = -1 , irow = -1;
      
      for(int i = 0; i < 5; i++)
      {
        double big = 0.0;
        
        for(int j = 0; j < 5; j++)
          if(ipiv[j] != 0)
            for(int k = 0; k < 5; k++) {
              if(ipiv[k] == -1) {
                if(std::abs(a[j][k]) >= big) {
                  big = std::abs(a[j][k]);
                  irow = j;
                  icol = k;
                }
              }
            }
        
        ++ipiv[icol];
        
        if(irow != icol)
          for(int l = 0; l < 5; l++)
            std::swap(a[irow][l], a[icol][l]);
        
        
        indxr[i] = irow;
        indxc[i] = icol;
        
        if(a[icol][icol] == 0.0)
          std::cerr<<"pdf_kernel::inverse(): Singular Matrix\n";
        
        double pivinv = 1.0/a[icol][icol];
        a[icol][icol] = 1.0;
        
        for(int l = 0; l < 5; l++) a[icol][l] *= pivinv;
        
        for(int ll = 0; ll < 5; ll++)
          if(ll != icol) {
            double dum = a[ll][icol];
            a[ll][icol] = 0.0;
            
            for(int l = 0; l < 5; l++)
              a[ll][l] -= a[icol][l]*dum;
          }
      }
      
      
      for(int l = 4; l >= 0; l--)
        if(indxr[l] != indxc[l])
          for(int k = 0; k < 5; k++)
            std::swap(a[k][indxr[l]], a[k][indxc[l]]);
    }
    
  }   //   namespace __details

  
  
  namespace qcd {
    namespace __qcd_details {
      
      /*-*-*-*-*-*-*-*-*-*-*-*- integrals of the LO splitting kernels   -*-*-*-*-*-*-*-*-*-*-*-*/
      
      static double lo1qq(double u, double v) {
        double vu = v-u, ov = 1.0-v, ou = 1.0-u;
        return Cf*(2.0*v*ou/vu*std::log(ou/ov)-u*v/vu*std::log(v/u)-0.5*v*(2.0+vu));
      }
      
      static double lo2qq(double u, double v) {
        double vu = v-u, ov = 1.0-v, ou = 1.0-u;
        if(v+1e-15 > 1.0) return Cf*(u*v/vu*std::log(v/u)+0.5*u*(2.0-vu));
        return Cf*(u*v/vu*std::log(v/u)-2.0*u*ov/vu*std::log(ou/ov)+0.5*u*(2.0-vu));
      }
      
      static double lo3qq(double u) {
        double ou = 1.0-u;
        return Cf*(u/ou*std::log(u)+2.0*std::log(ou)+0.5*u);
      }
      
      
      static double lo1qg(double u, double v) {
        double vu = v-u;
        return 0.5*(v*(1.0-vu)+v*vu*(u+2.0*v)/3.0-u*v/(v-u)*std::log(v/u));
      }
      
      static double lo2qg(double u, double v) {
        return 0.5*((u*u*(2.0*u-3.0*(1.0+v))+u*(3.0+6.0*v)+v*(v*v-3.0-3.0*v))/3.0+v*std::log(v/u))*u/(v-u);
      }
      
      
      static double lo1gq(double u, double v) {
        double vu = v-u;
        return Cf*(2.0*v*(1.0+u)/vu*std::log(v/u)-0.5*(4.0+v*(4.0-vu)));
      }
      
      static double lo2gq(double u, double v) {
        double vu = v-u;
        return Cf*(2.0*u*(1.0+v)/vu*std::log(u/v)+0.5*(4.0+u*(4.0+vu)));
      }
      
      
      static double lo1gg(double u, double v) {
        double vu = v-u, ov = 1.0-v, ou = 1.0-u;
        return 2.0*Ca*(v/vu*(ou*std::log(ou/ov)+(1.0+u)*std::log(v/u)) - 1.0 + v*(u*(u-3.0+v)+(3.0-2.0*v)*v-12.0)/6.0);
      }
      
      static double lo2gg(double u, double v) {
        double vu = v-u, ov = 1.0-v, ou = 1.0-u;
        if(v+1e-14 > 1.0) return 2.0*Ca*(u/vu*((1.0+v)*std::log(u/v)) + 1.0 + u*(12.0+ u*(2.0*u-v-3.0)-(v-3.0)*v)/6.0);
        return 2.0*Ca*(u/vu*(ov*std::log(ov/ou)+(1.0+v)*std::log(u/v)) + 1.0 + u*(12.0+ u*(2.0*u-v-3.0)-(v-3.0)*v)/6.0);
      }
      
      static double lo3gg(double u) {
        double ou = 1.0-u;
        return 2.0*Ca*(std::log(ou)-(1.0+u)/ou*std::log(u) + (u*(u-2.0)-17.0)/6.0);
      }
      
      /*-*-*-*-*-*-*-*-*-*-*-*- NLO splitting kernels   -*-*-*-*-*-*-*-*-*-*-*-*/
      constexpr double zeta2 = 1.64493406684822643647241516665;
      constexpr double zeta3 = 1.20205690315959428539973816151;
      
      inline double pqq(double x) { return (1.0+x*x)/(1.0-x);}
      inline double pqg(double x) { return 1.0 - 2.0*x*(1.0-x);}
      inline double pgq(double x) { return 2.0/x - 2.0 + x;}
      inline double pgg(double x) { return 1.0/(1.0-x) + 1.0/x - 2 + x*(1.0-x);}
      
      
      /*   Non-singlet kernels and its integrals   */
      double P1ns(int eta, unsigned nf, double x)
      {
        double H0 = std::log(x), H1 = -std::log(1.0-x);
        double H00 = 0.5*H0*H0, Hm0 = H0*std::log(1.0+x) + dilog(-x);
        
        double p1 = pqq(x)*(67.0/18.0 - zeta2 + 11.0/6.0*H0 + H00)
        + pqq(-x)*(zeta2 + 2.0*Hm0 - H00) + 14.0/3.0*(1.0-x);
        
        double p2 = pqq(x)*(5.0/9.0 + H0/3.0) + 2.0/3.0*(1.0-x);
        
        double p3 = 2.0*pqq(x)*(H0*H1 - 3.0/4.0*H0)
        - 2.0*pqq(-x)*(zeta2 + 2.0*Hm0 - H00)
        - (1.0-x)*(1 - 1.5*H0) - H0 - (1.0+x)*H00;
        
        
        /*   '+'  P1ns   */
        double res = Cf*(Ca*p1 - nf*p2 + Cf*p3);
        
        /*   '-'  P1ns   */
        if(-1 == eta)
          res += 4.0*Cf*(Cf-0.5*Ca)*(pqq(-x)*(zeta2 + 2.0*Hm0 - H00) - 2.0*(1.0-x) - (1.0+x)*H0);
        
        return res;
      }
      
      static std::pair<double, double> I1ns(int eta, unsigned nf, double x)
      {
        std::pair<double, double> p1, p2, p3, p4, res;
        
        double H0 = std::log(x), H1 = -std::log(1.0-x), Hm = std::log(1.0+x);
        double Hm0 = H0*Hm + dilog(-x), H01 = dilog(x), H001 = trilog(x);
        double H00m = -trilog(-x), L2 = trilog(1.0-x), L3 = trilog(x/(1.0+x)), x2 = x*x;
        
        /*   '+'  P1ns integrals  */
        p1.first = 2.0*H001 - 6.0*H00m - 11.0/3.0*H01 + 67.0/9.0*H1 - 2.0/3.0*Hm*Hm*Hm + 4.0*L3 - 2.0/9.0*x
        + H0*(11.0/3.0*H1 - 2.0*H01 + 2.0*(Hm*Hm-Hm0) + 7.0/6.0*x - 11.0/12.0*x2)
        + H0*H0*(H1 + Hm - 0.5*x2) - 269.0/72.0*x2 + Hm0*(x2 - 3.0 - 2.0*x) + (x2 + 2.0*(Hm-H1))*zeta2;
        
        p1.second = 2.0*H001 + 2.0*H00m - 11.0/3.0*H01 + 67.0/9.0*H1 + 2.0/3.0*Hm*Hm*Hm - 4.0*L3
        + H0*(151.0 - 36.0*(H01+Hm*Hm) + 66.0*H1 - 33.0*x)/18.0
        + H0*H0*(11.0/12.0 + H1 + Hm - x) - 59.0/9.0*x + 2.0*Hm0*(1.0 + x)
        -2.0*(H1 + Hm - x)*zeta2;
        
        
        p2.first = (40.0*H1 - 24.0*H01 + (16.0-19.0*x)*x)/36.0 + H0*(4.0*H1 - x*(2.0+x))/6.0;
        p2.second = H0*H0/6.0 + H0*(11.0/9.0 + 2.0/3.0*H1 - x/3.0) - 2.0/9.0*(3.0*H01-5.0*H1+4.0*x);
        
        
        p3.first = 12.0*H00m + 4.0/3.0*Hm*Hm*Hm + 4.0*L2 - 8.0*L3 + 5.5*x
        + Hm0*(6.0 + 4.0*x - 2.0*x2) - 0.5*H1*(5.0+8.0*H01-4.0*x-x2) + 5.0/8.0*x2
        + 0.25*H0*H0*(x2-8.0*Hm - 6.0*x) + (4.0*(H1-Hm) + 2.0*x - x2)*zeta2
        + H0*(4.0*(H1*H1-Hm*Hm+Hm0-x) + 0.25*x2 - H1*(2.0*x + x2));
        
        p3.second = H0*H0*H0/6.0 - 2.0*H001 - 4.0*H00m + H01 - 2.0*H1 - 4.0*H01*H1 - 4.0/3.0*Hm*Hm*Hm
        + 4.0*L2 + 8.0*L3 + 2.0*x*(1.0+H1) - 4.0*Hm0*(1.0 + x) + 0.5*H0*H0*(x-1 - 4.0*Hm)
        + H0*(2.0*H01-1.0 + 4.0*(H1*H1+Hm*Hm) + x - H1*(1.0+2.0*x) - 2.0*zeta2) + 4.0*(H1+Hm-0.5*x)*zeta2;
        
        
        /*   '-'  P1ns   */
        if(-1 == eta) {
          p4.first = -24.0*H00m - 8.0/3.0*Hm*Hm*Hm + 16.0*L3 - 12.0*x
          + H0*(8.0*(Hm*Hm-Hm0) + 4.0*x - 3.0*x2) + H0*H0*(4.0*Hm+2.0*x-x2)
          + 5.5*x2 + 4.0*Hm0*(x2-3.0-2.0*x) + 2.0*(4.0*Hm-2.0*x+x2)*zeta2;
          
          p4.second = -2.0/3.0*H0*H0*H0 + 8.0*Hm0*(1.0 + x)
          + H0*H0*(4.0*Hm - 2.0*(1.0 + x)) + (4.0*x-8.0*Hm)*zeta2
          + H0*(4.0*zeta2 - 8.0*(1.0 + Hm*Hm + x)) + 8.0/3.0*(3.0*H00m + Hm*Hm*Hm + 6.0*(x-L3));
          
          p3.first += p4.first;
          p3.second += p4.second;
          
          p1.first -= 0.5*p4.first;
          p1.second -= 0.5*p4.second;
        }
        
        
        res.first = Cf*(Ca*p1.first - nf*p2.first + Cf*p3.first);
        res.second = Cf*(Ca*p1.second - nf*p2.second + Cf*p3.second);
        
        return res;
      }
      
      
      /*   This is the regularized integral at x = 1. Thus
       *
       *    I1ns(eta,nf) = lim{x->1} (I1ns(eta,nf, x) + 2*Cf*Kg(nf)*log(1-x))
       */
      static std::pair<double, double> I1ns(int eta, unsigned nf)
      {
        std::pair<double, double> p1, p2, p3, p4, res;
        
        /*   '+'  P1ns integrals  */
        p1.first = -95.0/24.0 - 2.0/3.0*zeta2 + zeta3;
        p1.second = -59.0/9.0 - 11.0/3.0*zeta2;
        
        p2.first = -1.0/12.0 - 2.0/3.0*zeta2;
        p2.second = -8.0/9.0 - 2.0/3.0*zeta2;
        
        p3.first = 49.0/8.0 - 3.0*zeta2  + 2.0*zeta3;
        p3.second = 2.0 + 3.0*zeta2 + 2.0*zeta3;
        
        
        /*   '-'  P1ns   */
        if(-1 == eta) {
          p4.first = -6.5 + 6.0*zeta2 - 4.0*zeta3;
          p4.second = -4.0*(-4.0 + zeta2 + 2.0*zeta3);
          
          p3.first += p4.first;
          p3.second += p4.second;
          
          p1.first -= 0.5*p4.first;
          p1.second -= 0.5*p4.second;
        }
        
        
        res.first = Cf*(Ca*p1.first - nf*p2.first + Cf*p3.first);
        res.second = Cf*(Ca*p1.second - nf*p2.second + Cf*p3.second);
        
        return res;
      }
      
      
      /*   Kernel P1ps and its integrals
       *   NOTE: In the original paper a factor of nf is included. Here we don't have.
       */
      double P1ps(double x)
      {
        double H0 = std::log(x);
        
        double p1 = 20.0/9.0/x - 2.0 + 6.0*x - 4.0*H0
        + x*x*(8.0/3.0*H0 - 56.0/9.0)
        + (1.0+x)*(5.0*H0 - H0*H0);
        
        return Cf*p1;
      }
      
      static std::pair<double, double> I1ps(double x)
      {
        std::pair<double, double> res;
        
        double H0 = std::log(x), x2 = x*x;
        
        res.first = -5.0*x - 64.0/27.0*x2*x - 0.5*H0*H0*x*(2.0+x) + 1.5*x2
        + H0*(20.0/9.0 + 3.0*x + (8.0/9.0*x + 3.0)*x2);
        
        res.second = H0*H0*((0.5-x)-H0/3.0) + H0*(7.0*x-2.0 + 4.0/3.0*x2)
        - (20.0/(9.0*x) + 34.0/9.0*x2 + x);
        
        res.first *= Cf;
        res.second *= Cf;
        
        return res;
      }
      
      inline std::pair<double, double> I1ps()
      {
        return {-317.0/54.0*Cf, -7.0*Cf};
      }
      
      /*   Kernel P1qg and its integrals
       *   NOTE: In the original paper a factor of 2*nf is included. Here we don't have.
       */
      double P1qg(double x)
      {
        double H0 = std::log(x), H1 = -std::log(1.0-x);
        double H00 = 0.5*H0*H0, H11 = 0.5*H1*H1, Hm0 = H0*std::log(1.0+x) + dilog(-x);
        
        double p1 = 20.0/(9.0*x) - 2.0 + 25.0*x
        
        - 2.0*pqg(-x)*Hm0 - 2.0*pqg(x)*H11
        + x*x*(44.0/3.0*H0 - 218.0/9.0)
        
        + 4.0*(1.0-x)*(H00 - 2.0*H0 + x*H1)
        - 4.0*zeta2*x - 6.0*H00 + 9.0*H0;
        
        
        double p2 = 2.0*pqg(x)*(H0*H1 + H11  - zeta2)
        
        + 4.0*x*x*(H0 + H00 + 2.5)
        + 2.0*(1.0-x)*(H0 + H00 - 2.0*x*H1 + 7.25)
        - 7.5 - H00 - 0.5 * H0;
        
        return 0.5*(Ca*p1 + Cf*p2);
      }
      
      static std::pair<double, double> I1qg(double x)
      {
        std::pair<double, double> p1, p2, res;
        
        double H0 = std::log(x), H1 = -std::log(1.0-x), Hm = std::log(1.0+x);
        double Hm0 = H0*Hm + dilog(-x), H01 = dilog(x), H001 = trilog(x);
        double H00m = -trilog(-x), L2 = trilog(1.0-x), x2 = x*x;
        
        p1.first = -H0*H0*(x + x2) - 2.0/3.0*Hm0*(2.0 + 3.0*x + (2.0*x + 3.0)*x2)
        + 1.0/9.0*H0*(20.0 + 39.0*x + 48.0*x2*(x + 1.0))
        -27.0/18.0 - x*(64.0 + 94.0*x2)/9.0 - H1*H1*(-2.0/3.0 + x + x2*(2.0*x-3.0)/3.0)
        + 91.0/9.0*x2 + H1*(14.0 - 24.0*x + (42.0-32.0*x)*x2)/18.0 - 2.0*x2*zeta2;
        
        p1.second = 2.0*H0*H0*(0.25 - Hm - x - H0/6.0) - 2.0*Hm0*sqr(1.0+x)
        + H0*(H1*H1 - 2.0 + 2.0*Hm0 + 14.0*x + 25.0/3.0*x2)
        -20.0/(9.0*x) + 0.5*(4.0*L2 - 1.0 + 8.0*H00m - 2.0*H1*(3.0 + 2.0*H01 + H1))
        + (14.0 + (2.0-x)*H1*(3.0 + H1) - 160.0/9.0*x)*x + 2.0*(H1-2.0*x)*zeta2;
        
        
        p2.first = H0*H0*x*(0.5-0.5*x + 2.0/3.0*x2)
        + 22.0/27.0 + 4.0/3.0*H01 + 4.5*x + (10.0/3.0*x - 41.0/6.0)*x2
        + 2.0/3.0*H1*(1.0-x)*(1.0-2.0*x2) + H1*H1*(x-2.0/3.0 - x2*(1.0-2.0/3.0*x))
        + H0*(x*(11.0+8.0*x2) - 5.0*x2 - 4.0*H1*(2.0*(1.0-x2*x) - 3.0*x*(1.0-x)))/6.0
        + (2.0*x2 - 2.0/3.0*x*(3.0+2.0*x2))*zeta2;
        
        p2.second = H0*H0*H0/6.0 - 2.0*H001 - 2.0*H01*(1.0 - H1) + H1*H1 - 2.0*L2 - 12.5*x
        - 2.0*H1*x*(1.0-x+H1-0.5*H1*x)  + 5.0*x2 + H0*H0*(3.0/4.0 - x + x2)
        + H0*(7.0 + 2.0*H01 - H1*H1 - 2.0*x*(1.0-x) + 2.0*H1*(1.0- 2.0*x + x2) - 2.0*zeta2)
        - 2.0*(H1 - 2.0*x + x2)*zeta2;
        
        res.first = 0.5*(Ca*p1.first + Cf*p2.first);
        res.second = 0.5*(Ca*p1.second + Cf*p2.second);
        
        return res;
      }
      
      static std::pair<double, double> I1qg()
      {
        std::pair<double, double> p1, p2, res;
        
        p1.first = -161.0/18.0 + 4.0*zeta2/3.0;
        p1.second = -6.5 + 3.0*zeta3;
        
        p2.first = 49.0/27.0;
        p2.second = -7.5 - 2.0*zeta3;
        
        res.first = 0.5*(Ca*p1.first + Cf*p2.first);
        res.second = 0.5*(Ca*p1.second + Cf*p2.second);
        
        return res;
      }
      
      
      /*   Kernel P1gq and its integrals   */
      double P1gq(unsigned nf, double x)
      {
        double H0 = std::log(x), H1 = -std::log(1.0-x);
        double H00 = 0.5*H0*H0, H11 = 0.5*H1*H1, Hm0 = H0*std::log(1.0+x) + dilog(-x);
        
        double p1 = 1.0/x + 2.0*pgq(x)*(H0*H1 + H11  - 11.0/6.0*H1)
        - x*x*(8.0/3.0*H0 - 44.0/9.0) + 4.0*zeta2 - 2.0 - 7.0*H0 + 2.0*H00 - 2.0*H1*x
        + (1.0+x)*(2.0*H00 - 5.0*H0 + 37.0/9.0) - 2.0*pgq(-x)*Hm0;
        
        double p2 = 2.0/3.0*x - pgq(x)*(2.0/3.0*H1 - 10.0/9.0);
        
        double p3 = pgq(x)*(3.0*H1 - 2.0*H11)
        
        + (1.0+x)*(H00 - 3.5*(1.0 - H0))
        - 3.0*H00 + 1.0 - 1.5*H0 + 2.0*H1*x;
        
        return Cf*(Ca*p1 - nf*p2 + Cf*p3);
      }
      
      static std::pair<double, double> I1gq(unsigned nf, double x)
      {
        std::pair<double, double> p1, p2, p3, p4, res;
        
        double H0 = std::log(x), H1 = -std::log(1.0-x), Hm = std::log(1.0+x);
        double Hm0 = H0*Hm + dilog(-x), H01 = dilog(x), H001 = trilog(x);
        double H00m = -trilog(-x), L2 = trilog(1.0-x), x2 = x*x;
        
        p1.first = -4.0*H001 - 8.0*H00m - 31.0/3.0*H01 + H1*(1.5*H1- 5.5 + 4.0*H01)
        - 4.0*L2 + 533.0/18.0*x + H1*x*(25.0/3.0 - 2.0*H1)
        + x2*(77.0/36.0  + 0.5*H1*(H1 - 17.0/3.0) + 52.0/27.0*x)
        + Hm0*(3.0+4.0*x + x2) + 0.5*H0*H0*(8.0*Hm + 4.0*x + x2)
        + H0*(1.0 + 4.0*H01 - 2.0*H1*H1 - 4.0*Hm0 - 22.0*x - 3.0*x2 - 8.0/9.0*x*x2 + H1*(3.0-4.0*x+x2))
        + 4.0*(x-H1)*zeta2;
        
        p1.second = H0*H0*(2.0/3.0*H0 + 4.0*Hm-2.0 + x) - 2.0*Hm0*(2.0/x+1.0-x)
        + H0*(64.0/9.0 - 4.0*Hm0 + (2.0*H1*(H1-2.0/x+1.0+x) - 7.0*x - (25.0/3.0+4.0*H01+4.0/3.0*x2)) + 4.0*zeta2)
        -1.0/x + 4.0*H001 - 8.0*H00m + 28.0/3.0*H01 + 4.0*L2 + H1*(10.0/(3.0*x) + 7.0/3.0 - 4.0*H01 - 51.0/9.0*x)
        + 49.0/9.0*x + 28.0/9.0*x2 + H1*H1*(1.0 - 2.0/x + x) + 4.0*H1*zeta2;
        
        
        p2.first = 20.0/9.0*H0 + (13.0*x2 - 24.0*H01 - 22.0*x - 6.0*H1*(3.0-4.0*x+x2))/18.0;
        p2.second = -32.0/9.0*H0 + 2.0/9.0*(6.0*H01 + 5.0*(-2.0/x + x) - 3.0*H1*(-2.0/x + 1.0 + x));
        
        
        p3.first = H0*(H0*(0.25*x2-x) + 2.0*H1*H1 + 4.0*x + 1.5*x2)
        - 2.0*H01*(2.0*H1-3.0) + 4.0*L2 - 7.5*x + H1*H1*(2.0*x - 0.5*x2-3.0/2.0)
        - 3.0/2.0*x2 + H1*(1.0 - 3.0*x + 2.0*x2)
        + 4.0*H1*zeta2;
        
        p3.second =  H0*H0*(1.0 + 0.5*x - H0/3.0) + H0*(3.5 - 2.0*H1*H1 + 2.5*x)
        -10.0*H01 - 4.0*L2 - 3.0*x - H1*H1*(1.0 + x-2.0/x)
        + H1*(3.0 + 4.0*H01 + 3.0*x-6.0/x) - 4.0*H1*zeta2;
        
        
        res.first = Cf*(Ca*p1.first - nf*p2.first + Cf*p3.first);
        res.second = Cf*(Ca*p1.second - nf*p2.second + Cf*p3.second);
        
        return res;
      }
      
      static std::pair<double, double> I1gq(unsigned nf)
      {
        std::pair<double, double> p1, p2, p3, p4, res;
        
        p1.first = 3637.0/108.0 - 31.0/3.0*zeta2 - 10.0*zeta3;
        p1.second = 2.0/9.0*(34.0 + 51.0*zeta2 - 9.0*zeta3);
        
        p2.first = -0.5 - 4.0*zeta2/3.0;
        p2.second = 2.0/9.0*(-5.0 + 6.0*zeta2);
        
        p3.first = -9.0 + + 6.0*zeta2;
        p3.second = -3.0 - 10.0*zeta2;
        
        res.first = Cf*(Ca*p1.first - nf*p2.first + Cf*p3.first);
        res.second = Cf*(Ca*p1.second - nf*p2.second + Cf*p3.second);
        
        return res;
      }
      
      
      /*   Kernel P1gg and its integrals   */
      double P1gg(unsigned nf, double x)
      {
        double H0 = std::log(x), H1 = -std::log(1.0-x);
        double H00 = 0.5*H0*H0, Hm0 = H0*std::log(1.0+x) + dilog(-x);
        
        double p1 = 1.0-x - 10.0/9.0*pgg(x) - 13.0/9.0*(1.0/x - x*x)
        - 2.0/3.0*(1.0+x)*H0;
        
        
        double p2 = 27.0 + (1.0+x)*(11.0/3.0*H0 + 8.0*H00 - 13.5)
        
        + 2.0*pgg(-x)*(H00 - 2.0*Hm0 - zeta2)
        - 67.0/9.0*(1.0/x-x*x) - 12.0*H0 - 44.0/3.0*x*x*H0
        + 2.0*pgg(x)*(67.0/18.0 - zeta2 + H00 + 2.0*H0*H1);
        
        
        double p3 = 2.0*H0 + 2.0/3.0/x + 10.0/3.0*x*x - 12.0
        + (1.0+x)*(4.0 - 5.0*H0 - 2.0*H00);
        
        
        return Ca*(nf*p1  + Ca*p2) + Cf*nf*p3;
      }
      
      static std::pair<double, double> I1gg(unsigned nf, double x)
      {
        std::pair<double, double> p1, p2, p3, p4, res;
        
        double H0 = std::log(x), H1 = -std::log(1.0-x), Hm = std::log(1.0+x);
        double Hm0 = H0*Hm + dilog(-x), H01 = dilog(x), H001 = trilog(x);
        double H00m = -trilog(-x), L2 = trilog(1.0-x), L3 = trilog(x/(1.0+x)), x2 = x*x;
        
        p1.first = H0*(-23.0/9.0 - x*(2.0 + x)/3.0) + (-10.0*H1 - 8.0*x2 + x*(35.0 + 23.0/3.0*x2))/9.0;
        p1.second = -H0*H0/3.0 + H0*(19.0/9.0-2.0/3.0*x) + (-10.0*H1 + 23.0/x - 13.0*x + 11.5*x2)/9.0;
        
        
        p2.first = -2.0*(H001+H00m) - 22.0/3.0*H01 - 4.0*H01*H1 + 2.0/3.0*Hm*Hm*Hm
        + 4.0*(L2-L3) + 523.0/18.0*x + 8.0*H1*x - (29.0/9.0 +H1 - (4.0/9.0*H1 + 52.0/27.0)*x)*x2
        + H0*H0*(H1 + 3.0*Hm + 2.0/3.0*(3.0-x)*x2)
        + H0*(2.0*H01 + (22.0/3.0 + 4.0*H1)*H1 - 2.0*Hm*Hm - 2.0*Hm0 - 23.0*x - 8.0*H1*x)
        + H0*(2.0*H1 - 1.0/6.0 - 16.0/3.0*x - 4.0/3.0*H1*x)*x2
        + Hm0*(22.0/3.0 + 2.0*x2 + 4.0/3.0*x*(6.0+x2))
        + (2.0*H1 - 2.0*Hm + 4.0/3.0*x*(6.0+x2))*zeta2;
        
        p2.second = 2.0/3.0*(H0*H0*H0-Hm*Hm*Hm) + 6.0*H001 - 14.0*H00m - 2.0*H01 + (7.0-4.0*H01)*H1
        + 4.0*(L2+L3) - 4.0*H1/x - 85.0/18.0*x - 4.0*H1*x
        + H0*H0*(H1 + 5.0*Hm + 4.0*x - x2 - 1.0/6.0) + 14.0/3.0*x2 + H1*x2
        + Hm0*(4.0*x + 2.0*x2 - 2.0 - 4.0/x)
        + H0*(4.0*H1*H1 - 6.0*(H01+Hm0) + 181.0/18.0 + 2.0*Hm*Hm - 13.0/3.0*x
              - 25.0/3.0*x2 + H1*(2.0 - 4.0/x + 4.0*x - 2.0*x2) + 4.0*zeta2)
        + 67.0/9.0*H1  + 2.0*zeta2*(H1 + Hm + x2);
        
        
        p3.first = -0.5*H0*H0*x*(2.0+x) + H0*(2.0/3.0 - x - 2.0*x2) + 3.0*x2 + x*(10.0/9.0*x2-7);
        p3.second = -H0*H0*(H0/3.0 + 1.5 + x) - H0*(8.0 + 3.0*x) - 2.0/(3.0*x) + (7.0 + 5.0/3.0*x)*x;
        
        
        res.first = Ca*(nf*p1.first  + Ca*p2.first) + Cf*nf*p3.first;
        res.second = Ca*(nf*p1.second  + Ca*p2.second) + Cf*nf*p3.second;
        
        return res;
      }
      
      
      /*   This is the regularized integral at x = 1. Thus
       *
       *    I1gg(eta,nf) = lim{x->1} (I1gg(eta,nf, x) + 2*Ca*Kg(nf)*log(1-x))
       */
      static std::pair<double, double> I1gg(unsigned nf)
      {
        std::pair<double, double> p1, p2, p3, p4, res;
        
        p1.first = 104.0/27.0;
        p1.second = 43.0/18.0;
        
        p2.first = 1499.0/54.0 - 22.0/3.0*zeta2 - 7.0*zeta3;
        p2.second = -1.0/18.0 - zeta3;
        
        p3.first = -26.0/9.0;
        p3.second = 8.0;
        
        res.first = Ca*(nf*p1.first  + Ca*p2.first) + Cf*nf*p3.first;
        res.second = Ca*(nf*p1.second  + Ca*p2.second) + Cf*nf*p3.second;
        
        return res;
      }
      
      
      /*-*-*-*-*-*-*-*-*-*-*-*- integrals of the NLO splitting kernels   -*-*-*-*-*-*-*-*-*-*-*-*/
      
      static double nlo1ns(int eta, unsigned nf, double u, double v)
      {
        auto Iu = I1ns(eta, nf, u);
        auto Iv = I1ns(eta, nf, v);
        
        double res = Iv.first-u*Iv.second;
        res -= Iu.first-u*Iu.second;
        res *= v/(v-u);
        
        return res;
      }
      
      static double nlo1ps(double u, double v)
      {
        auto Iu = I1ps(u);
        auto Iv = v+1e-15 > 1.0 ? I1ps() : I1ps(v);
        
        double res = Iv.first-u*Iv.second;
        res -= Iu.first-u*Iu.second;
        res *= v/(v-u);
        
        return res;
      }
      
      static double nlo1qg(double u, double v)
      {
        auto Iu = I1qg(u);
        auto Iv = v+1e-15 > 1.0 ? I1qg() : I1qg(v);
        
        double res = Iv.first-u*Iv.second;
        res -= Iu.first-u*Iu.second;
        res *= v/(v-u);
        
        return res;
      }
      
      static double nlo1gq(unsigned nf, double u, double v)
      {
        auto Iu = I1gq(nf, u);
        auto Iv = v+1e-15 > 1.0 ? I1gq(nf) : I1gq(nf, v);
        
        double res = Iv.first-u*Iv.second;
        res -= Iu.first-u*Iu.second;
        res *= v/(v-u);
        
        return res;
      }
      
      
      static double nlo1gg(unsigned nf, double u, double v)
      {
        auto Iu = I1gg(nf, u);
        auto Iv = I1gg(nf, v);
        
        double res = Iv.first-u*Iv.second;
        res -= Iu.first-u*Iu.second;
        res *= v/(v-u);
        
        return res;
      }
      
      
      static double nlo2ns(int eta, unsigned nf, double u, double v)
      {
        auto Iu = I1ns(eta, nf, u);
        auto Iv = v+1e-15 > 1.0 ? I1ns(eta, nf) : I1ns(eta, nf, v);
        
        double res = v*Iv.second - Iv.first;
        res -= v*Iu.second - Iu.first;
        res *= u/(v-u);
        
        return res;
      }
      
      static double nlo2ps(double u, double v)
      {
        auto Iu = I1ps(u);
        auto Iv = v+1e-15 > 1.0 ? I1ps() : I1ps(v);
        
        double res = v*Iv.second - Iv.first;
        res -= v*Iu.second - Iu.first;
        res *= u/(v-u);
        
        return res;
      }
      
      static double nlo2qg(double u, double v)
      {
        auto Iu = I1qg(u);
        auto Iv = v+1e-15 > 1.0 ? I1qg() : I1qg(v);
        
        double res = v*Iv.second - Iv.first;
        res -= v*Iu.second - Iu.first;
        res *= u/(v-u);
        
        return res;
      }
      
      static double nlo2gq(unsigned nf, double u, double v)
      {
        auto Iu = I1gq(nf, u);
        auto Iv = v+1e-15 > 1.0 ? I1gq(nf) : I1gq(nf, v);
        
        double res = v*Iv.second - Iv.first;
        res -= v*Iu.second - Iu.first;
        res *= u/(v-u);
        
        return res;
      }
      
      
      static double nlo2gg(unsigned nf, double u, double v)
      {
        auto Iu = I1gg(nf, u);
        auto Iv = v+1e-15 > 1.0 ? I1gg(nf) : I1gg(nf, v);
        
        double res = v*Iv.second - Iv.first;
        res -= v*Iu.second - Iu.first;
        res *= u/(v-u);
        
        return res;
      }
      
      
      static double nlo3ns(int eta, unsigned nf, double u)
      {
        auto Iv = I1ns(eta, nf);
        auto Iu = I1ns(eta, nf, u);
        
        double res = Iv.first-u*Iv.second;
        res -= Iu.first-u*Iu.second;
        res /= 1.0-u;
        
        return res;
      }
      
      static double nlo3gg(unsigned nf, double u)
      {
        auto Iv = I1gg(nf);
        auto Iu = I1gg(nf, u);
        
        double res = Iv.first-u*Iv.second;
        res -= Iu.first-u*Iu.second;
        res /= 1.0-u;
        
        return res;
      }
      
      
      /*-*-*-*-*-*- Making the grids -*-*-*-*-*-*/
      class tauk_mapping
      {
      public:
        /*   constructor   */
        tauk_mapping(energy_sqr<double> q2max, const coupling& as) : Q2{q2max}
        {
          double asQ = as(q2max);
          asb = beta1(5)*asQ/(1.0-asQ*Kg(5));
        }
        
        /*   mapping functions   */
        energy_sqr<double> q2(double tau) const { return Q2*std::exp((std::exp(-beta1(5)*tau)-1.0)/asb);}
        double tauk(energy_sqr<double> q2) const { return -std::log(1.0 + asb*std::log(q2/Q2))/beta1(5);}
        
        /*   the rough alpha_s function    */
        double as(energy_sqr<double> q2) const { return 1.0/(1.0/asb+std::log(q2/Q2))/beta1(5);}
        
      private:
        energy_sqr<double> Q2;
        double asb;
      };
      
      
      void dglap_grid::_M_xt_grid()
      {
        /*  These parameters are hard wired. I don't think
         *  that the users should change them. We can set up
         *  a large grid over a big interval and that should
         *  be OK for every processes.
         */
        constexpr unsigned nt = 256u;
        constexpr double dx = 0.005;
        constexpr double xmin = 1e-8, xmax = 0.9999;
        
        /****  Don't change beyond this point!   ****/
        {
          double lx = std::log(xmin), lox = std::log(1.0-xmax);
          double dlx = -dx*lx, dlox = dx*lox;
          
          unsigned int k = 0U;
          double xk, x1 = xmin;
          std::set<double> xx({1.0});
          
          while((xk = xmin*std::exp((k++)*dlx))-x1 < dx)
            xx.insert(x1 = xk);
          
          k = 0U;
          double x2 = xmax;
          while(x2-(xk = 1.0-std::exp(lox-(k++)*dlox)) < dx)
            xx.insert(x2 = xk);
          
          unsigned int nx = 1 + (x2-x1)/dx;
          double dmx = (x2-x1)/nx;
          
          for(k = 1; k < nx; k++)
            xx.insert(dmx*k+x1);
          
          
          /*   copy the basepoint to the _M_x array   */
          _M_x.assign(xx.cbegin(), xx.cend());
        }
        
        /*  t grid   */
        {
          tauk_mapping map{_M_q2max, _M_as};
          
          double tmax = map.tauk(_M_m2T);
          double dt = tmax/nt, tc = tmax-map.tauk(_M_as.thrs(4)), tb = tmax-map.tauk(_M_as.thrs(5));
          std::set<double> tt{0.0, tmax};
          
          for(unsigned k = 1; k < nt; k++)
            tt.insert(dt*k);
          
          if(0 < tc && tc < tmax) tt.insert(tc);
          if(0 < tb && tb < tmax) tt.insert(tb);
          
          /*    copy the basepoint to the _M_t array   */
          _M_t.assign(tt.cbegin(), tt.cend());
        }
      }
      
      
      /*-*-*-*-*-*-*-*-*-*-*-*- calculating the weights  -*-*-*-*-*-*-*-*-*-*-*-*/
      
      __details::pdf_kernel dglap_grid::_M_lo_weights(unsigned i, unsigned k) const
      {
        double sp = _M_x[i]/_M_x[k+1];
        
        /*    set all elements of the kernel to zero   */
        __details::pdf_kernel res{0.0};
        
        if(i < k) {
          double s = _M_x[i]/_M_x[k], sm = _M_x[i]/_M_x[k-1];
          
          /*    gluon-gluon vertex      */
          res[0][0] = lo1gg(sp, s) + lo2gg(s, sm);
          
          /*    gluon-quark vertices    */
          res[0][1] = res[0][2] = res[0][3] = res[0][4] = lo1gq(sp, s) + lo2gq(s, sm);
          
          /*    quark-gluon vertices    */
          double sqg = lo1qg(sp, s) + lo2qg(s, sm);
          res[1][0] = 4.0*sqg;
          res[2][0] = 2.0*sqg;
          res[3][0] = 2.0*sqg;
          res[4][0] = 2.0*sqg;
          
          /*    quark-quark vertices    */
          res[1][1] = res[2][2] = res[3][3] = res[4][4] = lo1qq(sp, s) + lo2qq(s, sm);
          
          /*    non-singlet    */
          res[5][0] = res[1][1];
          res[5][1] = res[1][1];
        }
        
        /*    diagonal elements    */
        else if(i == k) {
          /*    gluon-gluon vertex      */
          res[0][0] = lo3gg(sp);
          
          /*    gluon-quark vertices    */
          res[0][1] = res[0][2] = res[0][3] = res[0][4] = lo1gq(sp, 1.0);
          
          /*    quark-gluon vertices    */
          double sqg = lo1qg(sp, 1.0);
          res[1][0] = 4.0*sqg;
          res[2][0] = 2.0*sqg;
          res[3][0] = 2.0*sqg;
          res[4][0] = 2.0*sqg;
          
          /*    quark-quark vertices    */
          res[1][1] = res[2][2] = res[3][3] = res[4][4] = lo3qq(sp);
          
          /*    non-singlet   */
          res[5][0] = res[1][1];
          res[5][1] = res[1][1];
        }
        
        /*    we shouldn't call this function with k < i    */
        else throw "qcd::dglap::_M_lo_weights(...) : We have some trouble...";
        
        return res;
      }
      
      
      __details::pdf_kernel dglap_grid::_M_nlo_weights(unsigned nf, unsigned i, unsigned k) const
      {
        double sp = _M_x[i]/_M_x[k+1];
        
        /*    set all elements of the kernel to zero   */
        __details::pdf_kernel res{0.0};
        
        if(i < k) {
          double s = _M_x[i]/_M_x[k], sm = _M_x[i]/_M_x[k-1];
          
          /*    gluon-gluon vertex   */
          res[0][0] = nlo1gg(nf, sp, s) + nlo2gg(nf, s, sm);
          
          /*    gluon-quark vertices   */
          res[0][1] = res[0][2] = res[0][3] = res[0][4] = nlo1gq(nf, sp, s) + nlo2gq(nf, s, sm);
          
          /*    quark-gluon vertices   */
          double Sqg = nlo1qg(sp, s) + nlo2qg(s, sm);
          res[1][0] = 4.0*Sqg;
          res[2][0] = 2.0*Sqg;
          res[3][0] = 2.0*Sqg;
          res[4][0] = 2.0*Sqg;
          
          /*    quark-quark vertices  (diagonal)  */
          double Splus = nlo1ns(1, nf, sp, s) + nlo2ns(1, nf, s, sm);
          double Sps = nlo1ps(sp, s) + nlo2ps(s, sm);
          
          res[1][1] = Splus + 2.0*Sps;
          res[2][2] = res[3][3] = res[4][4] = Splus + Sps;
          
          /*   quark-quark vertices (off diagonal)   */
          res[1][2] = res[1][3] = res[1][4] = 2.0*Sps;
          res[2][1] = res[2][3] = res[2][4] = Sps;
          res[3][1] = res[3][2] = res[3][4] = Sps;
          res[4][1] = res[4][2] = res[4][3] = Sps;
          
          /*    non-singlet    */
          res[5][0] = nlo1ns(-1, nf, sp, s) + nlo2ns(-1, nf, s, sm);
          res[5][1] = Splus;
        }
        
        /*    diagonal elements    */
        else if(i == k) {
          /*    gluon-gluon vertex   */
          res[0][0] = nlo3gg(nf, sp);
          
          /*    gluon-quark vertices   */
          res[0][1] = res[0][2] = res[0][3] = res[0][4] = nlo1gq(nf, sp, 1.0);
          
          /*    quark-gluon vertices   */
          double Sqg = nlo1qg(sp, 1.0);
          res[1][0] = 4.0*Sqg;
          res[2][0] = 2.0*Sqg;
          res[3][0] = 2.0*Sqg;
          res[4][0] = 2.0*Sqg;
          
          /*    quark-quark vertices  (diagonal)  */
          double Splus = nlo3ns(1, nf, sp);
          double Sps = nlo1ps(sp, 1.0);
          
          res[1][1] = Splus + 2.0*Sps;
          res[2][2] = res[3][3] = res[4][4] = Splus + Sps;
          
          /*   quark-quark vertices (off diagonal)   */
          res[1][2] = res[1][3] = res[1][4] = 2.0*Sps;
          res[2][1] = res[2][3] = res[2][4] = Sps;
          res[3][1] = res[3][2] = res[3][4] = Sps;
          res[4][1] = res[4][2] = res[4][3] = Sps;
          
          /*    non-singlet    */
          res[5][0] = nlo3ns(-1, nf, sp);
          res[5][1] = Splus;
        }
        
        /*    we shouldn't call this function with k < i    */
        else throw "qcd::dglap::_M_nlo_weights(...) : We have some trouble...";
        
        return res;
      }
      
      
      __details::pdf_kernel dglap_grid::_M_weights(unsigned k, unsigned l, unsigned it) const
      {
        /*   calculate the scale   */
        tauk_mapping map{_M_q2max, _M_as};
        
        energy_sqr<double> q2 = map.q2(_M_t.back()-_M_t[it]);
        energy_sqr<double> m2c = _M_as.thrs(4), m2b = _M_as.thrs(5);
        double asq2 =  _M_as(q2), asmap = map.as(q2);
        
        
        /*   number of active flavours   */
        unsigned nf = 3u;
        
        if(q2 > m2c) ++nf;
        if(q2 > m2b) ++nf;
        
        /*     LO contribution   */
        __details::pdf_kernel lo{_M_lo_weights(k,l)};
        
        if(q2 < m2c) lo[3][0] = 0.0;
        if(q2 < m2b) lo[4][0] = 0.0;
        if(k == l) lo[0][0] += G1g(nf);
        
        lo *= asq2/asmap;
        
        /*   this is just LO calculation    */
        if(_M_ord == order::lo)
          return lo/(1.0-asq2*Kg(nf));
        
        
        /*     NLO contribution   */
        __details::pdf_kernel nlo{_M_nlo_weights(nf, k,l)};
        
        if(q2 < m2c) nlo[3][0] = nlo[3][1] = nlo[3][2] = nlo[3][4] = 0.0;
        if(q2 < m2b) nlo[4][0] = nlo[4][1] = nlo[4][2] = nlo[4][3] = 0.0;
        
        if(k == l) {
          nlo[0][0] += G2g(nf);
          
          nlo[1][1] += G2q(nf);
          nlo[2][2] += G2q(nf);
          nlo[3][3] += G2q(nf);
          nlo[4][4] += G2q(nf);
          
          nlo[5][0] += G2q(nf);
          nlo[5][1] += G2q(nf);
        }
        
        nlo *= sqr(asq2)/asmap;
        nlo += lo;
        
        return nlo;
      }
      
      
      /*-*-*-*-*-*-*-*-*-*-*-*- do the evolution  -*-*-*-*-*-*-*-*-*-*-*-*/
      
      void dglap_grid::evolve(const std::function<void(double, double *)>& func)
      {
        /*   setting up the grid    */
        _M_xt_grid();
        
        /*    upper indices on the x,t grids   */
        unsigned nx = _M_x.size()-1;
        unsigned nt = _M_t.size();
        
        /*    at x = 1 the pdf and its derivative is zero  */
        __details::pdf_matrix dh{nx+1, nt};
        
        _M_h.resize(nx+1, nt);
        
        /*    the function values at the fitting scale   */
        double __f[11], dp, up;
        double *f = __f+5;
        
        /*   flavors in the pdf_array
         *      0 => gluon
         *      1 => singlet = u+ubar + d+dbar
         *      2 => s+sbar
         *      3 => c+cbar
         *      4 => b+bbar
         *      5 => d-dbar
         *      6 => u-ubar
         *      7 => s-sbar
         *      8 => T3 = u+ubar - (d+dbar)
         */
        for(unsigned k = 0; k <= nx; k++)
        {
          __details::pdf_array& h0 = _M_h[k][0];
          
          func(_M_x[k], f);
          
          dp = f[-1]+f[1];
          up = f[-2]+f[2];
          
          //---- singlet ----
          h0[0] = f[0];
          h0[1] = dp+up;
          h0[2] = f[-3]+f[3];
          h0[3] = f[-4]+f[4];
          h0[4] = f[-5]+f[5];
          
          //---- valence ----
          h0[5] = f[1]-f[-1];
          h0[6] = f[2]-f[-2];
          h0[7] = f[3]-f[-3];
          
          //---- T3 ----
          h0[8] = up-dp;
        }
        
        /*     the t derivatives at fitting scale    */
        for(unsigned k = 0; k <= nx; k++) {
          dh[k][0] = __details::pdf_array{0.0};
          
          for(unsigned l = k; l < nx; l++)
            dh[k][0] += _M_weights(k, l, 0)*_M_h[l][0];
        }
        
        /*    now we solve the equation iteratively   */
        static const __details::pdf_kernel one{1.0};
        
        for(unsigned int it = 1; it < nt; it++)
        {
          double dt = 0.5*(_M_t[it] - _M_t[it-1]);
          
          /*   the PDFs at x = 1 are always zero   */
          dh[nx][it] = __details::pdf_array{0.0};
          
          /*   we have to solve the equation for x < 1   */
          for(int k = nx-1; k >= 0; --k)
          {
            __details::pdf_array S{0.0};
            __details::pdf_kernel w{_M_weights(k, k, it)};
            
            for(unsigned l = k+1; l < nx; l++)
              S += _M_weights(k, l, it)*_M_h[l][it];
            
            S += w*(_M_h[k][it-1] + dt*dh[k][it-1]);
            w *= -dt;
            w += one;
            w.inverse();
            
            dh[k][it] = w*S;
            _M_h[k][it] = _M_h[k][it-1] + dt*(dh[k][it] + dh[k][it-1]);
          }
        }
      }
      
      
      /*-*-*-*-*-*-*-*-*- do some interpolation on the table after the evolution -*-*-*-*-*-*-*-*-*/
      template<typename RandomIterator>
      static inline unsigned locate(RandomIterator xb, RandomIterator xe, double x)
      {
        auto n = xe-xb;
        int ju = n, jm, jl = 0;
        
        while(ju-jl > 1) {
          jm=(ju+jl)/2;
          if(x >= xb[jm]) jl = jm;
          else ju = jm;
        }
        
        return static_cast<unsigned>(jl == n-1 ? jl-1 : jl);
      }
  
    
    
      double dglap_grid::_M_parton(int fl, unsigned ix, unsigned it) const
      {
        /*    trivial case    */
        if(std::abs(fl) == 6 || ix >= _M_x.size() || it >= _M_t.size()) return 0.0;
        
        const __details::pdf_array& h = _M_h[ix][it];
        
        /*   flavor combinations   */
        if(10 <= fl && fl < 19) return h[fl-10];
        
        /*   generic flavors   */
        switch(std::abs(fl)) {
          case 81: return h[1]+h[2]+h[3]+h[4]; break;
          case 82: return h[3]; break;
          case 83: return h[4]; break;
        }
        
        /*   usual flavors   */
        switch(fl) {
            //----- b and c quarks -----
          case -5: case 5: return 0.5*h[4]; break;
          case -4: case 4: return 0.5*h[3]; break;
            
            //----- strange quark -----
          case -3: return (h[2] - h[7])/2.0; break;
          case  3: return (h[2] + h[7])/2.0; break;
            
            //----- up quark -----
          case -2: return (h[1] + h[8] - 2.0*h[6])/4.0; break;
          case  2: return (h[1] + h[8] + 2.0*h[6])/4.0; break;
            
            //----- down quark -----
          case -1: return (h[1] - h[8] - 2.0*h[5])/4.0; break;
          case  1: return (h[1] - h[8] + 2.0*h[5])/4.0; break;
            
            //----- gluon -----
          case  0: return h[0]; break;
        }
        
        return 0.0;
      }
      
      
      double dglap_grid::parton(int fl, double x, energy_sqr<double> q2) const
      {
        /*    trivial case    */
        if(std::abs(fl) == 6 || x >= 0.999999) return 0.0;
        
        /*    finding the x, t interval   */
        tauk_mapping map{_M_q2max, _M_as};
        
        double t = _M_t.back()-map.tauk(q2);
        unsigned l = locate(_M_x.begin(), _M_x.end(), x);
        unsigned k = locate(_M_t.begin(), _M_t.end(), t);
        
  
        /*    the of the function in the corners of the interpolation square   */
        double f[] = {
          _M_parton(fl, l,   k), _M_parton(fl, l,   k+1),
          _M_parton(fl, l+1, k), _M_parton(fl, l+1, k+1)
        };
        
        /*    normalized variables    */
        double tt = (_M_t[k+1]-t)/(_M_t[k+1]-_M_t[k]);
        
        /*   When every point is well defined
         *   we intepolate the log of the function.
         */
        if(f[0] > 0.0 && f[1] > 0.0 && f[2] > 0.0 && f[3] > 0.0)
        {
          double xi = std::log(1.0/x-1.0);
          double xi1 = std::log(1.0/_M_x[l+1]-1.0);
          double xi2 = std::log(1.0/_M_x[l]-1.0);
          double xx = (xi2-xi)/(xi2-xi1);
          
          /*    interpolation in variable t    */
          double h1 = std::log(f[0])*tt + std::log(f[1])*(1.0-tt);
          double h2 = std::log(f[2])*tt + std::log(f[3])*(1.0-tt);
          
          /*    interpolation in variable x    */
          return std::exp(h1*(1.0-xx) + h2*xx);
        }
        
        /*     interpolation in variable t     */
        double h1 = f[0]*tt + f[1]*(1.0-tt);
        double h2 = f[2]*tt + f[3]*(1.0-tt);
        
        /*     interpolation in variable x     */
        double xx = (_M_x[l+1]-x)/(_M_x[l+1]-_M_x[l]);
        return h1*xx + h2*(1.0-xx);
      }
  
    }  //  namespace __qcd_details
    
 
    
    
    /*   I/O  operations   */
    void dglap::read(std::istream& is)
    {
      char c;
      
      duct::read(is, _M_anti, c);
      
      /*   if it is there we read the table    */
      if('A' == c) {
        _M_dglap = std::make_shared<__qcd_details::dglap_grid>();
        _M_dglap->read(is);
      }
      
      /*   otherwise reset it   */
      else {
        _M_dglap.reset();
        _M_anti = false;
      }
    }
    
    
    void dglap::write(std::ostream& os) const
    {
      if(_M_dglap) {
        duct::write(os, _M_anti, 'A');
        _M_dglap->write(os);
      } else duct::write(os, _M_anti, 'D');
    }
    


  }   //   namespace qcd
}   //   namespace duct
