/*  Copyright © 2015 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 31/10/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __binary_io_h__
#define __binary_io_h__ 1


/*   std includes   */
#include <iostream>
#include <cmath>
#include <limits>
#include <string>
#include <vector>
#include <array>
#include <map>
#include <utility>

/*   project headers   */


namespace duct
{
  namespace __details {
    
    template<class _Int>
    typename std::enable_if<std::is_integral<_Int>::value, void>::type
    packi(unsigned char *buf, _Int i)
    {
      int shift = sizeof(_Int)-1;
      
      do {
        *buf++ = i >> (8*(shift--));
      } while(shift > 0);
      
      *buf = i;
    }
    
    
    template<class _Int>
    typename std::enable_if<std::is_integral<_Int>::value, _Int>::type
    unpacki(const unsigned char *buf)
    {
      int shift = sizeof(_Int)-1;
      
      std::make_unsigned_t<_Int> i2 = 0;
      
      do {
        i2 |= (static_cast<_Int>(*(buf++)) <<  (8*(shift--)));
      } while(shift > 0);
      
      i2 |= static_cast<_Int>(*buf);
      
      /*   make it signed   */
      if(std::is_signed<_Int>::value)
      {
        using _Unsigned = std::make_unsigned_t<_Int>;
        
        _Int d = static_cast<_Int>(~_Unsigned(0) - i2);
        constexpr _Unsigned m = static_cast<_Unsigned>(~(_Unsigned(1u) << (8*sizeof(_Int)-1)));
        
        return (i2 <= m ? i2 : -1 - d);
      }
      
      return i2;
    }
    
    
    template<typename _Tp, bool _is_char = (sizeof(_Tp) == 1)>
    struct __packed_integral_helper;
    
    /*  non char integer   */
    template<typename _Tp>
    struct __packed_integral_helper<_Tp, false>
    {
      static std::ostream& write(std::ostream& os, const _Tp& n)
      {
        char buf[sizeof(_Tp)];
        
        packi(reinterpret_cast<unsigned char *>(buf), n);
        os.write(reinterpret_cast<const char *>(buf), sizeof(_Tp));
        
        return os;
      }
      
      static std::istream& read(std::istream& is, _Tp& n)
      {
        char buf[sizeof(_Tp)];
        
        is.read(reinterpret_cast<char *>(buf), sizeof(_Tp));
        n = unpacki<_Tp>(reinterpret_cast<const unsigned char *>(buf));
        
        return is;
      }
    };
    
    /*   char integer   */
    template<typename _Tp>
    struct __packed_integral_helper<_Tp, true>
    {
      static std::ostream& write(std::ostream& os, const _Tp& n) {
        return os.write(reinterpret_cast<const char *>(&n), sizeof(_Tp));
      }
      
      static std::istream& read(std::istream& is, _Tp& n) {
        return is.read(reinterpret_cast<char *>(&n), sizeof(_Tp));
      }
    };
    
    
    /*   empty template declaration   */
    template<typename _Tp>
    struct binary_io;
    
    
    /*   integral type specializations    */
    template<> struct binary_io<bool> : __packed_integral_helper<bool> {};
    
    template<> struct binary_io<char> : __packed_integral_helper<char> {};
    template<> struct binary_io<short> : __packed_integral_helper<short> {};
    template<> struct binary_io<int> : __packed_integral_helper<int> {};
    template<> struct binary_io<long> : __packed_integral_helper<long> {};
    template<> struct binary_io<long long> : __packed_integral_helper<long long> {};
    
    template<> struct binary_io<unsigned char> : __packed_integral_helper<unsigned char> {};
    template<> struct binary_io<unsigned short> : __packed_integral_helper<unsigned short> {};
    template<> struct binary_io<unsigned int> : __packed_integral_helper<unsigned int> {};
    template<> struct binary_io<unsigned long> : __packed_integral_helper<unsigned long> {};
    template<> struct binary_io<unsigned long long> : __packed_integral_helper<unsigned long long> {};
    
    template<> struct binary_io<wchar_t> : __packed_integral_helper<wchar_t> {};
    template<> struct binary_io<char16_t> : __packed_integral_helper<char16_t> {};
    template<> struct binary_io<char32_t> : __packed_integral_helper<char32_t> {};
    
    
    /*    floating point specialization  ONLY float and double   */
    template<>
    struct binary_io<double>
    {
      /*   convert a double int uint_fast64_t   */
      static uint_fast64_t pack(double f)
      {
        struct {
          unsigned char sign : 1;
          uint_fast16_t exponent : 11;
          uint_fast64_t mantissa : 52;
        } ieee;
        
        
        /*   decompose the number  */
        if(std::isnormal(f))
        {
          int ie;
          
          ieee.mantissa = std::ldexp(std::ldexp(std::frexp(std::abs(f), &ie),1)-1.0, 52);
          ieee.exponent = ie + (1<<10) - 2;
        }
        
        else if(std::isnan(f))
        {
          ieee.mantissa = 0x0008000000000000;
          ieee.exponent = 0x7FF;
        }
        else if(std::isinf(f))
        {
          ieee.mantissa = 0;
          ieee.exponent = 0x7FF;
        }
        else
        {
          ieee.mantissa = 0;
          ieee.exponent = 0;
        }
        
        /*   sign bit   */
        ieee.sign = std::signbit(f);
        
        return (uint_fast64_t(ieee.sign)<<63) | (uint_fast64_t(ieee.exponent)<<52) | ieee.mantissa;
      }
      
      /*   convert back to double   */
      static double unpack(uint_fast64_t id)
      {
        bool s = id >> 63;
        uint_fast64_t mantissa = id & 0x000FFFFFFFFFFFFFULL;
        uint_fast64_t exponent = (id >> 52) & 0x00000000000007FFULL;
        
        /*   zero   */
        if(0 == mantissa && 0 == exponent)
          return (s ? -0.0 : 0.0);
        
        /*   NaN    */
        else if(0x0008000000000000ULL == mantissa && 0x7FFULL == exponent)
          return (s ? -std::nan("") : std::nan(""));
        
        /*   infinity   */
        else if(0 == mantissa && 0x7FFULL == exponent)
          return (s ? -1.0 : 1.0)*std::numeric_limits<double>::infinity();
        
        /*   normal    */
        else
          return (s ? -1.0 : 1.0)*std::ldexp(1.0+std::ldexp(mantissa, -52), exponent+1 - (1<<10));
      }
      
      static std::ostream& write(std::ostream& os, double d)
      {
        char buf[sizeof(uint_fast64_t)];
        
        packi(reinterpret_cast<unsigned char *>(buf), pack(d));
        os.write(reinterpret_cast<const char *>(buf), sizeof(uint_fast64_t));
        
        return os;
      }
      
      static std::istream& read(std::istream& is, double& d)
      {
        char buf[sizeof(uint_fast64_t)];
        
        is.read(reinterpret_cast<char *>(buf), sizeof(uint_fast64_t));
        d = unpack(unpacki<uint_fast64_t>(reinterpret_cast<const unsigned char *>(buf)));
        
        return is;
      }
    };
    
    
    template<>
    struct binary_io<float>
    {
      /*   convert a float int uint_fast32_t   */
      static uint_fast32_t pack(float f)
      {
        struct {
          unsigned char sign : 1;
          uint_fast8_t exponent : 8;
          uint_fast32_t mantissa : 23;
        } ieee;
        
        
        /*   decompose the number  */
        if(std::isnormal(f))
        {
          int ie;
          
          ieee.mantissa = std::ldexp(std::ldexp(std::frexp(std::abs(f), &ie),1)-1.0, 23);
          ieee.exponent = ie + (1<<7) - 2;
        }
        
        else if(std::isnan(f))
        {
          ieee.mantissa = 0x400000;
          ieee.exponent = 0xFF;
        }
        else if(std::isinf(f))
        {
          ieee.mantissa = 0;
          ieee.exponent = 0xFF;
        }
        else
        {
          ieee.mantissa = 0;
          ieee.exponent = 0;
        }
        
        /*   sign bit   */
        ieee.sign = std::signbit(f);
        
        return (uint_fast32_t(ieee.sign)<<31) | (uint_fast32_t(ieee.exponent)<<23) | ieee.mantissa;
      }
      
      /*   convert back to float   */
      static double unpack(uint_fast32_t id)
      {
        bool s = id >> 31;
        uint_fast32_t mantissa = id & 0x0007FFFFFUL;
        uint_fast32_t exponent = (id >> 23) & 0xFFUL;
        
        /*   zero   */
        if(0 == mantissa && 0 == exponent)
          return (s ? -0.0 : 0.0);
        
        /*   NaN    */
        else if(0x400000UL == mantissa && 0xFFUL == exponent)
          return (s ? -std::nan("") : std::nan(""));
        
        /*   infinity   */
        else if(0 == mantissa && 0xFFUL == exponent)
          return (s ? -1.0 : 1.0)*std::numeric_limits<double>::infinity();
        
        /*   normal    */
        else
          return (s ? -1.0 : 1.0)*std::ldexp(1.0+std::ldexp(mantissa, -23), exponent+1 - (1<<7));
      }
      
      
      static std::ostream& write(std::ostream& os, float d)
      {
        char buf[sizeof(uint_fast32_t)];
        
        packi(reinterpret_cast<unsigned char *>(buf), pack(d));
        os.write(reinterpret_cast<const char *>(buf), sizeof(uint_fast32_t));
        
        return os;
      }
      
      static std::istream& read(std::istream& is, float& d)
      {
        char buf[sizeof(uint_fast32_t)];
        
        is.read(reinterpret_cast<char *>(buf), sizeof(uint_fast32_t));
        d = unpack(unpacki<uint_fast32_t>(reinterpret_cast<const unsigned char *>(buf)));
        
        return is;
      }
    };
  
  }   /*   namespace __details    */
  
  
  
  /*    write functions for the standard types   */
  inline std::ostream& write(std::ostream& os, bool n) { return __details::binary_io<bool>::write(os, n);}
  inline std::ostream& write(std::ostream& os, char n) { return __details::binary_io<char>::write(os, n);}
  inline std::ostream& write(std::ostream& os, short n) { return __details::binary_io<short>::write(os, n);}
  inline std::ostream& write(std::ostream& os, int n) { return __details::binary_io<int>::write(os, n);}
  inline std::ostream& write(std::ostream& os, long n) { return __details::binary_io<long>::write(os, n);}
  inline std::ostream& write(std::ostream& os, long long n) { return __details::binary_io<long long>::write(os, n);}
 
  inline std::ostream& write(std::ostream& os, unsigned char n) { return __details::binary_io<unsigned char>::write(os, n);}
  inline std::ostream& write(std::ostream& os, unsigned short n) { return __details::binary_io<unsigned short>::write(os, n);}
  inline std::ostream& write(std::ostream& os, unsigned int n) { return __details::binary_io<unsigned int>::write(os, n);}
  inline std::ostream& write(std::ostream& os, unsigned long n) { return __details::binary_io<unsigned long>::write(os, n);}
  inline std::ostream& write(std::ostream& os, unsigned long long n) { return __details::binary_io<unsigned long long>::write(os, n);}

  inline std::ostream& write(std::ostream& os, wchar_t n) { return __details::binary_io<wchar_t>::write(os, n);}
  inline std::ostream& write(std::ostream& os, char16_t n) { return __details::binary_io<char16_t>::write(os, n);}
  inline std::ostream& write(std::ostream& os, char32_t n) { return __details::binary_io<char32_t>::write(os, n);}

  inline std::ostream& write(std::ostream& os, float n) { return __details::binary_io<float>::write(os, n);}
  inline std::ostream& write(std::ostream& os, double n) { return __details::binary_io<double>::write(os, n);}
 
  inline std::ostream& write(std::ostream& os, const long double& n) {
    return  os.write(reinterpret_cast<const char *>(&n), sizeof(long double));
  }

  
  /*    read functions for the standard types   */
  inline std::istream& read(std::istream& is, bool& n) { return __details::binary_io<bool>::read(is, n);}
  inline std::istream& read(std::istream& is, char& n) { return __details::binary_io<char>::read(is, n);}
  inline std::istream& read(std::istream& is, short& n) { return __details::binary_io<short>::read(is, n);}
  inline std::istream& read(std::istream& is, int& n) { return __details::binary_io<int>::read(is, n);}
  inline std::istream& read(std::istream& is, long& n) { return __details::binary_io<long>::read(is, n);}
  inline std::istream& read(std::istream& is, long long& n) { return __details::binary_io<long long>::read(is, n);}
  
  inline std::istream& read(std::istream& is, unsigned char& n) { return __details::binary_io<unsigned char>::read(is, n);}
  inline std::istream& read(std::istream& is, unsigned short& n) { return __details::binary_io<unsigned short>::read(is, n);}
  inline std::istream& read(std::istream& is, unsigned int& n) { return __details::binary_io<unsigned int>::read(is, n);}
  inline std::istream& read(std::istream& is, unsigned long& n) { return __details::binary_io<unsigned long>::read(is, n);}
  inline std::istream& read(std::istream& is, unsigned long long& n) { return __details::binary_io<unsigned long long>::read(is, n);}
  
  inline std::istream& read(std::istream& is, wchar_t& n) { return __details::binary_io<wchar_t>::read(is, n);}
  inline std::istream& read(std::istream& is, char16_t& n) { return __details::binary_io<char16_t>::read(is, n);}
  inline std::istream& read(std::istream& is, char32_t& n) { return __details::binary_io<char32_t>::read(is, n);}
  
  inline std::istream& read(std::istream& is, float& n) { return __details::binary_io<float>::read(is, n);}
  inline std::istream& read(std::istream& is, double& n) { return __details::binary_io<double>::read(is, n);}

  inline std::istream& read(std::istream& is, long double& n) {
    return  is.read(reinterpret_cast<char *>(&n), sizeof(long double));
  }

  /*     forward declarations    */
  template<typename _Tp, std::size_t _N> std::ostream& write(std::ostream&, const std::array<_Tp, _N>&);
  template<typename _Tp, std::size_t _N> std::istream& read(std::istream&, std::array<_Tp, _N>&);

    
    
    
    
  /*    binary write/read of a C like array    */
  template<typename _Tp, std::size_t _N>
  inline std::ostream& write(std::ostream& os, const _Tp (&n)[_N]) {
    for(const _Tp& x : n) write(os, x);
    return os;
  }
  
  template<typename _Tp, std::size_t _N>
  inline std::istream& read(std::istream& is, _Tp (&n)[_N]) {
    for(_Tp& x : n) read(is, x);
    return is;
  }
  
  
  /*    binary write/read of an array    */
  template<typename _Tp, std::size_t _N>
  inline std::ostream& write(std::ostream& os, const std::array<_Tp, _N>& n)
  {
    for(const _Tp& x : n) write(os, x);
    return os;
  }
  
  template<typename _Tp, std::size_t _N>
  inline std::istream& read(std::istream& is, std::array<_Tp, _N>& n)
  {
    for(_Tp& x : n) read(is, x);
    return is;
  }
  
  
  /*    binary write/read of a string  */
  inline std::ostream& write(std::ostream& os, const std::string& ltx)
  {
    write(os, ltx.size());
    for(auto c : ltx) write(os, c);
    return os;
  }
  
  inline std::istream& read(std::istream& is, std::string& ltx)
  {
    std::string::size_type s;
    
    read(is, s);
    ltx.resize(s);
    
    for(unsigned k = 0; k < s; ++k)
      read(is, ltx[k]);
    
    return is;
  }
 
  
  /*    binary write/read of a vector  */
  template<class _Tp, class _Allocator>
  inline std::ostream& write(std::ostream& os, const std::vector<_Tp, _Allocator>& ltx)
  {
    write(os, ltx.size());
    for(const auto& c : ltx) write(os, c);
    return os;
  }
  
  template<class _Tp, class _Allocator>
  inline std::istream& read(std::istream& is, std::vector<_Tp, _Allocator>& ltx)
  {
    typename std::vector<_Tp, _Allocator>::size_type s;
    
    read(is, s);
    ltx.resize(s);
    
    for(unsigned k = 0; k < s; ++k)
      read(is, ltx[k]);
    
    return is;
  }
  
  
 
  /*    binary write/read of a pair  */
  template<class _Tp, class _Up>
  inline std::ostream& write(std::ostream& os, const std::pair<_Tp, _Up>& p)
  {
    write(os, p.first, p.second);
    return os;
  }
  
  template<class _Tp, class _Up>
  inline std::istream& read(std::istream& is, std::pair<_Tp, _Up>& p)
  {
    read(is, p.first, p.second);
    return is;
  }
  
 
  /*    binary write/read of a map  */
  template<class Key, class T, class Compare, class Allocator>
  inline std::ostream& write(std::ostream& os, const std::map<Key, T, Compare, Allocator>& ltx)
  {
    write(os, ltx.size());
    for(const auto& c : ltx) write(os, c);
    return os;
  }
  
 
  template<class Key, class T, class Compare, class Allocator>
  std::istream& read(std::istream& is, std::map<Key, T, Compare, Allocator>& ltx)
  {
    typename std::map<Key, T, Compare, Allocator>::size_type s;
    std::pair<Key, T> tmp;
    
    read(is, s);
    ltx.clear();
    
    for(unsigned k = 0; k < s; ++k) {
      read(is, tmp);
      ltx.insert(tmp);
    }
    
    return is;
  }
  

  
  
  /*   Variadic template definition for multiple read/write operations   */
  template<class _Msg, class _Args>
  std::ostream& write(std::ostream& os, const _Msg& msg, const _Args& arg)
  {
    write(os, msg);
    return write(os, arg);
  }

  template<class _Msg, class _Args>
  std::istream& read(std::istream& is, _Msg& msg, _Args& arg)
  {
    read(is, msg);
    return read(is, arg);
  }

  
  template<class _Msg, class ..._Args>
  std::ostream& write(std::ostream& os, const _Msg& msg, const _Args& ...arg)
  {
    write(os, msg);
    return write(os, arg...);
  }
  
  template<class _Msg, class ..._Args>
  std::istream& read(std::istream& is, _Msg& msg, _Args& ...arg)
  {
    read(is, msg);
    return read(is, arg...);
  }
  
  
  
  /*   binary read/write of a tuple   */
  namespace __details {
    template<class _Tuple, std::size_t N>
    struct binary_io_tuple_unpacker
    {
      /*   binary read from a stream   */
      static void write_tuple(std::ostream& os, const _Tuple& x) {
        binary_io_tuple_unpacker<_Tuple, N-1>::write_tuple(os, x);
        write(os, std::get<N-1>(x));
      }
      
      /*   binary read from a stream   */
      static void read_tuple(std::istream& is, _Tuple& x) {
        binary_io_tuple_unpacker<_Tuple, N-1>::read_tuple(is, x);
        read(is, std::get<N-1>(x));
      }
    };
    
    
    template<class _Tuple>
    struct binary_io_tuple_unpacker<_Tuple, 1>
    {
      /*   binary read from a stream   */
      static void write_tuple(std::ostream& os, const _Tuple& x) {
        write(os, std::get<0>(x));
      }
      
      /*   binary read from a stream   */
      static void read_tuple(std::istream& is, _Tuple& x) {
        read(is, std::get<0>(x));
      }
    };
    
  }   //   namespace __details
  
  
  template<class ..._Args>
  std::ostream& write(std::ostream& os, const std::tuple<_Args...>& x)
  {
    __details::binary_io_tuple_unpacker<std::tuple<_Args...>, sizeof...(_Args)>::write_tuple(os, x);
    return os;
  }
  
  template<class ..._Args>
  std::istream& read(std::istream& is, std::tuple<_Args...>& x)
  {
    __details::binary_io_tuple_unpacker<std::tuple<_Args...>, sizeof...(_Args)>::read_tuple(is, x);
    return is;
  }
  
  

  
}


#endif
