/*  Copyright (C) 2016 Zoltán Nagy
 *  Created by Zoltán Nagy on 5/2/16.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __duct_cxx_cloned_ptr_h__
#define __duct_cxx_cloned_ptr_h__ 1


/*    This code is based on Jonathan Coe's implementation. 
 *    See https://github.com/jbcoe/cloned_ptr  
 */

/*   std headers   */
#include <memory>
#include <type_traits>


namespace duct {

  namespace __details {

    template<typename _Tp>
    struct default_copier
    {
      _Tp *operator()(const _Tp& t) const { return new _Tp(t);}
    };
    
  
    
    template <typename _Tp>
    struct control
    {
      virtual ~control() = default;
      virtual std::unique_ptr<control> clone() const = 0;
      virtual _Tp *release() = 0;
      virtual _Tp *ptr() = 0;
    };
    
    
    template <typename _Tp, typename _Up, typename _Copier = default_copier<_Up>, typename _Deleter = std::default_delete<_Up>>
    struct object : public control<_Tp>
    {
      explicit object(_Up *u, _Copier c = _Copier{}, _Deleter d = _Deleter{})
      : copier{std::move(c)}, pointer{u, std::move(d)} {}
      
      std::unique_ptr<control<_Tp>> clone() const override {
        return std::make_unique<object>(copier(*pointer), copier, pointer.get_deleter());
      }
      
      _Tp * release() override { return pointer.release(); }
      _Tp * ptr() override { return pointer.get(); }
      
      _Copier copier;
      std::unique_ptr<_Up, _Deleter> pointer;
    };
  
    
    template <typename _Tp, typename _Up>
    struct delegating : public control<_Tp>
    {
      explicit delegating(std::unique_ptr<control<_Up>> b)
      : delegate{std::move(b)} {}
      
      std::unique_ptr<control<_Tp>> clone() const override {
        return std::make_unique<delegating>(delegate->clone());
      }
      
      _Tp * release() override { return delegate->release();}
      _Tp * ptr() override { return delegate->ptr();}
      
      std::unique_ptr<control<_Up>> delegate;
    };
    
    
    
    template <typename _Tp, typename _Up>
    struct downcasting : public control<_Tp>
    {
      explicit downcasting(std::unique_ptr<control<_Up>> b)
      : delegate{std::move(b)} {}
      
      std::unique_ptr<control<_Tp>> clone() const override {
        return std::make_unique<downcasting>(delegate->clone());
      }
      
      _Tp * release() override { return static_cast<_Tp *>(delegate->release());}
      _Tp * ptr() override { return static_cast<_Tp *>(delegate->ptr());}
      
      std::unique_ptr<control<_Up>> delegate;
    };
    

    template <typename _Tp, typename _Up>
    struct dyncasting : public control<_Tp>
    {
      explicit dyncasting(std::unique_ptr<control<_Up>> b)
      : delegate{std::move(b)} {}
      
      std::unique_ptr<control<_Tp>> clone() const override {
        return std::make_unique<dyncasting>(delegate->clone());
      }
      
      _Tp * release() override { return dynamic_cast<_Tp *>(delegate->release());}
      _Tp * ptr() override { return dynamic_cast<_Tp *>(delegate->ptr());}
      
      std::unique_ptr<control<_Up>> delegate;
    };
    
    
    template <typename _Tp, typename _Up>
    struct constcasting : public control<_Tp>
    {
      explicit constcasting(std::unique_ptr<control<_Up>> b)
      : delegate{std::move(b)} {}
      
      std::unique_ptr<control<_Tp>> clone() const override {
        return std::make_unique<constcasting>(delegate->clone());
      }
      
      _Tp * release() override { return const_cast<_Tp *>(delegate->release());}
      _Tp * ptr() override { return const_cast<_Tp *>(delegate->ptr());}
      
      std::unique_ptr<control<_Up>> delegate;
    };
    


    
  }   //   namespace __details
  
  
  
  template <typename _Tp>
  class cloned_ptr
  {
  public:
    /*   trivial consructors & destructor   */
    cloned_ptr() noexcept : _M_ptr{nullptr} {}
    cloned_ptr(std::nullptr_t) noexcept : _M_ptr{nullptr} {}
    ~cloned_ptr() = default;
    
    
    /*   capturing a bare ponter    */
    template<typename _Up, typename _Copier = __details::default_copier<_Up>,
    typename _Deleter = std::default_delete<_Up>, typename _X = std::enable_if_t<std::is_convertible<_Up*, _Tp*>::value>>
    explicit cloned_ptr(_Up *u, _Copier copier = _Copier{}, _Deleter deleter = _Deleter{})
    : _M_ptr{u ? std::make_unique<__details::object<_Tp, _Up, _Copier, _Deleter>>(u, std::move(copier), std::move(deleter)) : nullptr} {}
    
    
    /*    copy constructors   */
    cloned_ptr(const cloned_ptr &p) noexcept
    : _M_ptr{p._M_ptr ? p._M_ptr->clone() : nullptr} {}
    
    
    template<typename _Up, typename _X = std::enable_if_t<!std::is_same<_Tp, _Up>::value && std::is_convertible<_Up*, _Tp*>::value>>
    cloned_ptr(const cloned_ptr<_Up> &p) : _M_ptr{std::make_unique<__details::delegating<_Tp, _Up>>(std::move(cloned_ptr<_Up>{p}._M_ptr))} {}
    
    
    /*   move constructors   */
    cloned_ptr(cloned_ptr&& p) noexcept
    : _M_ptr{std::move(p._M_ptr)} {}
    
    
    template <typename _Up, typename _X = std::enable_if_t<!std::is_same<_Tp, _Up>::value && std::is_convertible<_Up*, _Tp*>::value>>
    cloned_ptr(cloned_ptr<_Up>&& p) : _M_ptr{std::make_unique<__details::delegating<_Tp, _Up>>(std::move(p._M_ptr))} {}
      
    
    
    /*    copy assignments    */
    cloned_ptr &operator=(const cloned_ptr &p) noexcept
    {
      if(&p != this) {
        if(p._M_ptr) _M_ptr = p._M_ptr->clone();
        else _M_ptr.reset();
      }
      
      return *this;
    }
    
    
    template <typename _Up, typename _X = std::enable_if_t<!std::is_same<_Tp, _Up>::value && std::is_convertible<_Up*, _Tp*>::value>>
    cloned_ptr &operator=(const cloned_ptr<_Up> &p)
    {
      if(p._M_ptr) _M_ptr = std::make_unique<__details::delegating<_Tp, _Up>>(std::move(cloned_ptr<_Up>{p}._M_ptr));
      else _M_ptr.reset();
      return *this;
    }

    
    
    /*    move assignments    */
    cloned_ptr &operator=(cloned_ptr&& p) noexcept {
      if(&p != this) _M_ptr = std::move(p._M_ptr);
      return *this;
    }
    
    
    template <typename _Up, typename _X = std::enable_if_t<!std::is_same<_Tp, _Up>::value && std::is_convertible<_Up*, _Tp*>::value>>
    cloned_ptr &operator=(cloned_ptr<_Up> &&p) {
      _M_ptr = std::make_unique<__details::delegating<_Tp, _Up>>(std::move(p._M_ptr));
      return *this;
    }

    
    /*    modifiers   */
    _Tp * release() noexcept { return (_M_ptr ? _M_ptr->release() : nullptr);}
    void swap(cloned_ptr& p) noexcept{ std::swap(_M_ptr, p._M_ptr);}
    
    
    template <typename _Up = _Tp>
    std::enable_if_t<std::is_convertible<_Up *, _Tp *>::value>
    reset(_Up *u = nullptr)
    {
      if(static_cast<_Tp *>(u) != (_M_ptr ? _M_ptr->ptr() : nullptr))
        *this = cloned_ptr<_Up>{u};
    }
    
    
    /*   observers    */
    _Tp * get() const noexcept { return _M_ptr ? _M_ptr->ptr() : nullptr;}
    _Tp *operator->() const noexcept { return _M_ptr ? _M_ptr->ptr() : nullptr;}
    _Tp& operator*() const noexcept { return *get();}
    
    /*   conversion to bool   */
    operator bool() const noexcept { return _M_ptr ? (bool) (_M_ptr->ptr()) : false;}
    
  private:
    std::unique_ptr<__details::control<_Tp>> _M_ptr;

  
    /*   Friends    */
    template<typename> friend class cloned_ptr;
    template<typename T, typename U> friend cloned_ptr<T> const_pointer_cast(const cloned_ptr<U>&);
    template<typename T, typename U> friend cloned_ptr<T> dynamic_pointer_cast(const cloned_ptr<U>&);
    template<typename T, typename U> friend cloned_ptr<T> static_pointer_cast(const cloned_ptr<U>&);
  };
  
    
  /*   cloned_ptr factory   */
  template <typename _Tp, typename... _Args>
  cloned_ptr<_Tp> make_cloned_ptr(_Args&& ...args) {
    return cloned_ptr<_Tp>(new _Tp(std::forward<_Args>(args)...));
  }

  
  /*   castings   */
  template <typename _Tp, typename _Up>
  cloned_ptr<_Tp> static_pointer_cast(const cloned_ptr<_Up>& p)
  {
    cloned_ptr<_Tp> t;
    t._M_ptr = std::make_unique<__details::downcasting<_Tp, _Up>>(std::move(cloned_ptr<_Up>{p}._M_ptr));
    return t;
  }
  

  template <typename _Tp, typename _Up>
  cloned_ptr<_Tp> dynamic_pointer_cast(const cloned_ptr<_Up>& p)
  {
    cloned_ptr<_Tp> t;
    if(dynamic_cast<_Tp *>(p.get()))
      t._M_ptr = std::make_unique<__details::dyncasting<_Tp, _Up>>(std::move(cloned_ptr<_Up>{p}._M_ptr));
    return t;
  }
    
  
  template <typename _Tp, typename _Up>
  cloned_ptr<_Tp> const_pointer_cast(const cloned_ptr<_Up>& p)
  {
    cloned_ptr<_Tp> t;
    t._M_ptr = std::make_unique<__details::constcasting<_Tp, _Up>>(std::move(cloned_ptr<_Up>{p}._M_ptr));
    return t;
  }
      
      
  /*    comparisons    */
  template <typename _Tp, typename _Up>
  bool operator==(const cloned_ptr<_Tp>& x, const cloned_ptr<_Up>& y) {
    return x.get() == y.get();
  }

  template <typename _Tp, typename _Up>
  bool operator!=(const cloned_ptr<_Tp>& x, const cloned_ptr<_Up>& y) {
    return x.get() != y.get();
  }

  template<typename _Tp, typename _Up>
  bool operator<(const cloned_ptr<_Tp>& x, const cloned_ptr<_Up>& y) {
    return std::less<std::common_type<_Tp*, _Up*>>()(x.get(), y.get());
  }

  template<typename _Tp, typename _Up>
  bool operator<=(const cloned_ptr<_Tp>& x, const cloned_ptr<_Up>& y) {
    return !(y < x);
  }
  
  template<typename _Tp, typename _Up>
  bool operator>(const cloned_ptr<_Tp>& x, const cloned_ptr<_Up>& y) {
    return y < x;
  }
      
  template<typename _Tp, typename _Up>
  bool operator>=(const cloned_ptr<_Tp>& x, const cloned_ptr<_Up>& y) {
    return !(x < y);
  }
      
  template<typename _Tp, typename _Up>
  bool operator==(const cloned_ptr<_Tp>& x, std::nullptr_t) {
    return !x;
  }
      
  template <typename _Tp, typename _Up>
  bool operator==(std::nullptr_t, const cloned_ptr<_Tp>& x) {
    return !x;
  }
      
  
  template<typename _Tp, typename _Up>
  bool operator!=(const cloned_ptr<_Tp>& x, std::nullptr_t) {
    return (bool) x;
  }
      
  template<typename _Tp, typename _Up>
  bool operator!=(std::nullptr_t, const cloned_ptr<_Tp>& x) {
    return (bool) x;
  }
      
      
 template<typename _Tp, typename _Up>
  bool operator<(const cloned_ptr<_Tp>& x, std::nullptr_t) {
    return std::less<std::common_type<_Tp*, _Up*>>()(x.get(), nullptr);
  }
      
  template<typename _Tp, typename _Up>
  bool operator<(std::nullptr_t, const cloned_ptr<_Up>& y) {
    return std::less<std::common_type<_Tp*, _Up*>>()(nullptr, y.get());
  }
      
  template<typename _Tp, typename _Up>
  bool operator<=(const cloned_ptr<_Tp>& x, std::nullptr_t) {
    return !(nullptr < x);
  }
      
  template<typename _Tp, typename _Up>
  bool operator<=(std::nullptr_t, const cloned_ptr<_Up>& y) {
    return !(y < nullptr);
  }
      
  template<typename _Tp, typename _Up>
  bool operator>=(const cloned_ptr<_Tp>& x, std::nullptr_t) {
    return !(x < nullptr);
  }
      
  template<typename _Tp, typename _Up>
  bool operator>=(std::nullptr_t, const cloned_ptr<_Up>& y) {
    return !(nullptr < y);
  }

    
}    //    namespace duct

#endif
