/*  Copyright (C) 2002-2017 Zoltán Nagy
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __duct_cxx_concepts_h__
#define __duct_cxx_concepts_h__ 1


/*   std includes   */
#include <utility>
#include <type_traits>
#include <complex>
#include <concepts>


namespace duct {


  /*  Concept of real numbers.
   *  We have only float, double and long double available.
   */
  template<typename _Tp>
  concept real_numbers = std::floating_point<_Tp>;

  /*   Concept of complex numbers.
   *   We have only the three std::complex variants here.
   */
  template<typename _Tp>
  concept complex_numbers = (std::same_as<std::complex<float>, std::remove_cv_t<_Tp>>
                             || std::same_as<std::complex<double>, std::remove_cv_t<_Tp>>
                             || std::same_as<std::complex<long double>, std::remove_cv_t<_Tp>>);
  
  
  /*   Concept for real and complex numbers.     */
  template<typename _Tp>
  concept complex_and_real_numbers = real_numbers<_Tp> || complex_numbers<_Tp>;
  
  
  
  template<typename _Tp, typename _Xp>
  using result_of_add_t = std::decay_t<decltype(std::declval<_Tp>()+std::declval<_Xp>())>;
  
  template<typename _Tp, typename _Xp>
  using result_of_sub_t = std::decay_t<decltype(std::declval<_Tp>()-std::declval<_Xp>())>;
  
  template<typename _Tp, typename _Xp>
  using result_of_mul_t = std::decay_t<decltype(std::declval<_Tp>()*std::declval<_Xp>())>;
  
  template<typename _Tp, typename _Xp>
  using result_of_div_t = std::decay_t<decltype(std::declval<_Tp>()/std::declval<_Xp>())>;
  
 

  
  /*     tagging dirac spinors    */
  struct bra {};
  struct ket {};
  
  
  /*   concept for making sure that we use either bra or ket as template argument   */
  template<class _Tp>
  concept BraKetTags = std::is_same<_Tp, bra>::value || std::is_same<_Tp, ket>::value;
  
  
  /*  bra => ket and ket => bra   */
  template<BraKetTags _Tag>
  using adjoint_tag_of = std::conditional_t<std::is_same<_Tag, bra>::value, ket, bra>;

  
}   //   namespace duct

#endif
