/*  Copyright (C) 2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 3/31/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __duct_cxx_memory_resource_h__
#define __duct_cxx_memory_resource_h__ 1


/*   std includes   */
#include <memory>
#include <type_traits>
#include <memory_resource>


/*   project includes    */
#include <cxx/stackvector.h>




/*    Unique pointer with polymorphic allocator     */
namespace duct {
  
  /*   simple name space alias to the std::pmr library   */
  namespace pmr = std::pmr;

}  //   namespace duct





/*   Simple and fast memory resource to build big dynamic objects    */
namespace duct {
  
  
  class memory_block
  {
    /*   Private types    */
    using _Size = std::size_t;
    
  public:
    /*    default constructor   */
    memory_block()
    : _M_begin{nullptr}, _M_current{nullptr}, _M_end{nullptr} {}
    
    
    /*   constructor    */
    memory_block(char *b, char *e)
    : _M_begin{b}, _M_current{b}, _M_end{e} {}
    
    memory_block(void *b, _Size bytes)
    : _M_begin{reinterpret_cast<char *>(b)}
    , _M_current{reinterpret_cast<char *>(b)},
    _M_end{reinterpret_cast<char *>(b)+bytes} {}
    
    
    
    /*   access to the memory block   */
    void * begin() const { return _M_begin;}
    void * current() const { return _M_current;}
    void * end() const { return _M_end;}
    
    
    /*   size infos   */
    _Size size() const { return _M_end-_M_begin;}
    _Size available() const { return _M_end-_M_current;}
    _Size used() const { return _M_current-_M_begin;}
    
    
    /*   returns the avaible space on the current buffer    */
    _Size available(_Size alignment, _Size& wasted) const
    {
      const _Size align = alignment - 1u;
      const _Size align_mask = ~align;
      
      const _Size addr = reinterpret_cast<_Size>(_M_current);
      const _Size aligned_addr = (addr + align) & align_mask;
      
      wasted = static_cast<_Size>(aligned_addr - addr);
      
      return (available() <= wasted ? 0u : available() - wasted);
    }
    
    
    /*    remaining storage on the current buffer    */
    _Size available(_Size alignment) const {
      _Size dummy;
      return available(alignment, dummy);
    }
    
    
    /*    allocate memory on this block.
     *    if there is no enough space, return nullptr
     */
    void * allocate(_Size bytes, _Size alignment)
    {
      std::size_t aligner = 0u;
      
      if(!_M_begin || available(alignment, aligner) < bytes)
        return nullptr;
      
      /*    we have sufficient storage on this bloc      */
      char * ptr = _M_current + aligner;
      
      _M_current = ptr + bytes;
      
      return ptr;
    }
    
    
    /*    reset the current pointer to the begining   */
    void reset() { _M_current = _M_begin;}
    
  private:
    /*    layout of the block   */
    char *_M_begin, *_M_current, *_M_end;
  };
  
  
  
  
  class monotonic_buffer_resource : public pmr::memory_resource
  {
    /*   Private types    */
    using _Size = std::size_t;
    
    static constexpr _Size _S_initial_next_buffer_size = 32u*sizeof(void*);
    
  public:
    /*    default constructor    */
    explicit monotonic_buffer_resource(pmr::memory_resource *upstream = nullptr) noexcept
    : _M_upstream{upstream ? upstream : pmr::get_default_resource()}
    , _M_next_size{_S_initial_next_buffer_size}
    , _M_has_external{false}
    {
      _M_blocks.emplace_back(_M_upstream->allocate(_M_next_size), _M_next_size);
      _M_current = _M_blocks.begin();
    }
    
    /*  destructor   */
    virtual ~monotonic_buffer_resource() { release();}
    
    /*    constructors    */
    explicit monotonic_buffer_resource(std::size_t initial_size, pmr::memory_resource* upstream = nullptr) noexcept
    : _M_upstream{upstream ? upstream : pmr::get_default_resource()}
    , _M_next_size{std::max(_S_initial_next_buffer_size, initial_size)}
    , _M_has_external{false}
    {
      _M_blocks.emplace_back(_M_upstream->allocate(_M_next_size), _M_next_size);
      _M_current = _M_blocks.begin();
    }
    
    
    monotonic_buffer_resource(void* buffer, std::size_t buffer_size, pmr::memory_resource* upstream = nullptr) noexcept
    : _M_upstream{upstream ? upstream : pmr::get_default_resource()}
    , _M_next_size{std::max(_S_initial_next_buffer_size, buffer_size)}
    , _M_has_external{true}
    {
      _M_blocks.emplace_back(buffer, buffer_size);
      _M_current = _M_blocks.begin();
    }
    
    
    /*   copy is deleted   */
    monotonic_buffer_resource(const monotonic_buffer_resource&) = delete;
    monotonic_buffer_resource operator=(const monotonic_buffer_resource&) = delete;
    
    
    
    /*    release all allocated memory   */
    void release() noexcept
    {
      for(_Size k = (_M_has_external ? 1u : 0u); k < _M_blocks.size(); k++) {
        if(_M_blocks[k].begin()) _M_upstream->deallocate(_M_blocks[k].begin(), _M_blocks[k].size());
          }
    }
    
    /*    returns the upstream memory resource   */
    pmr::memory_resource * upstream_resource() const noexcept { return _M_upstream;}
    
    
    /*   resets the buffers
     *   It doens't deallocate the memory block, it just resets the current pointer
     *   on every blocks. All the stored objects will be overwriten when we make new
     *   allocations.
     */
    void reset() noexcept
    {
      for(auto& block : _M_blocks)
        block.reset();
        
      _M_current = _M_blocks.begin();
    }
    
    /*   iterating over the memory blocks and do something (read only access)  */
    template<class _Callback>
    void info(_Callback&& doit) const
    {
      for(const auto& block : _M_blocks)
        doit(block);
    }
    
  protected:
    /*     allocate and deallocate    */
    void * do_allocate(_Size bytes, _Size alignment)
    {
      void *ptr = _M_current->allocate(bytes, alignment);
      
      /*   it was succesful   */
      if(ptr) return ptr;
      
      
      /*   try to find space on other blocks   */
      for(auto ib = _M_blocks.begin(); ib < _M_blocks.end(); ++ib)
        if(ib != _M_current)
        {
          ptr = ib->allocate(bytes, alignment);
          
          if(ptr)
          {
            if(_M_current->available() < ib->available())
              _M_current = ib;
            
            return ptr;
          }
        }
      
      
      /*   there is no space on other blocks, make a new block  */
      _M_increase_next_buffer_size(bytes);
      _M_blocks.emplace_back(_M_upstream->allocate(_M_next_size), _M_next_size);
      _M_increase_next_buffer_size();
      _M_current = _M_blocks.end()-1;
      
      
      return _M_current->allocate(bytes, alignment);
    }
    
    /*    deallocate does nothing    */
    void do_deallocate(void *, std::size_t, std::size_t) noexcept {}
    
    /*   compare memory resources   */
    bool do_is_equal(const pmr::memory_resource& other) const noexcept {
      return this == &other;
    }
    
    
  private:
    /*   upstream memory resource    */
    pmr::memory_resource *_M_upstream;
    
    
    /*    current buffer    */
    memory_block *_M_current;
    
    
    /*   we can have max 64 block. this is not a limitiation
     *   since the block size has geometric progression
     */
    stackvector<memory_block, 64u> _M_blocks;
    
    /*   size of the next buffer    */
    _Size _M_next_size;
    
    
    /*   true if we have external block    */
    bool _M_has_external;
    
    
  private:
    /*   increase the next buffer size    */
    void _M_increase_next_buffer_size() {
      _M_next_size = (static_cast<_Size>(-1)/2 < _M_next_size ? static_cast<_Size>(-1)/2 : 2u*_M_next_size);
    }
    
    
    /*    increase next buffer size at least to a given size    */
    void _M_increase_next_buffer_size(_Size n)
    {
      if(_M_next_size < n)
      {
        _Size tmp = n & (n-1);
        
        /*   it is power of 2   */
        if(!tmp || static_cast<_Size>(-1)/2 < n) {
          _M_next_size = n;
          return;
        }
        
        
        /*  find the next power of 2   */
        while(tmp) {
          n = tmp;
          tmp = n & (n-1);
        }
        
        _M_next_size = n << 1;
      }
    }
  };
  
}      //   namespace duct


#endif
