/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 10/1/18.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __cxx_random_h__
#define __cxx_random_h__ 1



/*   std includes   */
#include <random>
#include <utility>


/*   project headers   */


namespace duct {
  
  /*  I use a fixed random engine type. This way I can avoid to carry an extra
   *  template parameter. This can be modified only here and it requires to
   *  recomplie the whole library.
   */
  using random_engine_type = std::mt19937_64;

  
  
  /*  This is simple function to select an element from a list.
   *  The probability weight of a given elemnt is provided by the _UnaryOretion type argument.
   */
  template<class _RandomEngine, class _InputIt, class _UnaryOperation>
  inline _InputIt select_randomly(_RandomEngine& rne, _InputIt first, _InputIt last, _UnaryOperation op)
  {
    double rho = std::uniform_real_distribution<double>{0.0,1.0}(rne);
    double oprob = op(*first);
    
    while(first != last && oprob < rho) oprob += op(*(++first));
    
    return first;
  }

  
  /*   This is just like previous function but the random number is provided as argument.   */
  template<class _InputIt, class _UnaryOperation>
  inline _InputIt select_randomly(_InputIt first, _InputIt last, double rho, _UnaryOperation op)
  {
    double oprob = op(*first);
    while(first != last && oprob < rho) oprob += op(*(++first));
    return first;
  }

  
  
  /*  This is simple function to select  elements from a list.
   *  The probability weight of a given element is provided by the _UnaryOperation type argument.
   */
  template<typename _InputIt>
  struct select_signed_pair_randomly_result_t
  {
    _InputIt selectedP, selectedM;
    double normP, normM;
  };
  
  
  template<class _RandomEngine, class _InputIt, class _UnaryOperation>
  select_signed_pair_randomly_result_t<_InputIt>
  select_signed_pair_randomly(_RandomEngine& rne, _InputIt first, _InputIt last, _UnaryOperation op)
  {
    double rho = std::uniform_real_distribution<double>{0.0,1.0}(rne);
    
    /*   normalizations    */
    double normP = 0.0, normM = 0.0;
    
    for(auto it = first; it != last; ++it)
      if(double v = op(*it); v > 0.0) normP += v;
      else normM -= v;
    

    /*   select a positive weight    */
    auto ip = first;
    
    if(normP > 0.0)
    {
      double oprob = std::max(op(*ip), 0.0), r = rho*normP;
      
      while(ip != last && oprob < r)
        if(auto w = op(*(++ip)); w > 0.0)
          oprob += w;
    }
    else ip = last;
    
    
    /*   select a negative weight    */
    auto im = first;
    
    if(normM > 0.0)
    {
      double oprob = -std::min(op(*im), 0.0), r = rho*normM;
    
      while(im != last && oprob < r)
      if(auto w = op(*(++im)); w < 0.0)
        oprob -= w;
    }
    else im = last;

    
    return {ip, im, normP, normM};
  }
  
}

#endif
