/*  Copyright (C) 2012-2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 5/3/12.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */

//   std includes
#include <cmath>
#include <limits>
#include <complex>

//   project headers
#include "proc-dyjets.h"
#include "bits/qcd-color.h"


namespace duct {
  
  //----- FIXME: these constats should be gotten from the model description file -----
  constexpr energy<double> Mz = 91.182_GeV, Gz = 2.4952_GeV;
  constexpr energy_sqr<double> M2z = Mz*Mz;

  constexpr double s2w = 0.231;
  constexpr double swcw = 0.84294484;   // 2.0*std::sqrt(s2w*(1.0-s2w));
  
  //----- electroweak couplings ------
  constexpr double Ql = -1.0, Vl = (-0.5 + 2.0*s2w)/swcw, Al = -0.5/swcw;
  constexpr double Qu = 2.0/3.0, Vu = (0.5 - 4.0*s2w/3.0)/swcw, Au = 0.5/swcw;
  constexpr double Qd = -1.0/3.0, Vd = (-0.5 + 2.0*s2w/3.0)/swcw, Ad = -0.5/swcw;

  
  
  dyjets::dyjets(const model_ref& mdl, energy<double> Ecm, energy<double> mL, energy<double> mH,
                 double sfac, int hdr1, int hdr2,
                 std::initializer_list<int> lpts, std::initializer_list<int> partons)
  : _M_mdl{mdl}, _M_Ecm{Ecm}, _M_scalefac{sfac*sfac}, _M_lpts(lpts), _M_partons(partons)
  {
    /*   incoming hadrons   */
    _M_hadronA = hdr1;
    _M_hadronB = hdr2;
    
    /*    the pdf functions   */
    _M_pdf1 = _M_mdl.pdf(qcd::order::lo, hdr1);
    _M_pdf2 = _M_mdl.pdf(qcd::order::lo, hdr2);

    /*    tau limits   */
    auto s = _M_Ecm*_M_Ecm;
    double tauL = mL*mL/s, tauH = mH*mH/s;
    
    /*   coupling part of the normalization    */
    double Ng = (Qu*Qu+Qd*Qd)*Ql*Ql;
    double Nz = (std::pow((Vu+Vl)*(Au+Al), 2)+std::pow((Vd+Vl)*(Ad+Al), 2))*std::log(s/M2z);
   
    /*   the full normalization   */
    double pL = -Mz/Gz*std::atan((M2z-s*tauL)/(Mz*Gz));
    double pH = -Mz/Gz*std::atan((M2z-s*tauH)/(Mz*Gz));
    double lL = std::pow(std::log(tauL),2);
    double lH = std::pow(std::log(tauH),2);
    
    double Ngz = 0.5*(lL-lH)*Ng + (pH-pL)*Nz;
    double pg = 0.5*(lL-lH)*Ng/Ngz;
    
    /*   the tau probability density    */
    this->_M_rhotau = [=](double tau) -> double {
      auto pp = s*tau - M2z;
      return (Nz*s*M2z/(pp*pp + M2z*Gz*Gz) - Ng*std::log(tau)/tau)/Ngz;
    };
    
    /*   generating tau   */
    this->_M_generate_tau = [=](double rho, double chi) -> double {
      if(rho < pg) return std::exp(-std::sqrt(chi*lH + (1.0-chi)*lL));
      return Mz/s*(Mz - Gz*std::tan(-Gz/Mz*(pL+chi*(pH-pL))));
    };
  }

  
  double dyjets::_M_select_flavors(shower_stage<hhc>& hard, double tau, double eta, double x, double rho)
  {
    double norm = 0.0, __fa[11], __fb[11];
    double *fa = __fa+5, *fb = __fb+5;
    const energy_sqr<double> &m2F = hard.scale;
    
    /*   precalculating the pdfs  */
    for(int i : _M_partons) {
      fa[ i] = _M_pdf1->parton( i, eta, m2F);
      fb[-i] = _M_pdf2->parton(-i, tau/eta, m2F);
    }

    /*   calculating the matrix element square with the pdfs  */
    unsigned ip = 0u;
    struct { int fq, fl; double w;} proc[50];
    
    for(int fl : _M_lpts) {
      /*   processes q + qbar -> l + lbar and
       *   processes qbar + q -> l + lbar   
       */
      for(int iq : _M_partons)
        if(iq != 0 && -6 < iq && iq < 6) {
          proc[ip].fq = -iq;
          proc[ip].fl =  fl;
          norm += (proc[ip++].w = fa[iq]*fb[-iq]*_M_process_qqll(iq > 0 ? x : 1.0-x, tau, iq, fl));
        }
    }
    
    //----- selecting subprocess ----
    double cproc = proc[0].w;
    
    ip = 0U;
    while(cproc < rho*norm)
      cproc += proc[++ip].w;
 
    /*   Assigning the flavours   */
    hard.state[-3].flavor = -proc[ip].fl;
    hard.state[-2].flavor =  proc[ip].fl;
    hard.state[-1].flavor =  proc[ip].fq;
    hard.state[ 0].flavor = -proc[ip].fq;
   
    
    return norm/tau;   //  We have a 1/tau because the pdfs are the eta*f(eta,mu2) functions.
  }

  
  double dyjets::_M_process_qqll(double x, double tau, int fq, int fl) const
  {
    /*    masses invariants    */
    energy_sqr<double> s = _M_Ecm*_M_Ecm, m2l = _M_mdl.mass2(fl);
    energy_sqr<double> sab = s*tau;
    energy_sqr<double> sa1 = x*sab;
    
    /*  trivial cuts  */
    if(x*(1.0-x) <= m2l/sab) return 0.0;
      
    /*    Z propagator factor   */
    std::complex<double> chi = 1.0/std::complex<double>((sab-M2z)/sab, Mz*Gz/sab);
    std::complex<double> A, B, C, D;
    
    /*    calulate EW coupling factor   */
    switch(std::abs(fq)) {
      //----- down-like quarks -----
      case 1: case 3: case 5: {
        A = Ql*Qd + Vd*Vl*chi;
        B = Vd*Al*chi;
        C = Ad*Vl*chi;
        D = Ad*Al*chi;
      } break;
        
      //------ up-like quarks -----
      case 2: case 4: case 6: {
        A = Ql*Qu + Vu*Vl*chi;
        B = Vu*Al*chi;
        C = Au*Vl*chi;
        D = Au*Al*chi;
      } break;
    }
    
    /*   derived dot products   */
    auto sa2 = sab-sa1;
    auto X1 = sa1*sa1+sa2*sa2, X3 = 2.0*m2l*sab;
    
    return ((std::norm(A) + std::norm(C))*(X1 + X3) + (std::norm(B) + std::norm(D))*(X1 - X3)
            + 2.0*std::real(A*std::conj(D)+B*std::conj(C))*(sa2*sa2 - sa1*sa1))/(sab*sab);
  }
  
  
  void dyjets::generate(random_engine_type& random, shower_stage<hhc>& hard, shower_vetos& vetos)
  {
    std::uniform_real_distribution<double> rng(0.0,1.0);
  
    /*   resizing the state, setting the model  */
    hard.state.resize(2,0);

    
    /*    incoming hadrons   */
    hard.state.hadron[-1].momentum = {0.5*_M_Ecm, 0.0_GeV,0.0_GeV, 0.5*_M_Ecm};
    hard.state.hadron[ 0].momentum = {0.5*_M_Ecm, 0.0_GeV,0.0_GeV,-0.5*_M_Ecm};
    hard.state.hadron[-1].key = _M_hadronA;
    hard.state.hadron[ 0].key = _M_hadronB;

    /*   selecting tau, eta and x   */
    double tau = _M_generate_tau(rng(random), rng(random));
    double eta = tau*std::pow(1.0/tau, rng(random));
    double etb = tau/eta, x = rng(random);
    auto s = _M_Ecm*_M_Ecm;
    auto shat = s*tau;
    
    /*    hard scale    */
    hard.scale = _M_scalefac*shat;
    
    /*   selecting the flavor   */
    hard.weights.born.main = _M_select_flavors(hard, tau, eta, x, rng(random));
    hard.weights.born.main /= _M_rhotau(tau);
    hard.weights.born.main *= -std::log(tau)/tau;
    hard.weights.nlo.main = 0.0;

    /*   For the outgoing momenta we use Sudakov paremetrization.   */
    auto m2l = _M_mdl.mass2(hard.state[-2].flavor);
    auto mT = sqrt(x*(1.0-x)*shat - m2l);
    
    /*      + and - momenta      */
    auto na = hard.state.hadron[-1].momentum*eta;
    auto nb = hard.state.hadron[ 0].momentum*etb;
    
    /*     reconstruction the momenta   */
    double phi = 2.0*M_PI*rng(random);
    lorentzvector<energy<double>> kT{0.0_GeV, mT*std::cos(phi), mT*std::sin(phi), 0.0_GeV};
    
    hard.state[-3].momentum =      x *na + (1.0-x)*nb - kT;
    hard.state[-2].momentum = (1.0-x)*na +      x *nb + kT;
    hard.state[-1].momentum = na;
    hard.state[ 0].momentum = nb;
    
    /*   color connections    */
    using colorlink = parton_cstate::colorlink;
    auto& cstate = hard.state;
    
    cstate[-3].colorL = cstate[-3].colorR = colorlink::dead();
    cstate[-2].colorL = cstate[-2].colorR = colorlink::dead();
    cstate[-1].colorL = cstate[-1].colorR = (hard.state[-1].flavor > 0 ? colorlink::quark( 0) : colorlink::antiquark( 0));
    cstate[ 0].colorL = cstate[ 0].colorR = (hard.state[ 0].flavor > 0 ? colorlink::quark(-1) : colorlink::antiquark(-1));
    
    /*   color suppression of the state and color weight  */
    cstate.color_weight = 1.0;

    
    /*   relaxing the vetos   */
    vetos = shower_vetos{};
    
    /*   alpha_qed coupling   */
    //double aqed = _M_mdl.alpha_qed(shat);
    
    double alpha_MZ = 1.0/127.944;     // From PDG Chapter 10, page 4
    double betamod = 20.0/9.0/M_PI;    // 1/3/pi * (1 + 1 + 1 + (4/9 + 4/9 + 1/9 + 1/9 + 1/9)*Nc)
    double aqed = alpha_MZ/(1.0 - alpha_MZ*betamod*std::log(shat/M2z));
    
    /*    adjusting the weight     */
    hard.weights.born.main *= 2.0*M_PI*aqed*aqed*1.0_GeV2/(s*qcd::Nc);
  }
  

}











