/*  Copyright (C) 2012-2013 Zoltn Nagy
 *  Created by Zoltn Nagy on 16/2/12.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */

//   std includes
#include <cmath>
#include <limits>
#include <complex>

//   project headers
#include <proc-epajets.h>


namespace duct {
  
  epajets::epajets(const model_ref& mdl, energy<double> Ecm)
  : _M_Ecm(Ecm), _M_mdl(mdl)
  {
    //----- FIXME: these constats should be gotten from the model description file -----
    constexpr double s2w = 0.231;
    constexpr energy<double> Mz = 91.182_GeV, Gz = 2.4952_GeV;

    //---- Z propagator factor -----
    energy_sqr<double> s = _M_Ecm*_M_Ecm;
    std::complex<double> chi = 1.0/std::complex<double>((s-Mz*Mz)/s, Mz*Gz/s)/(4.0*s2w*(1.0-s2w));
    double rchi = std::real(chi), schi = std::norm(chi);
    
    //---- electroweek couplings -----
    double gev = -0.5 + 2.0*s2w, gea = -0.5;
    double guv = 0.5 - 4.0*s2w/3.0, gua = 0.5;
    double gdv = -0.5 + 2.0*s2w/3.0, gda = -0.5;
    double ge2 = gev*gev+gea*gea;
    
    double uvv = 4.0/9.0 - 4.0*gev*guv/3.0*rchi + ge2*guv*guv*schi, uaa = ge2*gua*gua*schi; 
    double dvv = 1.0/9.0 + 2.0*gev*gdv/3.0*rchi + ge2*gdv*gdv*schi, daa = ge2*gda*gda*schi; 
    
    //----- normalized cross sections ----
    double p[] = {dvv+daa, uvv+uaa, dvv+daa, uvv+uaa, dvv+daa};
    
    //----- commulative probabilities -----
    int n = 0; _M_cp[0] = p[0];
    while(++n < 5) {
      _M_cp[n] = p[n];
      _M_cp[n] += _M_cp[n-1];
    }
  }
  
  void epajets::generate(random_engine_type& random, shower_stage<epa>& hard, shower_vetos& vetos)
  {
    std::uniform_real_distribution<double> rng(0.0,1.0);
        
    double phi = (2.0*M_PI)*rng(random);
    double c = 2.0*rng(random)-1.0;
    double s = std::sqrt(1.0-c*c);
    double r = _M_cp[4]*rng(random);
    
    /*    select the quark flavor   */
    int n = 0; while(_M_cp[n++] < r);

    energy<double> E = 0.5*_M_Ecm;
    energy<double> px = E*s*std::cos(phi), py = E*s*std::sin(phi);

    
    /*    resizing the state    */
    hard.state.resize(0,2);
 
    
    /*       flavours       */
    hard.state[-1].flavor =  11;
    hard.state[ 0].flavor = -11;
    hard.state[ 1].flavor =  n;
    hard.state[ 2].flavor = -n;
    
    /*       momenta       */
    hard.state[-1].momentum = {E, 0.0_GeV,0.0_GeV,-E};
    hard.state[ 0].momentum = {E, 0.0_GeV,0.0_GeV, E};
    hard.state[ 1].momentum = {E,  px,  py,  E*c};
    hard.state[ 2].momentum = {E, -px, -py, -E*c};
    
    /*    color connections    */
    hard.state[-1].colorL = hard.state[-1].colorR = parton_cstate::colorlink::dead();
    hard.state[ 0].colorL = hard.state[ 0].colorR = parton_cstate::colorlink::dead();

    insert_open_stringL({1,2}, hard.state);
    insert_open_stringR({1,2}, hard.state);
    
    /*      color suppression of the state    */
    hard.state.color_weight = 1.0;

    /*    weight    */
    hard.weights.born.main = _M_cp[4];
    hard.weights.nlo.main = 0.0;
    
    /*    starting scale of the shower    */
    hard.scale = sqr(_M_Ecm);

    /*   vetos    */
    vetos = shower_vetos{};
  }
  
  
}


