/*  azimuth-duct.cc
 *  Created by Zoltan Nagy on 20.08.20.
 *  Copyright © 2020 Zoltán Nagy. All rights reserved.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes    */
#include <cmath>

/*   project headers    */
#include "bits/rs-shower-scheme.h"



namespace duct {
  namespace resum {
    
    namespace __details {
    
      constexpr double __one_pi = M_PI;
      constexpr double __two_pi = 2.0*M_PI;
      
       
      azimuthal_angle_duct::azimuthal_angle_duct(const resummation_event<epa>& p, const resum_splitting_vars& zeta)
      {
        /*   for two parton sate we don't have azimuthal correlation   */
        if(2 == p.upper()) {
          is_simple = true;
          return;
        }
        
        /*    aliases   */
        const long double &x0 = zeta.x0, &y = zeta.y, &u = zeta.z, &xi = zeta.xi;
        
        /*    standard variables    */
        long double lmd = (1.0-x0)/(1.0+x0)*(1.0+y);
        
        /*    basic variables   */
        is_simple = false;
        B = x0/(x0 + u*(1.0-x0));
        A = 2.0*(xi+(1.0-xi)*x0);
        C = xi*(1.0-u) + (1.0-xi)*u*x0;
        S = 2.0*std::sqrt(std::abs(xi*(1.0-xi)*u*(1.0-u)*x0));
        
        /*    more variables    */
        long double Cm = xi*(1.0-u) - (1.0-xi)*u*x0;
        long double D = std::sqrt(B*B + 2.0*C*B + Cm*Cm);
        
        I = (A + 2.0*B)/D;
        R = (B+C+S)/D;
        
        
        /*  pure collinear part for quark emitter  */
        if(p[zeta.emitter].flavor)
        {
          long double h = (sqr(lmd+y-1.0) + 4.0*y)/(2.0*lmd);
          long double u1 = 1.0-u+x0*u, gam = (1.0+y+lmd)/2.0;
          
          Coll = -2.0 + h*(-1.0 + (1.0 + x0 - x0/u1)/u1) + gam*(1.0-u);
        }
        
        /*   and for gluon emitter   */
        else
        {
          long double Dlt = 2.0*x0/((x0+u*(1.0-x0))*(1.0-u*(1.0-x0)));
          
          Coll = -2.0 + u*(1.0-u)*(1.0 - 0.5*Dlt*(2.0-Dlt));
        }
      }
      
      
      long double azimuthal_angle_duct::function_phi(long double phi) const
      {
        long double cphi = std::cos(__two_pi*phi);
        long double ret = (A + 2.0*B)/(B+C - S*cphi) + Coll;
        
        return ret;
      }
      
      
      long double azimuthal_angle_duct::primitive_phi(long double phi) const
      {
        phi *= __two_pi;
        
        /*   if phi is 2pi then we have simpler expression   */
        if(std::abs(1.0-phi/__two_pi) < 1e-13 || phi >= __two_pi)
          return I - 2.0;
        
        /*   phi is smaller than 2pi   */
        double ctg = std::tan(0.5*phi), one = phi <= __one_pi ? 0.0 : __one_pi;
        double ret = I*(one + std::atan(R*ctg)) + 0.5*Coll*phi;
        
        return ret/__one_pi;
      }
      
      
      long double azimuthal_angle_duct::inverse_phi(long double rho) const
      {
        /*    special case is very simple   */
        if(is_simple) return rho*__two_pi;
        
        /*   otherwise we have to use Newton method to find the root   */
        long double lo = 0.0, hi = 1.0, mi = rho;
        long double mi1=-10.0, mi2=-10.0, u, v, prim;
        unsigned int iter = 0U;
        
        //----- normalization ----
        rho *= primitive_phi(1.0);
        
        do {
          //----- start with Newton method -----
          mi2 = mi1; mi1 = mi;
          mi += (rho - (prim = primitive_phi(mi)))/function_phi(mi);
          
          if(mi2 > 0.0) {
            u = mi1-mi2;
            v = mi-2.0*mi1+mi2;
            if(v != 0.0) mi = mi2 - u*u/v;
          }
          
          //---- adjusting the boundaries to the old value -----
          if(prim > rho) { if(mi1 < hi) hi = mi1;}
          else { if(mi1 > lo) lo = mi1;}
          
          //----- if it doesn't work we use bisect method-----
          if(mi < 0.0 || mi > 1.0) mi = 0.5*(lo+hi);
        } while(std::abs(mi1-mi) > 1e-8*mi1 && ++iter < 30U);
        
        return mi*__two_pi;
      }
      
    }  //  namespace __details
    
    
    
  } //  namespace resum
} //namespace duct

