/*  rs-shower-scheme.h
 *  Created by Zoltan Nagy on 20.01.20.
 *  Copyright © 2020 Zoltán Nagy. All rights reserved.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __rs_shower_scheme_h__
#define __rs_shower_scheme_h__ 1


/*   std includes    */
#include <ratio>
#include <cxx/random.h>


/*   project headers    */
#include <bits/rs-resum-stage.h>




namespace duct {
  namespace resum {
  
    /*  Splitting variables
     *  This class is a collection of variables for reconstracting
     *  the event after a splitting in a real splitting.
     */
    struct resum_splitting_vars
    {
      /*  splitting type : 1 => 1st stage, 20-25 => a term, 3 => b term real, 4 => bterm virtual   */
      int type;

      /*  label of the emitter and the possible spectator   */
      int emitter, spectator;

      /*  flavor splitting variable   */
      int fl, fm;

      /*  momentum splitting variables   */
      long double x0, y, z, phi;

      /*    the jacobian   */
      long double jac;

      /*   some hard parameters   */
      long double xi, xl, ymax;
    };


    /*   This concept defines the most essention properties of a shower schele that we need
     *   for testing resummation properties.
     */
    template<class _Proc, class _Scheme>
    concept ShowerScheme = ProcessTag<_Proc> &&
    requires(const long double *r, resummation_event<_Proc>& q,
             const resummation_event<_Proc>& p, const resum_splitting_vars& zeta)
    {
      typename _Scheme::ordering_type;
      
      {_Scheme::splitting_function(p, p, zeta)} -> std::same_as<long double>;
      {_Scheme::ordering(p, zeta)} -> std::same_as<long double>;
      {_Scheme::mapping(q, p, zeta)};
      {_Scheme::safety_cuts(p, zeta)} -> std::same_as<bool>;
      {_Scheme::event_vars(p, zeta)} -> std::same_as<resum_event_vars>;

      /*    Requirements for the splitting variable generation    */
      {_Scheme::generate_vars(r, int{1}, int{2}, p)} -> std::same_as<resum_splitting_vars>;
    };

  }
}


/*
 *     Deductor schemes
 */
namespace duct {
  namespace resum {
    

    /*   Helper functions and classes    */
    namespace __details {

      /*   calss for generating the azimuthal angle */
      class azimuthal_angle_duct
      {
      public:
        /*   nontrivial constructor   */
        azimuthal_angle_duct(const resummation_event<epa>&, const resum_splitting_vars&);
        
        long double function_phi(long double) const;
        long double primitive_phi(long double) const;
        long double inverse_phi(long double) const;
 
      private:
        bool is_simple;
        long double A, B, C, S, I, R, Coll;
      };

      
      /*    momentum mapping for the final state
       *     (only massless partons!)
       */
      template<ProcessTag _Proc>
      void momentum_mapping_massless_fin(resummation_event<_Proc>& p, const resummation_event<_Proc>& q,
                                         const resum_splitting_vars& vars)
      {
        /*    making space for the extra parton   */
        p = q;
        p.resize_by(1);
        
        /*   iterators   */
        auto p1 = begin<qcd_out>(p), pn = end<qcd_out>(p)-1;
        auto pl = p1+vars.emitter-1;
        

        /*    y dependent variables    */
        const long double &y = vars.y, &z = vars.z, &x0 = vars.x0;
        const long double Fg = (1.0+y)/(1.0+x0);
        
        /*    momenta kl, nl and Pl    */
        const lorentzvector<long double> Q = momentum_sum<incomings>(p);
        const lorentzvector<long double> &kl = pl->momentum, nl = vars.xl*Q - pl->momentum;
        const lorentzvector<long double> Pl = Fg*(kl + x0*nl);
        
        
        /*    Lorentz boost   */
        const long double op = y*(1.0+x0)/(1.0-x0*y), om = -y/Fg, nk = kl*nl;
        
        auto boost = [=](const lorentzvector<long double>& qnull) -> lorentzvector<long double> {
          return qnull + kl*(op*(nl*qnull)/nk) + nl*(om*(kl*qnull)/nk);
        };
        
        /*    boosting the non-qcd momenta    */
        for(auto& ki : range<non_qcd>(p))
          ki.momentum = boost(ki.momentum);
        
        /*    boosting the qcd momenta    */
        for(auto pi = p1; pi != pn; ++pi)
          if(pi != pl) pi->momentum = boost(pi->momentum);
        
        
        /*    the transverse momentum in the c.m. of Pl   */
        long double mT = std::sqrt(std::abs(z*(1.0-z)*x0*2.0*nk));
        lorentzvector<long double> kT{0.0, mT*std::cos(vars.phi), mT*std::sin(vars.phi), 0.0};
        
        /*    boosting the transverse momentum back to the original frame    */
        lorentzvector<long double> ql = kl, qk = p[vars.spectator].momentum;
        threevector<long double> bvec = Q.boostVector();
        
        ql.boost(-bvec);
        long double rPhi = ql.phi(), rTheta = ql.theta();
        
        if(vars.spectator != vars.emitter) {
          qk.boost(-bvec);
          qk.rotate(3,-rPhi);
          qk.rotate(2, -rTheta);
          kT.rotate(3, qk.phi());
        }
        
        kT.rotate(2, rTheta);
        kT.rotate(3, rPhi);
        kT.boost(bvec);
        
        /*    the daughter momenta & flavors    */
        pl->flavor = vars.fl;
        pl->momentum = Fg*(z*kl + (1.0-z)*x0*nl - kT);
        
        pn->flavor = vars.fm;
        pn->momentum = Pl-pl->momentum;
      }
    
      
      /*   deductor splitting functions   */
      long double splitting_function_full(const resummation_event<epa>&, const resum_splitting_vars&);
      long double splitting_function_approx(const resummation_event<epa>&, const resum_splitting_vars&);
      long double splitting_function_test(const resummation_event<epa>&, const resum_splitting_vars&);

      /*   generate variables for deductor phase space mapping   */
      resum_splitting_vars generate_vars_duct(const long double *, int, int, const resummation_event<epa>&, long double);
      resum_splitting_vars generate_vars_duct_kT(const long double *, int, int, const resummation_event<epa>&, long double);

    }   //   namespace __details
  


    /*   Class for defining the deductor shower scheme for testing resummation properties.
     *   This is just a declaration because at the moment we have some expression only for
     *   e+e- processes. So, we define it via specialization.
     */
    template<ProcessTag, OrderingTag, unsigned = 0u>
    struct deductor_scheme;
    
   
    
    template<OrderingTag _Ordering, unsigned _ApproxLevel>
    requires (_ApproxLevel < 3u)
    struct deductor_scheme<epa, _Ordering, _ApproxLevel>
    {
      /*   public types   */
      using ordering_type = _Ordering;
      
      /*   splitting function   */
      static long double splitting_function(const resummation_event<epa>&, const resummation_event<epa>& p,
                                            const resum_splitting_vars& zeta)
      {
        /*   simplified splitting function (not positive definite) */
        if constexpr(_ApproxLevel == 1u)
          return __details::splitting_function_approx(p, zeta);
    
        /*   simple splitting function with angular cut    */
        if constexpr(_ApproxLevel == 2u)
          return __details::splitting_function_test(p, zeta);
        
        /*   full deductor splitting function   */
        return __details::splitting_function_full(p, zeta);
      }
      
      
      /*   the momentum and flavor mapping   */
      static void mapping(resummation_event<epa>& q, const resummation_event<epa>& p, const resum_splitting_vars& zeta) {
        __details::momentum_mapping_massless_fin(q, p, zeta);
      }
      
      
      /*   provides some safety cuts    */
      static constexpr bool safety_cuts(const resummation_event<epa>&, const resum_splitting_vars& zeta)
      {
        /*   angular ordering   */
        if constexpr(std::is_same_v<_Ordering, ordering::angular>) {
          const long double kT = zeta.z*(1.0-zeta.z)*zeta.y*zeta.xl;
          return (kT > 1.0e-5L);
        }
        
        return true;
      }
      

      /*   generate the splitting variables with some importance sampling   */
      static resum_splitting_vars generate_vars(const long double *r, int l, int k, const resummation_event<epa>& p)
      {
        /*   lambda ordering   */
        if constexpr(std::is_same_v<_Ordering, ordering::lambda>)
          return __details::generate_vars_duct(r, l, k, p, p.scale);
        
        if constexpr(std::is_same_v<_Ordering, ordering::kT>)
          return __details::generate_vars_duct_kT(r, l, k, p, p.scale);
        
        /*     angular ordering    */
        return __details::generate_vars_duct(r, l, k, p, 1.0L);
      }
      
      
      /*   defines the odreding variable   */
      static long double ordering(const resummation_event<epa>&, const resum_splitting_vars& zeta)
      {
        /*    kT ordering   */
        if constexpr(std::is_same_v<_Ordering, ordering::kT>)
          return zeta.z*(1.0-zeta.z)*zeta.y*zeta.xl;
        
        
        /*   angular ordering   */
        if constexpr(std::is_same_v<_Ordering, ordering::angular>)
        {
          const long double &x0 = zeta.x0, &z = zeta.z;
          return x0/((1.0L-z+x0*z)*(z+x0*(1.0L-z)));
        }
        
        /*    lambda ordering    */
        return zeta.y;
      }

      /*   event variables to study approximated observable   */
      static resum_event_vars event_vars(const resummation_event<epa>&, const resum_splitting_vars& zeta)
      {
        resum_event_vars vars;
        
        /*   mapping parameters   */
        vars.y = zeta.y;
        vars.z = zeta.z;
        vars.phi = zeta.phi;
        vars.emitter = zeta.emitter;
        vars.spectator = zeta.spectator;
 
        /*    scale of the strong coupling   */
        vars.asscale = (1.0-zeta.z)*zeta.y*zeta.xl;
        
        if constexpr(std::is_same_v<_Ordering, ordering::kT>)
          vars.asscale *= zeta.z;
        
        /*   derived quantities   */
        vars.lambda = zeta.y;
        vars.kappaT = zeta.z*(1.0-zeta.z)*zeta.y*zeta.xl;

        return vars;
      }
    };
    
  }   //   namespace resum
}   //   namespace duct






/*
 *     Catani-Seymour schemes
 */
namespace duct {
  namespace resum {
    
    /*   Helper functions and classes    */
    namespace __details {
      
      template<ProcessTag _Proc>
      lorentzvector<long double> cs_tranverse_new_fff(const resummation_event<_Proc>& p, int l, int k, long double phi)
      {
        const auto &pl = p[l].momentum, &pk = p[k].momentum;
        const auto nl = unit(pl), nk = unit(pk);
        
        /*   we have only two partons   */
        if(2 == p.upper())
        {
          auto plk = pl+pk, qk = pk;
          auto bvec = plk.boostVector();
          auto kT = lorentzvector<long double>{0.0, std::cos(phi), std::sin(phi), 0.0};
          
          qk.boost(-bvec);
          kT.rotate(2, qk.theta());
          kT.rotate(3, qk.phi());
          kT.boost(bvec);
          kT *= -mag(plk);
          
          return kT;
        }
        
        /*   special case when the dipole is almost back to back   */
        if(mag(cross(nl,nk)) < 1.0e-12)
        {
          auto kT = lorentzvector<long double>{0.0, std::cos(phi), std::sin(phi), 0.0};
          auto mlk = nk-nl;
          
          kT.rotate(2, mlk.theta());
          kT.rotate(3, mlk.phi());
          kT *= -mag(pl+pk);
          
          return kT;
        }
        
        /*    general case     */
        const long double c2lk = 1.0 + dot(nl, nk);
        const long double clk = std::sqrt(c2lk), E = std::sqrt(2.0*pl[0]*pk[0]);
        const long double cphi = E*std::cos(phi), sphi = E*std::sin(phi);

        return {clk*cphi, ((nl+nk)*cphi + cross(nl,nk)*sphi)/clk};
      }
 
      
      template<ProcessTag _Proc>
      lorentzvector<long double> cs_tranverse_fff(const resummation_event<_Proc>& p, int l, int k, long double phi)
      {
        const auto &pl = p[l].momentum, &pk = p[k].momentum;
        
        /*   we have only two partons   */
        auto plk = pl+pk, qk = pk;
        auto bvec = plk.boostVector();
        auto kT = lorentzvector<long double>{0.0, std::cos(phi), std::sin(phi), 0.0};
        
        qk.boost(-bvec);
        kT.rotate(2, qk.theta());
        kT.rotate(3, qk.phi());
        kT.boost(bvec);
        kT *= -mag(plk);
        
        return kT;
      }
      

      
      /*    momentum mapping for the final state
       *     (only massless partons!)
       */
      template<ProcessTag _Proc>
      void catani_seymour_mapping_fff(resummation_event<_Proc>& p, const resummation_event<_Proc>& q,
                                      const resum_splitting_vars& vars)
      {
        /*    making space for the extra parton   */
        p = q;
        p.resize_by(1);
        
        /*   iterators   */
        auto p1 = begin<qcd_out>(p), pn = end<qcd_out>(p)-1;
        auto pl = p1+vars.emitter-1;
        
        
        /*   splitting variables   */
        const long double &y = vars.y, &z = vars.z;
        const long double oz = 1.0-z;
        
        /*   make the emission happen   */
        auto kT = std::sqrt(z*oz*y)*cs_tranverse_fff(q, vars.emitter,vars.spectator, vars.phi);

        /*   the spectator parton gets rescaled   */
        p[vars.spectator].momentum *= 1.0-y;
        
        /*   daughter partons   */
        pl->flavor = vars.fl;
        pl->momentum = z*q[vars.emitter].momentum + oz*y*q[vars.spectator].momentum + kT;
        
        pn->flavor = vars.fm;
        pn->momentum = oz*q[vars.emitter].momentum + z*y*q[vars.spectator].momentum - kT;;
      }
      
      
      /*   generate the splitting variables with some importance sampling   */
      resum_splitting_vars generate_vars_cs(const long double *, const long double&);
      resum_splitting_vars generate_vars_cs(const long double *, int, int, const resummation_event<epa>&, const long double&);
      resum_splitting_vars generate_vars_cs_kT(const long double *, int, int, const resummation_event<epa>&, const long double&);
    }   //   namespace __details
    
    
    
    /*   Class for defining the Catani-Seymour shower scheme for testing resummation properties.
     *   This is just a declaration because at the moment we have some expression only for
     *   e+e- processes. So, we define it via specialization.
     */
    template<ProcessTag, OrderingTag>
    struct catani_seymour_scheme;
    
    
    
    /*   Class for defining the deductor scheme for e+e- annihilation.
     */
    template<OrderingTag _Ordering>
    struct catani_seymour_scheme<epa, _Ordering>
    {
      /*   public types   */
      using ordering_type = _Ordering;

 
      /*   splitting function   */
      static long double splitting_function(const resummation_event<epa>&, const resummation_event<epa>& p,
                                            const resum_splitting_vars& zeta)
      {
        return 2.0 - (1.0-zeta.z*(1.0-zeta.y))*(p[zeta.emitter].flavor ? 1.0+zeta.z : 2.0-zeta.z*(1.0-zeta.z));
      }
      
      
      /*   the momentum and flavor mapping   */
      static void mapping(resummation_event<epa>& q, const resummation_event<epa>& p, const resum_splitting_vars& zeta) {
        __details::catani_seymour_mapping_fff(q, p, zeta);
      }
      

      /*   generate the splitting variables with some importance sampling   */
      static resum_splitting_vars generate_vars(const long double *r, int l, int k, const resummation_event<epa>& p)
      {
        if constexpr(std::is_same_v<_Ordering, ordering::lambda>)
          return __details::generate_vars_cs(r, l, k, p, p.scale);
        
        if constexpr(std::is_same_v<_Ordering, ordering::kT>)
          return __details::generate_vars_cs_kT(r, l, k, p, p.scale);

        
        /*    for other ordering   */
        auto zeta = __details::generate_vars_cs(r, 1.0);
        
        zeta.emitter = l;
        zeta.spectator = k;
        
        zeta.fl = p[l].flavor;
        zeta.fm = 0;  //  we have only gluon emission (at least at the moment)
        
        return zeta;
      }

      
      /*   provides some safety cuts    */
      static bool safety_cuts(const resummation_event<epa>&, const resum_splitting_vars&) {
        return true;
      }
 
      
      /*   defines the ordering variable   */
      static long double ordering(const resummation_event<epa>& p, const resum_splitting_vars& zeta)
      {
        const long double &y = zeta.y, &z = zeta.z;
        const auto Q = momentum_sum<incomings>(p);
        const auto plk = p[zeta.emitter].momentum + p[zeta.spectator].momentum;
        const long double Q2 = mag2(Q);
        
        /*    kT ordering   */
        if constexpr(std::is_same_v<_Ordering, ordering::kT>)
          return y*z*(1.0-z)*mag2(plk)/Q2;
        
        /*   angular ordering   */
        if constexpr(std::is_same_v<_Ordering, ordering::angular>)
        {
          const long double xl = 2.0*(p[zeta.emitter].momentum*Q)/Q2;
          return y/((1.0-z*(1.0-y))*(z+y*(1.0-z)))*sqr(xl);
        }
        
        /*    lambda ordering    */
        return 0.5*y*mag2(plk)/(p[zeta.emitter].momentum*Q);
      }

      /*   event variables to study approximated observable   */
      static resum_event_vars event_vars(const resummation_event<epa>& p, const resum_splitting_vars& zeta)
      {
        resum_event_vars vars;
        
        /*   mapping parameters   */
        vars.type = zeta.type;
        vars.y = zeta.y;
        vars.z = zeta.z;
        vars.phi = zeta.phi;
        vars.emitter = zeta.emitter;
        vars.spectator = zeta.spectator;


        /*   derived quantities   */
        const auto Q = momentum_sum<incomings>(p);
        const auto plk = p[zeta.emitter].momentum + p[zeta.spectator].momentum;
        const long double slk = mag2(plk), Q2 = mag2(Q);
        
        vars.lambda = zeta.y*0.5*slk/(p[zeta.emitter].momentum*Q);
        vars.kappaT = zeta.z*(1.0-zeta.z)*zeta.y*slk/Q2;
        
        /*    scale of the strong coupling   */
        vars.asscale = (1.0-zeta.z)*zeta.y*slk/Q2;
  
        if constexpr(std::is_same_v<_Ordering, ordering::kT>)
          vars.asscale *= zeta.z;

        return vars;
      }
    };

    
    
  }   //   namespace resum
}   //   namespace duct


    
/*
 *     PanLocal schemes
 */
namespace duct {
  namespace resum {
        
    
    
    /*   Class for defining the PanLocal shower scheme for testing resummation properties.
     *   This is just a declaration because at the moment we have some expression only for
     *   e+e- processes. So, we define it via specialization.
     */
    template<ProcessTag, typename>
    struct panlocal_scheme;
    
    
    
    /*   Class for defining the deductor scheme for e+e- annihilation.
     */
    template<std::intmax_t _Num, std::intmax_t _Den>
    class panlocal_scheme<epa, std::ratio<_Num, _Den>>
    {
      /*   private types   */
      using _Beta = std::ratio<_Num, _Den>;

    public:
      /*   public types   */
      using ordering_type = duct::ordering::kT;
      using cs_scheme = catani_seymour_scheme<epa, duct::ordering::kT>;
      

      /*   splitting function   */
      static long double splitting_function(const resummation_event<epa>&, const resummation_event<epa>& p,
                                            const resum_splitting_vars& zeta)
      {
        const long double &z = zeta.z, &y = zeta.y;
        const long double oz = 1.0-z, oy = 1.0-y;

        const auto Q = momentum_sum<incomings>(p);
        const long double sl = 2.0*(p[zeta.emitter].momentum*Q), sk = 2.0*(p[zeta.spectator].momentum*Q);
        long double eta = 0.5*std::log(sl/sk*(1.0/z-1.0)/y);
        long double eta2 = eta*eta;

        /*   don't allow leaking to the other hemisphere   */
        if(eta <= -1.0) return 0.0;
        
        /*    rapidity cut   */
//        constexpr double beta = static_cast<double>(_Beta::num)/static_cast<double>(_Beta::den);
//        const double slk = mag2(p[zeta.emitter].momentum + p[zeta.spectator].momentum), Q2 = mag2(Q);
//        double k2T = z*(1.0-z)*y*slk;
//
//        if(2.0*std::abs(eta) + std::log(k2T) + beta*std::log(Q2*slk/(sl*sk)) > 0.0)
//          return 0.0;
        
        
        /*   the emission is in the allowed rapidity range   */
        long double g = (eta < 1.0L ? (eta*(eta2*(3.0L*eta2-10.0L) + 15.0L) + 8.0L)/16.0L : 1.0L);
        long double sfunc =  2.0L*z/(1.0L-z) + (p[zeta.emitter].flavor ? oz : z*oz);
        //long double sfunc =  2.0*z*oy/(g*oz*oy + (1.0-g)*y) + (p[zeta.emitter].flavor ? oz : z*oz);

        return (1.0-z*oy)*g*sfunc;
      }
      
      
      /*   the momentum and flavor mapping   */
      static void mapping(resummation_event<epa>& q, const resummation_event<epa>& p, const resum_splitting_vars& zeta) {
         return __details::catani_seymour_mapping_fff(q, p, zeta);
      }
      
      
      /*   generate the splitting variables with some importance sampling   */
      static resum_splitting_vars generate_vars(const long double *r, int l, int k, const resummation_event<epa>& p)
      {
        constexpr long double beta = static_cast<long double>(_Beta::num)/static_cast<long double>(_Beta::den);
        return __details::generate_vars_cs_kT(r, l, k, p, std::pow(p.scale, 1.0L/(1.0L+beta)));
      }
      
      
      /*   provides some safety cuts    */
      static bool safety_cuts(const resummation_event<epa>&, const resum_splitting_vars&) {
        return true;
      }
      
      
      /*   defines the ordering variable   */
      static long double ordering(const resummation_event<epa>& p, const resum_splitting_vars& zeta)
      {
        const auto Q = momentum_sum<incomings>(p);
        const auto plk = p[zeta.emitter].momentum + p[zeta.spectator].momentum;
        const long double Q2 = mag2(Q), slk = mag2(plk);
       
        /*   ordering variable   */
        double v2 = zeta.y*zeta.z*(1.0-zeta.z)*slk/Q2;
        
        if constexpr(_Beta::num)
        {
          const long double sl = 2.0*(p[zeta.emitter].momentum*Q), sk = 2.0*(p[zeta.spectator].momentum*Q);
          constexpr long double beta = static_cast<long double>(_Beta::num)/static_cast<long double>(_Beta::den);
          
          v2 *= std::pow(Q2*slk/(sl*sk)*std::exp(-std::abs(std::log(sl/sk*(1.0/zeta.z-1.0)/zeta.y))), beta);
        }
        
        return v2;
      }
      
      /*   event variables to study approximated observable   */
      static resum_event_vars event_vars(const resummation_event<epa>& p, const resum_splitting_vars& zeta) {
        return cs_scheme::event_vars(p, zeta);
      }
    };
    
  }   //   namespace resum
}   //   namespace duct
    
    
 
/*
 *     Deductor local schemes
 */
namespace duct {
  namespace resum {
        
    
    
    
    /*   Class for defining the Catani-Seymour shower scheme for testing resummation properties.
     *   This is just a declaration because at the moment we have some expression only for
     *   e+e- processes. So, we define it via specialization.
     */
    template<ProcessTag, OrderingTag>
    struct ductlocal_scheme;
    
    
    
    /*   Class for defining the deductor scheme for e+e- annihilation.
     */
    template<OrderingTag _Ordering>
    struct ductlocal_scheme<epa, _Ordering>
    {
      /*   public types   */
      using ordering_type = _Ordering;
      using cs_scheme = catani_seymour_scheme<epa, _Ordering>;

 
      /*   splitting function   */
      static long double splitting_function(const resummation_event<epa>&, const resummation_event<epa>& p,
                                            const resum_splitting_vars& zeta)
      {
        const long double &z = zeta.z, &y = zeta.y, oz = 1.0-zeta.z;
        long double ret = 2.0*z;
        
        /*   trival case   */
        if(p.upper() == 2) {
          ret *= (z + y*oz)/(y + oz*(z + y*oz));
        }
        
        /*   more the two partons in the final state   */
        else {
          const auto Q = momentum_sum<incomings>(p);
          const auto plk = p[zeta.emitter].momentum + p[zeta.spectator].momentum;
          const long double Q2 = mag2(Q), slk = mag2(plk);
          const long double sl = 2.0*(p[zeta.emitter].momentum*Q), sk = 2.0*(p[zeta.spectator].momentum*Q);
          
          long double Ap = z*sl + oz*y*sk, Am = z*sl - oz*y*sk;
          long double D = std::sqrt(sqr(oz)*(sqr(Am)+ 4.0*z*oz*y*Q2*slk) + 2.0*y*oz*sk*Ap + sqr(y*sk));
          
          ret *= oz*(sqr(Am)+ 4.0*z*oz*y*Q2*slk) + 2.0*y*sk*Ap;
          ret /= D;
          ret /= y*sk + D;
        }
        
        /*    pure collinear part   */
        ret += (p[zeta.emitter].flavor ? oz : z*oz);

        
        return (1.0-z*(1.0-y))*ret;
      }
      
      
      /*   the momentum and flavor mapping   */
      static void mapping(resummation_event<epa>& q, const resummation_event<epa>& p, const resum_splitting_vars& zeta) {
        __details::catani_seymour_mapping_fff(q, p, zeta);
      }
      
      
      /*   generate the splitting variables with some importance sampling   */
      static resum_splitting_vars generate_vars(const long double *r, int l, int k, const resummation_event<epa>& p) {
        return cs_scheme::generate_vars(r, l,k, p);
      }
      
      
      /*   provides some safety cuts    */
      static bool safety_cuts(const resummation_event<epa>&, const resum_splitting_vars&) {
        return true;
      }
      
      
      /*   defines the ordering variable   */
      static long double ordering(const resummation_event<epa>& p, const resum_splitting_vars& zeta) {
        return cs_scheme::ordering(p, zeta);
      }
      
      /*   event variables to study approximated observable   */
      static resum_event_vars event_vars(const resummation_event<epa>& p, const resum_splitting_vars& zeta) {
        return cs_scheme::event_vars(p, zeta);
      }
    };
    
  }   //   namespace resum
}   //   namespace duct
    
    
#endif
