/*  shower-scheme-duct.cc
 *  Created by Zoltan Nagy on 21.01.20.
 *  Copyright © 2020 Zoltán Nagy. All rights reserved.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes    */
#include <cmath>


/*   project headers    */
#include "cxx/math-support.h"
#include "bits/rs-shower-scheme.h"



namespace duct {
  namespace resum {
  
    /*   Helper functions and classes    */
    namespace __details {

      
      long double splitting_function_full(const resummation_event<epa>& p, const resum_splitting_vars& zeta)
      {
        /*    aliases   */
        const long double &x0 = zeta.x0, &y = zeta.y, &u = zeta.z, &xi = zeta.xi;
        
        
        /*   special case is simple   */
        long double ret = 0.0, lmd = (1.0-x0)/(1.0+x0)*(1.0+y);
        
        if(p.upper() == 2)
          ret = 2.0*u/(x0/(x0 + u*(1.0-x0)) + 1.0-u);
        
        /*   general case   */
        else
        {
          /*    basic variables   */
          long double A = 2.0*(xi+(1.0-xi)*x0);
          long double B = x0/(x0 + u*(1.0-x0));
          long double Cp = xi*(1.0-u) + (1.0-xi)*u*x0;
          long double Cm = xi*(1.0-u) - (1.0-xi)*u*x0;

          ret = (A + 2.0*B)/std::sqrt(B*B + 2.0*Cp*B + Cm*Cm) - 2.0;
        }
        
        
        /*  pure collinear part for quark emitter  */
        if(p[zeta.emitter].flavor)
        {
          long double h = (sqr(lmd+y-1.0) + 4.0*y)/(2.0*lmd);
          long double u1 = 1.0-u+x0*u, gam = (1.0+y+lmd)/2.0;
          ret += h*(-1.0 + (1.0 + x0 - x0/u1)/u1) + gam*(1.0-u);
        }
        
        /*   and for gluon emitter   */
        else
        {
          long double Dlt = 2.0*x0/((x0+u*(1.0-x0))*(1.0-u*(1.0-x0)));
          ret += u*(1.0-u)*(1.0 - 0.5*Dlt*(2.0-Dlt));
        }
        
        return ret;
      }
      
      
      long double splitting_function_approx(const resummation_event<epa>& p, const resum_splitting_vars& zeta)
      {
        /*    aliases   */
        const long double &x0 = zeta.x0, &z = zeta.z, &xi = zeta.xi;
        
        /*   soft part   */
        long double ret = 2.0*xi/std::sqrt(sqr(xi)*(1.0-z)*(1.0-z+2.0*x0)+sqr(x0*(2.0-xi))) - 2.0;
        
        /*  pure collinear part   */
        ret += p[zeta.emitter].flavor ? 1.0-z : z*(1.0-z);
        
        return ret;
      }
      
      
      long double splitting_function_test(const resummation_event<epa>& p, const resum_splitting_vars& zeta)
      {
        /*    aliases   */
        const long double &x0 = zeta.x0, &z = zeta.z;
        /*   angular ordering condition   */
        if(zeta.xi*(1.0-z+x0*z)*(z+x0*(1.0-z)) < x0)
          return 0.0;

        return 2.0/(1.0-z + x0) - 2.0 + (p[zeta.emitter].flavor ? 1.0-z : z*(1.0-z));
      }
      
      
   
      
      static resum_splitting_vars generate_vars(const long double *rng, const long double& xl, long double xi, const long double& yL)
      {
        const long double ymax = xl/sqr(1.0+std::sqrt(std::abs(1.0-xl)));
        const long double x0max = (yL < ymax ? 4.0/xl*yL/sqr(1.0+yL+std::sqrt(std::abs((1.0-yL/ymax)*(1.0-ymax*yL)))) : 1.0);

        /*  parameters to control the importance sampling    */
        long double epsilon = 1.0e-9L, beta = 4.0;

        while(1.0e-15L + std::abs(x0max*epsilon) < 1.0e-9L)
          if((epsilon *= 10.0L) > 1.0e-3L) break;
        
               
        const long double eps = epsilon*x0max, lmax = -std::log(x0max);
        const long double leps = -std::log(eps), beta2 = beta*beta;
        const long double n1 = 0.5*(leps*leps-lmax*lmax), n2 = beta*(beta + leps);


        /*   generate the x0 variable  */
        long double x0, jac = n1+n2;

        if(*(rng++) < n1/jac) {
          x0 = std::exp(-std::sqrt(leps*leps - 2.0*(*(rng++))*n1));
          jac /= -std::log(x0);
        }
        else
        {
          long double r = -(*(rng++))*n2/beta2;
          long double fe = M_E*std::pow(eps, -1.0/beta);

          x0 = eps*std::pow(r/lambert_Wm1(r/fe), beta);
          jac /= std::pow(x0/eps, 1.0/beta);
          jac /= -std::log(x0);
        }

        /*    y variable  */
        long double t = 4.0*x0*xl/sqr(1.0+x0);
        long double y = t/sqr(1.0+std::sqrt(1.0-t));
        long double lmd = (1.0-x0)/(1.0+x0)*(1.0+y);

        /*    jacobian of y -> x0 and the phase space jacobian (a factor of lmd) */
        jac *= (1.0-x0)/std::sqrt(1.0-t)/(1.0+x0);
        jac *= lmd;

        /*   lorentz transform of the xi parameter   */
        xi = xi/(xi+(1.0-xi)*sqr(1.0+2.0*y/(1.0-y+lmd)));


        /*   generate the z variable   */
        long double zint = std::log(0.5*(xi*(1.0+x0)+std::sqrt(sqr(xi*(1.0+x0))+sqr(2.0*x0)*(1.0-xi)))/x0);
        long double r = 2.0*x0*std::exp((1.0-*(rng++))*zint);
        long double z = 1.0+x0 - 0.5*r/xi + 2.0*sqr(x0)/r*(1.0/xi-1.0);

        jac *= zint*std::sqrt(sqr(xi)*(1.0-z)*(1.0-z+2.0*x0)+sqr(x0*(2.0-xi)))/xi;


        /*   generating the phi variable   */
        long double phi = 2.0*M_PI*(*(rng++));

        /*   returning the splitting variables and the jacobian  */
        resum_splitting_vars zeta;

        zeta.x0 = x0;
        zeta.y = y;
        zeta.z = z;
        zeta.phi = phi;
        zeta.jac = jac;
 
        zeta.xi = xi;
        zeta.xl = xl;
        zeta.ymax = ymax;
        
        
        return zeta;
      }


  
      resum_splitting_vars generate_vars_duct(const long double *r, int l, int k, const resummation_event<epa>& p, long double yL)
      {
        /*   xl and xi variable   */
        auto Q = p[-1].momentum+p[0].momentum;
        
        long double xl = 2.0L*(Q*p[l].momentum)/mag2(Q);
        long double xi = (p[l].momentum*p[k].momentum)/(Q*p[k].momentum)/xl;
        
        /*    generate variables   */
        auto zeta = __details::generate_vars(r+1, xl, xi, yL);
        
        zeta.emitter = l;
        zeta.spectator = k;
        
        zeta.fl = p[l].flavor;
        zeta.fm = 0;  //  we have only gluon emission (at least at the moment)
        
        
        /*   generating the phi variable   */
        zeta.phi = azimuthal_angle_duct{p, zeta}.inverse_phi(*r);
        
        
        return zeta;
      }
      
     
    
      /*==================================================================================================================*/
    
      
      static resum_splitting_vars generate_vars_kT(const long double *rng, const long double& xl, long double xi, const long double& yT)
      {
        const long double ymax = xl/sqr(1.0+std::sqrt(std::abs(1.0-xl)));
        const long double xT = (yT < ymax ? 4.0/xl*yT/sqr(1.0+yT+std::sqrt(std::abs((1.0-yT/ymax)*(1.0-ymax*yT)))) : 1.0);

        /*  parameters to control the importance sampling    */
        long double epsilon = 1.0e-9L, beta = 4.0;
        
        while(1.0e-15L + std::abs(xT*epsilon) < 1.0e-9L)
          if((epsilon *= 10.0L) > 1.0e-3L) break;
        
        const long double eps = epsilon*xT;
        const long double leps = -std::log(eps), beta2 = beta*beta;
        const long double n1 = 0.5*leps*leps, n2 = beta*(beta + leps);
        
        
        /*   generate the x0 variable  */
        long double x0, jac = n1+n2;
        
        if(*(rng++) < n1/jac) {
          x0 = std::exp(-std::sqrt(leps*leps - 2.0*(*(rng++))*n1));
          jac /= -std::log(x0);
        }
        else
        {
          long double r = -(*(rng++))*n2/beta2;
          long double fe = M_E*std::pow(eps, -1.0/beta);
          
          x0 = eps*std::pow(r/lambert_Wm1(r/fe), beta);
          jac /= std::pow(x0/eps, 1.0/beta);
          jac /= -std::log(x0);
        }
        
        /*    y variable  */
        long double t = 4.0*x0*xl/sqr(1.0+x0);
        long double y = t/sqr(1.0+std::sqrt(1.0-t));
        long double lmd = (1.0-x0)/(1.0+x0)*(1.0+y);
        
        /*    jacobian of y -> x0 and the phase space jacobian (a factor of lmd) */
        jac *= (1.0-x0)/std::sqrt(1.0-t)/(1.0+x0);
        jac *= lmd;
        
        /*   lorentz transform of the xi parameter   */
        xi = xi/(xi+(1.0-xi)*sqr(1.0+2.0*y/(1.0-y+lmd)));
        
        
        /*   generate the z variable when x0 < xT  */
        long double z, r, rz = *(rng++);
        
        if(x0 < xT)
        {
          long double zint = std::log(0.5*(xi*(1.0+x0)+std::sqrt(sqr(xi*(1.0+x0))+sqr(2.0*x0)*(1.0-xi)))/x0);
          r = 2.0*x0*std::exp(rz*zint);
          jac *= zint;
        }
        
        /*   generate the z variable when 1 > x0 > xT  */
        else
        {
          auto F = [=](const long double z) -> long double {
            const long double oz = 1.0L-z;
            return -std::log(xi*(oz+x0) + std::sqrt(sqr(xi)*oz*(oz+2.0L*x0)+sqr(x0*(2.0L-xi))));
          };
          
          const long double zp = 0.5L*(1.0L+std::sqrt(1.0L-yT/y));
          const long double zm = 1.0L-zp, F0 = F(0.0L), F1 = F(1.0L);
          const long double Nm = F(zm)-F0, Np = F1-F(zp);
          
          /*    generating z in [0,zm]   */
          r = std::exp(*(rng++) < Nm/(Nm+Np) ? -F0-Nm*rz : -F1+Np*rz);
          jac *= Nm+Np;
        }
        
        z = 1.0+x0 - 0.5*r/xi + 2.0*sqr(x0)/r*(1.0/xi-1.0);
        jac *= std::sqrt(sqr(xi)*(1.0-z)*(1.0-z+2.0*x0)+sqr(x0*(2.0-xi)))/xi;
        
        /*   generating the phi variable   */
        long double phi = 2.0*M_PI*(*(rng++));
        
        /*   returning the splitting variables and the jacobian  */
        resum_splitting_vars zeta;
        
        zeta.x0 = x0;
        zeta.y = y;
        zeta.z = z;
        zeta.phi = phi;
        zeta.jac = jac;
        
        zeta.xi = xi;
        zeta.xl = xl;
        zeta.ymax = ymax;

        
        return zeta;
      }
      
      
      
      resum_splitting_vars generate_vars_duct_kT(const long double *r, int l, int k, const resummation_event<epa>& p, long double kTprev)
      {
        /*   xl and xi variable   */
        auto Q = p[-1].momentum+p[0].momentum;
        
        const long double xl = 2.0L*(Q*p[l].momentum)/mag2(Q);
        const long double xi = (p[l].momentum*p[k].momentum)/(Q*p[k].momentum)/xl;
        
        /*    generate variables   */
        resum_splitting_vars zeta = __details::generate_vars_kT(r+1, xl, xi, 4.0L*kTprev/xl);;
      
        zeta.emitter = l;
        zeta.spectator = k;
        
        zeta.fl = p[l].flavor;
        zeta.fm = 0;  //  we have only gluon emission (at least at the moment)
        
        /*   generating the phi variable   */
        zeta.phi = azimuthal_angle_duct{p, zeta}.inverse_phi(*r);

        
        return zeta;
      }
      
  
      
    }   //   namespace __details
    
  }   //  namespace resum
}   //   namespace duct

